/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane;

import java.awt.Color;
import java.awt.Image;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.swing.ImageIcon;
import javax.swing.JApplet;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.UIManager;

import com.nilo.plaf.nimrod.NimRODLookAndFeel;

import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.ui.SplashScreen;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.utils.Tree;
import fr.greyc.pantheon.ariane.xml.XMLElement;
import fr.greyc.pantheon.ariane.xml.XMLFileInputStream;

/**
 * <code>Ariane</code> is the main class of the software.
 * It reads the configuration file and initializes the application.
 * It uses a Splash Screen to pass the time.
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Apr 11, 2000
 * @version Mathieu JOETS - Sep 07, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Oct 06, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 10, 2004 (Look&Feel Revision)
 * @version Régis CLOUARD - Jan 16, 2007 (add icon)
 * @version Frédéric LAIR - Sep 04 , 2007
 * @version Régis CLOUARD - Nov 25 , 2007 (new release)
 */
public class Ariane {
	/** The base name of the software ... en français dans le texte. */
	public static final String TITLE_BASE = "Au fil d'Ariane ...";

	/** The current version identifier. */
	public static final String VERSION_ID = "3.1.1 (Xxx xx, 2013)";

	/** The large list of contributors. */
	public static final String[] AUTHORS = { "Régis CLOUARD",
		"François RIOULT",
		"Sébastien LEFÈVRE (1999)",
		"Richard LOBRÉAU (1999)",
		"Adrien SCHRÈQUE (2000)",
		"Hicham AMADI (2000)",
		"Mathieu JOETS (2001)",
		"Guillaume CHATELET (2002)",
		"Alban DOUBLET (2002)",
		"Frédéric LAIR (2007)",
		"Look and Feel: Nilo J. GONZALEZ (see http://personales.ya.com/nimrod)"
	};

	///// 
	///// C O N S T A N T S
	/////

	/** The default extension for Ariane's filenames (aws::Ariane WorkSpace). */
	public static final String DEFAULT_FILE_EXT= ".aws";

	/** The name of operator library configuration file. */
	public static final String LIBRARY_CONFIG_FILENAME = "ariane.cfg";

	/** The name of the static configuration path: $HOME/.greyc/ariane. */
	private static final String CONFIG_STATIC_DIRNAME = ".greyc"+File.separator+"ariane";

	/** The name of the dynamic configuration path: $JARPATH/ariane */
	private static final String CONFIG_DYNAMIC_DIRNAME = "ariane";

	/** The name of the user preferences file. */
	private static final String PREFS_FILENAME = "ariane.ini";

	/** The name of the subroutines file. */
	private static final String SUBROUTINE_FILENAME = "subroutines.cfg";

	/** The name of the operator file. */
	private static final String OPERATOR_FILENAME = "operators.cfg";

	/** The main frame that support all other frames. */
	public static Workspace workspace;

	/** The operators library path. */
	private static String _libraryPath=".";

	/** The configuration path name. */
	private static String _configurePathName=null;

	/** 
	 * The documentation path. Each document is
	 * supposed located in _docPath/fr/ or _docPath/en/ or ...
	 */
	private static String _docPath="";

	/** The HOME path. */
	private static String _homePath;

	/** The jar path. */
	private static String _jarPath;

	/** List of paths to search when loading dynamic libraries. */
	private static String _javaLibraryPath = null;

	/** The history size for do/undo management. */
	public static final int HISTORYSIZE=10;

	/** The directory where are stored all icons. */
	public static final String iconDir = "fr/greyc/pantheon/ariane/properties/img/";

	/** The directory where are stored all resources. */
	private static final String resourceDir = "fr/greyc/pantheon/ariane/properties/locale/";

	/** The name of the resource file. */
	private static final String resourceFile = "messages";

	/** The ResourceBundle to load properties. */
	public static ResourceBundle lang;

	/** The URL for manual files. */
	public static String urlManual;

	/** The URL for properties files. */
	private static String url;

	/** The applet that supports all the Ariane interface. */
	private JApplet _applet=null;

	/** The frame that supports all the Ariane interface. */
	static public JFrame frame=null;

	/**
	 * Called only by the applet version.
	 * 
	 * @param applet the applet.
	 */
	public Ariane( ArianeApplet applet ) {
		applet.getContentPane().add(new JLabel("essai2"));
		_libraryPath=".";
		_homePath=".";

		_applet=applet;
		if (_homePath == null) _homePath=System.getProperty("user.home",".");
		String configFilename=getLibraryPath()+File.separator+"lib"+File.separator+LIBRARY_CONFIG_FILENAME;
		// The name of the input graph file if any.
		initializeInterface(false,configFilename,null,null);
	}

	/**
	 * Sets the global context variables
	 * homePath
	 * javaLibraryPath
	 * jarPath
	 */
	private void setGlobalContextVariables() {
		_homePath=System.getProperty("user.home",".");
		_javaLibraryPath =System.getProperty("java.library.path");
		// Tips: getting the absolute file name of the jar.
		Class<?> c = Ariane.class;
		URL tmp= c.getResource(c.getSimpleName() + ".class");
		String s = tmp.getFile();
		int xi = s.indexOf('!'); // Jar file uses the ! meta character
		if (xi>0) s = s.substring(0,xi);
		try {
			URL u = new URL(s);
			_jarPath = new File( u.toURI() ).getParent();
		} catch ( URISyntaxException e1 ) {
			_jarPath = new File( s).getParent();
		} catch ( MalformedURLException e2 ) {
			_jarPath = new File( s).getParent();
		} catch ( Exception e3 ) {
			_jarPath = new File( s).getParent();
		}
	}

	/**
	 * Called only by the application version.
	 * 
	 * @param args the input arguments list.
	 */
	public Ariane( String[] args ) {
		String operatorLibraryFilename = null;
		String configDirName=null;
		ArrayList<String> graphFilenames=new ArrayList<String>();

		setGlobalContextVariables();

		// Parse arguments
		boolean doSplash = true;
		for (int i=0; i < args.length; i++) {
			if (args[i].startsWith("-")) {
				if (args[i].equalsIgnoreCase("-version") || args[i].equalsIgnoreCase("-v")) {
					System.err.println("Ariane \""+VERSION_ID+"\"");
					System.exit(0);
				} else if (args[i].equalsIgnoreCase("-nosplash")) {
					doSplash=false;
				} else if (args[i].equalsIgnoreCase("-config")) {
					if (++i == args.length) {
						System.err.println("No configuration directory specified ... option -config ignored.");
					} else {
						configDirName=args[i];
					}
				} else if (args[i].equalsIgnoreCase("-lib")) {
					if (++i == args.length) {
						System.err.println("No operator library configuration file specified... option -lib ignored.");
					} else {
						operatorLibraryFilename=args[i];
					}
				} else {
					if (args[i].equalsIgnoreCase("-help") || args[i].equalsIgnoreCase("-h")){
						System.err.println("Usage: ariane [options] [fichier-graphe ..]");
						System.err.println("Options include: ");
						System.err.println("  -help -h -?	display this message");
						System.err.println("  -version -v	print product version and exit");
						System.err.println("  -nosplash	launch Ariane without the splash screen");
						System.err.println("  -lib <file>	use<fichier> instead of $PANDOREHOME/lib/pandore.cfg");
						System.exit(0);
					}else {
						System.err.println("Option '" + args[i] + "' unknown option, ignored...");
					}
				}
			} else {
				graphFilenames.add(args[i]);
			}
		}
		frame=new JFrame();
		initializeInterface(doSplash, operatorLibraryFilename, graphFilenames, configDirName);
	}

	/**
	 * Returns the name of the file that stores
	 * the preferences.
	 * 
	 * @return the preference file name.
	 */
	public static final String getPreferencesFilename() {
		return _configurePathName+File.separator+PREFS_FILENAME;
	}

	/**
	 * Returns the name of the file that stores
	 * the subroutines.
	 * 
	 * @return the absolute file name.
	 */
	public static final String getSubroutineFilename() {
		return _configurePathName+File.separator+SUBROUTINE_FILENAME;
	}

	/**
	 * Returns the name of the file that stores
	 * the personal operators.
	 * 
	 * @return the absolute file name.
	 */
	public static final String getPersonalOperatorFilename() {
		return _configurePathName+File.separator+OPERATOR_FILENAME;
	}

	/**
	 * Sets the new path to the library of operators
	 * and modify the related execution environment.
	 * 
	 * @param path the new path.
	 */
	public static final void setLibraryPath( String path ) {
		_libraryPath = path;
		System.setProperty("java.library.path",_javaLibraryPath + File.pathSeparatorChar + _libraryPath+File.separatorChar+"lib");
	}

	/**
	 * Returns the current library path.
	 * 
	 * @return the library path.
	 */
	public static final String getLibraryPath( ) {
		return _libraryPath;
	}

	/**
	 * Sets the new doc path.
	 * 
	 * @param path the new path.
	 */
	public static final void setDocPath( String path ) {
		_docPath = path;
	}

	/**
	 * Gets the current doc path.
	 * 
	 * @return the current doc path.
	 */
	public static final String getDocPath( ) {
		return _docPath+"/"+PreferenceChooser.locale.getLanguage();
	}

	/**
	 * Sets the current language to the specified language.
	 * Loads related resources form the resource files.
	 * 
	 * @param locale a locale (eg., Fr_fr, En_us).
	 */
	public static final void setLocale( Locale locale ) {
		url = resourceDir+ locale.getLanguage()+"/";
		lang = ResourceBundle.getBundle(url+resourceFile, locale);
		urlManual =url+"manual/";
	}

	/**
	 * Loads the configuration file.
	 * 
	 * @param configFile the file that contains the configuration.
	 * @return the XML tree of or null.
	 */
	private static final LinkedList<XMLElement> readConfigFile( File configFile ) {
		try {
			return readXMLFile(configFile);
		} catch(Exception e) {
			System.err.println(Ariane.lang.getString("error")+e.getLocalizedMessage());
			return null;
		}
	}

	/**
	 * Loads an XML file.
	 * 
	 * @param configFile the file that contains the configuration.
	 * @return the XML tree of or null.
	 * @throws IOException bad file format.
	 */
	public static final LinkedList<XMLElement> readXMLFile( File configFile ) throws IOException {
		try {
			XMLFileInputStream convert = new XMLFileInputStream(configFile);
			convert.parse();
			convert.close();
			return convert.getElements();
		} catch(Exception e) {
			throw new IOException(e.getLocalizedMessage());
		}
	}

	/**
	 * Locates the configuration file.
	 * The sequence of configuration files is:
	 * 0: the specified directory.
	 * 1: jarPath/ariane/ariane.ini
	 * 2: $HOME/.greyc/ariane/ariane.ini
	 * and for sake of backwards compatibility
	 * 4: jarPath/.ariane
	 * 4: jarPath/ariane.cfg
	 * 3: $HOME/.greyc/ariane.ini
	 * 5: $HOME/.greyc/ariane.cfg
	 * 6: $HOME/.ariane
	 *
	 * @param dirname a specified configuration directory or null.
	 * @return the configuration file or null.
	 */
	private File locateConfigFile( String dirname ) {
		File f;

		// 0. specified dir name
		if (dirname!=null) {
			f=new File(dirname+File.separator+PREFS_FILENAME);
			if (f.exists()) {
				_configurePathName = f.getAbsoluteFile().getParent();
				return f;
			}
		}

		// 1. jarPath/ariane/ariane.ini
		f=new File(_jarPath+File.separator+CONFIG_DYNAMIC_DIRNAME+File.separator+PREFS_FILENAME);
		if (f.exists()) {
			_configurePathName = f.getParent();
			return f;
		}

		// 2 jarPath/ariane.cfg
		f=new File(_jarPath+File.separator+"ariane.cfg");
		if (f.exists()) {
			_configurePathName = _jarPath+File.separator+CONFIG_DYNAMIC_DIRNAME;
			return f;
		}

		// 3. jarPath/.ariane
		f=new File(_jarPath+File.separator+".ariane");
		if (f.exists()) {
			_configurePathName = _jarPath+File.separator+CONFIG_DYNAMIC_DIRNAME;
			return f;
		}

		// 4 $HOME/.greyc/ariane/ariane.ini
		f=new File(_homePath+File.separator+CONFIG_STATIC_DIRNAME+File.separator+PREFS_FILENAME);
		if (f.exists()) {
			_configurePathName = f.getParent();
			return f;
		}

		// 5 $HOME/.greyc/ariane.cfg
		f=new File(_homePath+File.separator+".greyc"+File.separator+"ariane.cfg");
		if (f.exists()) {
			_configurePathName = _homePath+File.separator+CONFIG_STATIC_DIRNAME;
			return f;
		}

		// 6 $HOME/.ariane
		f=new File(_homePath+File.separator+".ariane");
		if (f.exists()) {
			_configurePathName = _homePath+File.separator+CONFIG_STATIC_DIRNAME;
			return f;
		}
		// default configuration file
		_configurePathName = _homePath+File.separator+CONFIG_STATIC_DIRNAME;
		return null;
	}

	/**
	 * Initializes the interface.
	 * 
	 * @param doSplash does the splash screen?
	 * @param operatorLibraryFilename an operators library configuration filename or null.
	 * @param graphFilenames a set of input graph filenames or null.
	 * @param configDirName a specified configuration directory.
	 */
	private void initializeInterface( boolean doSplash, String operatorLibraryFilename, ArrayList<String> graphFilenames, String configDirName ){
		SplashScreen splash=null;
		if (doSplash){
			splash = new SplashScreen(iconDir+"Splash.jpg");
			splash.showProgress("Loading Ariane settings...",0);
			splash.setVisible(true);
		}
		
		if (doSplash) splash.showProgress("Reading preferences file ...",10);
		PreferenceChooser.initPreferences();

		LinkedList<XMLElement> preferenceHandler=null;

		File configFile=locateConfigFile(configDirName);
		if (configFile!=null) {
			preferenceHandler=readConfigFile(configFile);
		}

		String defaultLibraryFilename = null;
		if (preferenceHandler!= null) {
			PreferenceChooser.load(preferenceHandler);
			defaultLibraryFilename=getLibraryPath()+File.separator+"lib"+File.separator+LIBRARY_CONFIG_FILENAME;
			File f = new File(defaultLibraryFilename);
			if (!f.exists()) {
				// Keep this for sake of compatibility
				defaultLibraryFilename=getLibraryPath()+File.separator+"lib"+File.separator+"pandore.cfg";
				File f1 = new File(defaultLibraryFilename);
				if (!f1.exists()) {
					defaultLibraryFilename=null;
				}
			}
		}

		// Build the tree of operators
		if (doSplash) splash.showProgress("Reading common operators...",30);
		Tree itemTree = new Tree(null);
		LinkedList<XMLElement> libraryHandler = null;
		if (operatorLibraryFilename != null) {
			libraryHandler=readConfigFile(new File(operatorLibraryFilename));
		} else if (defaultLibraryFilename != null && PreferenceChooser.getBoolean("UseOperatorLibrary")) {
			libraryHandler=readConfigFile(new File(defaultLibraryFilename));
		}
		if (libraryHandler != null) {
			itemTree.setCommonOperators(libraryHandler);
		}
		libraryHandler = null;
		// Add personal operators.

		if (doSplash) splash.showProgress("Reading personal operators ...",40);
		// new version
		LinkedList<XMLElement> personalOperatorHandler = null;
		File personalOperatorFile=new File(Ariane.getPersonalOperatorFilename());
		if (personalOperatorFile.exists()) {
			personalOperatorHandler=readConfigFile(personalOperatorFile);
			itemTree.addUserOperators(personalOperatorHandler);
		} else {
			// Keep this for sake of compatibility
			if (preferenceHandler != null) {
				itemTree.addUserOperators(preferenceHandler);
				File d = personalOperatorFile.getParentFile();
				d.mkdirs();
				configFile.renameTo(personalOperatorFile);
			}
		}

		// Add the subroutines stored in the init file.
		if (doSplash) splash.showProgress("Reading subroutines ...",50);
		File subRoutineFile=new File(Ariane.getSubroutineFilename());
		if (subRoutineFile.exists()) {
			LinkedList<XMLElement> subRoutineHandler=readConfigFile(subRoutineFile);
			itemTree.addSubroutines(subRoutineHandler);
		}

		if (doSplash) splash.showProgress("Building graphical interfaces ...",60);
		installUI();
		workspace = new Workspace(itemTree);

		if (doSplash) splash.showProgress("Initializing windows contents ...",70);
		// Case of Applet or not.

		// The icon :: should be a gif image.
		ImageIcon logo64=new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"logo64.png"));
		ImageIcon logo32=new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"logo32.png"));
		frame.setIconImages(Arrays.asList(new Image[] {logo64.getImage(), logo32.getImage()}));
		if (_applet == null) {
			frame.setJMenuBar(workspace.buildMenuBar()); 
			workspace.initFrame(frame);
			workspace.initContent();
			frame.setContentPane(workspace);
			frame.pack();
		} else {
			_applet.setJMenuBar(workspace.buildMenuBar());
			workspace.initContent();
			// The icon.
			_applet.setContentPane(workspace);
		}

		Graph[] g=null;
		if (graphFilenames.size()>0) {
			if (doSplash) splash.showProgress("Reading the input graph(s) ...",90);
			g = new Graph[graphFilenames.size()];
			for ( int i=0; i< graphFilenames.size(); i++) {
				g[i]=workspace.loadGraph(new File(graphFilenames.get(i)));
			}
			File file = new File(graphFilenames.get(0));
			Workspace.setCurrentDir(file.getAbsoluteFile().getParent());
		}

		if (doSplash) splash.showProgress("Starting session ...",100);

		// Close the SplashScreen.
		if (splash!=null){
			splash.setVisible(false);
			splash.dispose();
			splash = null;
		}

		workspace.display(g);
		if (_applet==null) {
			frame.setVisible(true);
		}

		// First use of the software: configure directories.
		if (getLibraryPath().equalsIgnoreCase(".") && PreferenceChooser.getBoolean("UseOperatorLibrary")) {
			workspace.configurationAction();
		}
	}

	/**
	 * Installs the L&F preferences.
	 */
	private void installUI () {
		try {
			NimRODLookAndFeel NimRODLF = new NimRODLookAndFeel();
			UIManager.setLookAndFeel(NimRODLF);
//			UIManager.setLookAndFeel("com.sun.java.swing.plaf.Metal");
//			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch ( Exception e) { }
		Color x = UIManager.getColor("Panel.background");
		UIManager.put("TabbedPane.contentAreaColor",x);
		UIManager.put("TabbedPane.background",new Color(150,150,190));
		UIManager.put("TabbedPane.selected",new Color(90,90,120));
		UIManager.put("TabbedPane.selectedForeground",Color.white);
		UIManager.put("OptionPane.yesIcon", new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"gtk-yes.png")));
		UIManager.put("OptionPane.noIcon", new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"gtk-no.png")));
		UIManager.put("OptionPane.cancelIcon", new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"gtk-cancel.png")));
		UIManager.put("OptionPane.okIcon", new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"gtk-apply.png")));
	}

	private static void changeDefaultCharsetSet( String  charsetEncoder ) throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		Class<Charset> c = Charset.class;
		Field defaultCharsetField = c.getDeclaredField("defaultCharset");
		defaultCharsetField.setAccessible(true);
		defaultCharsetField.set(null, Charset.forName(charsetEncoder));
	}

	/**
	 * Creates the application (i.e., the workspace)
	 * and initializes all the components.
	 * Displays a <code>SplashScreen</code> during initialization process.
	 * 
	 * @param args the list of input arguments.
	 */
	static public void main( String[] args ) {
		try {
			changeDefaultCharsetSet("utf-8");
		} catch ( Exception e) { }
		new Ariane(args);
	}

	/**
	 * The List of all error codes used in the project.
	 */
	public enum ErrorCode {
		/** Bottom of the stack error. */
		NOERROR,
		/** Missing parameter value. */
		MISSINGPARAMETER,
		/** Missing input data file. */
		MISSINGINPUT,
		/** Command runtime error during execution. */
		RUNTIME,
		/** Input or Output errors during execution. */
		INPUTOUTPUT,
		/** Command execution is interrupted. */
		INTERRUPTEDEXECUTION,
		/** Missing file. */
		MISSINGFILE,
		/** Bad file name. */
		UNKNOWNFILE,
		/** Arithmetics error. */
		MISSINGFORMULA,
		/** Arithmetics error. */
		ARITHMETICS,
	}
}
