/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.generator;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureForItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureIfItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureMergeItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureSelectItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureWhileItem;
import fr.greyc.pantheon.ariane.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.graph.slot.InSlot;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotMask;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;

import java.io.File;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedList;
import java.util.StringTokenizer;

/**
 * The <code>ShellGenerator</code> generates bash Shell code
 * for the specified graph.
 * It operates in three steps:<ol>
 * <li>transforms the graph into a list of operators usable for the topological sort;
 * <li>executes the topological sort;
 * <li>generates the related code.
 * </ol>
 *
 * @version Régis CLOUARD - dec 04, 2007
 * @version Régis CLOUARD - Jul 03, 2010 (fixed bugs)
 * @version Régis CLOUARD - Nov 19, 2011 (fixed bugs)
 * @version Régis CLOUARD - Aug 25, 2012 (add system + script items)
 */ 
public final class BashGenerator extends CodeGenerator {
	/** The list of generated output images. */
	private final ArrayList<String> _outputImages = new ArrayList<String>();

	/**
	 * Creates the Bash generator.
	 * 
	 * @param chosenFile the name of the file where to store the bash script.
	 * @param graph the graph to be saved.
	 */
	public BashGenerator( File chosenFile, Graph graph) {
		super(chosenFile,graph);
	}

	/**
	 * Generates the header code of the Shell file and
	 * the code that checks the argument line consistency.
	 * 
	 * @param fp the output stream.
	 * @param filename the name of the code file.
	 * @param arguments the list of arguments defined in the graph.
	 * @param inputs the list of loaders defined in the graph.
	 * @param outputs the list of savers defined in the graph.
	 */
	@Override
	protected void generateScriptHeader( FileOutputCode fp, String filename, ArrayList<GraphItem> arguments, ArrayList<GraphItem> inputs, ArrayList<GraphItem> outputs) {
		// The current date.
		Date creationDate=new Date();
		DateFormat dateFormat=DateFormat.getDateInstance(DateFormat.MEDIUM);
		fp.println("#!/bin/bash");
		fp.println("#");
		fp.println("# @(#)"+filename+"				"+dateFormat.format(creationDate));
		fp.println("#");
		fp.println("# This file was automatically generated by Ariane.");
		fp.println("# Ariane is part of the Pantheon Project.");
		fp.println("#\n# GREYC - ENSICAEN & University of Caen, France");
		fp.println("");
		
		fp.println("IFS=$(echo -en \"\\n\\b\")");
		fp.println("");

		fp.println("# Check argument line consistency.");
		fp.println("if [ $# -ge 1 -a \"$1\" = \"-h\" -o $# -ne "+(arguments.size()+inputs.size()+outputs.size())+" ]");
		fp.println("then");

		fp.print("echo \"Usage: $0 ",1);
		for (int i=0;i<arguments.size();i++) {
			fp.print(" "+arguments.get(i).getNickName());
		}
		for (int i=0;i<inputs.size();) {
			fp.print(" input-file"+(++i));
		}
		for (int i=0;i<outputs.size();) {
			fp.print(" output-file"+(++i));
		}
		fp.println("\"");
		fp.print("echo \"  example: $0",1);
		for (int i=0;i<arguments.size();i++) {
			if (arguments.get(i).getFormula()!=null && arguments.get(i).getFormula().length()>0) {
				fp.print(" "+arguments.get(i).getFormula());
			} else {
				fp.print(" "+arguments.get(i).getNickName());
			}
		}
		for (int i=0;i<inputs.size();i++) {
			fp.print(" "+inputs.get(i).getNickName());
		}
		for (int i=0;i<outputs.size();i++) {
			String name = outputs.get(i).getNickName();
			if (name.equals("saver")) name+=outputs.get(i).getId();
			fp.print(" "+name);
		}
		fp.println("\"");

		fp.println("exit",1);
		fp.println("fi");
	}

	/**
	 * Generates the footer code of the script file.
	 * 
	 * @param fp the output stream.
	 */
	@Override
	protected void generateScriptFooter( FileOutputCode fp ) {
		fp.println("\nmain $*");
	}

	/**
	 * Generates utility functions required to execute the script.
	 * 
	 * @param fp the output stream.
	 */
	@Override
	protected void generateUtilitySubroutinesDefinition( FileOutputCode fp ) { }

	/**
	 * Creates the name of the image from the specified slot number,
	 * item and graph identifier.
	 * 
	 * @param slotIndex the identifier of the slot.
	 * @param itemIndex the identifier of the item.
	 * @param num_graph the graph identifier
	 * @return a name that is supposed unique.
	 */
	private String createImageName( int slotIndex, int itemIndex, int num_graph ) {
		String name= PreferenceChooser.createTemporaryName(num_graph, itemIndex, slotIndex);
		_outputImages.add(name);
		return name;
	}

	/**
	 * Returns the command line for getting the
	 * input images for the specified item slot of the specified
	 * graph identifier.
	 * 
	 * @param slot the related slot.
	 * @param num_graph the graph identifier
	 * @return a name that is supposed to unique.
	 */
	private String getInputImage( InSlot slot, int num_graph ){
		return PreferenceChooser.createTemporaryName(num_graph, slot.getConnectedSlot().getGraphItem().getId(), slot.getConnectedSlot().getRank());
	}

	/**
	 * Creates a variable name with the specified arguments.
	 * 
	 * @param graph the corresponding graph.
	 * @param item the corresponding item.
	 * @param slot the corresponding slot.
	 * @return the name of the variable.
	 */
	private String variable( int graph, int item, int slot ) {
		return "var"+graph+"_"+item+"_"+slot;
	}

	/**
	 * Returns the parameters line command of the specified item.
	 * 
	 * @param item the related item.
	 * @param slotIndex the slot index.
	 * @param graphId the current graph identifier.
	 * @return a value or a command to get the value.
	 */
	private String getParameter( GraphItem item, int slotIndex, int graphId ) {
		InSlotParameter isp = (InSlotParameter) item.getSlot("InSlotParameter", slotIndex);
		if (isp.getConnectedSlot() == null)
			return "'"+((InSlotParameter)item.getSlot("InSlotParameter", slotIndex)).getValue()+"'";
		else
			return "$"+variable(graphId,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank());
	}

	/**
	 * Returns the parameters line command of the specified item.
	 * 
	 * @param item the related item.
	 * @param slotIndex the slot index.
	 * @param graphId the current graph identifier.
	 * @return a value or a command to get the value.
	 */
	private String getNumericalParameter( GraphItem item, int slotIndex, int graphId ) {
		InSlotParameter isp = (InSlotParameter) item.getSlot("InSlotParameter", slotIndex);
		if (isp.getConnectedSlot() == null)
			return ((InSlotParameter)item.getSlot("InSlotParameter", slotIndex)).getValue();
		else
			return "$"+variable(graphId,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank());
	}

	/**
	 * Translates the specified logical formula into Shell.
	 * 
	 * @param formula the logical formula.
	 * @param parameter0 the value of the parameter0.
	 * @param parameter1 the value of the parameter1.
	 * @return the Shell code for the formula.
	 */
	private String translateLogicalFormula( String formula, String parameter0, String parameter1 ) {
		// Parse the boolean expression into Shell.
		char[] st = formula.toCharArray();
		String commandShell = "";
		int index=0;
		while (index<st.length) {
			if (st[index]=='&') {
				commandShell += " && ";
				index++;
			} else if (st[index]=='|') {
				commandShell += " || ";
				index++;
			} else if (st[index]=='<') {
				if (st[index+1]=='=') {
					commandShell += " <= ";
					index+=2;
				} else {
					commandShell += " < ";
					index++;
				}
			} else if (st[index]=='>') {
				if (st[index+1]=='=') {
					commandShell += " >= ";
					index+=2;
				} else {
					commandShell += " > ";
					index++;
				}
			} else if (st[index]=='=') {
				if (st[index+1]=='=') {
					commandShell += " == ";
					index+=2;
				} else {
					index++; // degenerated case!
				}
			} else if (st[index]=='!') {
				if (st[index+1]=='=') {
					commandShell += " != ";
					index+=2;
				} else {
					commandShell += " ! ";
					index++;
				}
			} else if (st[index]=='(') {
				commandShell += " ";
				index++;
			} else if (st[index]==')') {
				commandShell += " ";
				index++;
			} else if (st[index]=='$') {
				if (st[index+1]=='0') {
					commandShell += parameter0;
				} else {
					commandShell += parameter1;
				}
				index+=2;
			} else {
				commandShell += st[index++];
			}
		}
		return commandShell;
	}

	/**
	 * Generates code for the main.
	 * 
	 * @param graph the graph to encode.
	 * @param fp the output stream.
	 * @param arguments the list of argument.
	 * @param inputs the list of inputs.
	 */
	@Override
	protected void generateMainPrologue( Graph graph, FileOutputCode fp, ArrayList<GraphItem> arguments, ArrayList<GraphItem> inputs, ArrayList<GraphItem> outputs ) {
		fp.println("\n#\n# The main program\n#");
		fp.println("function main() {");
		fp.println("rm -f "+PreferenceChooser.TEMPORARY_FILE_PREFIX+"?*"+PreferenceChooser.TEMPORARY_FILE_EXT,1);
		for ( GraphItem item : arguments ) {
			fp.println(variable(1,item.getId(),0)+"=$1; shift",1);
		}
		for ( GraphItem item : inputs ) {
			fp.println("ln -sf $1 "+createImageName(0,item.getId(),1)+"; shift",1);
		}
	}

	/**
	 * Generates code to erase all temporary files.
	 * 
	 * @param graph the graph to encode.
	 * @param fp the output stream.
	 */
	@Override
	protected void generateMainEpilogue( Graph graph, FileOutputCode fp ) {
		//		fp.println("rm -f "+PreferenceChooser.TEMPORARY_FILE_PREFIX+"*"+PreferenceChooser.TEMPORARY_FILE_EXT, 1);
		fp.println("}");
	}

	/**
	 * Generates the code for loading an image.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( LoaderItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception { }

	/**
	 * Generates the code for saving an image.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( SaverItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		if (((item.getInSlotImages().get(0)).getConnectedSlot()) != null) {
			fp.println("cp "+getInputImage(item.getInSlotImages().get(0),num_graph)+" $1; shift",indent);
			fp.println("ln -sf "
					+ getInputImage(item.getInSlotImages().get(0),num_graph)
					+ " "
					+ createImageName(0,item.getId(),num_graph)
					+ ";",indent);
		}
	}

	/**
	 * Generates the code for viewer item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ViewerItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		String commandShell = item.getViewerCommandName();
		if ((item.getInSlotImages().get(0).getConnectedSlot()) != null) {
			fp.println(commandShell+" "+getInputImage(item.getInSlotImages().get(0),num_graph),indent);
			fp.println("ln -sf "
					+ getInputImage(item.getInSlotImages().get(0),num_graph)
					+ " "
					+ createImageName(0,item.getId(),num_graph)
					+ ";",indent);
		}
	}

	/**
	 * Generates the code for executing an operator.
	 * 
	 * @param item the operator to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( OperatorItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		LinkedList<String> maskLine = new LinkedList<String>();
		LinkedList<String> parameterLine = new LinkedList<String>();
		LinkedList<String> inputLine = new LinkedList<String>();
		LinkedList<String> outputLine = new LinkedList<String>();

		String codeName = OperatorItem.getCommandPart(item.getBinary());
		if (codeName==null) {
			throw new Exception(Ariane.lang.getString("Generator.error2")+item.getName());
		}

		// Get parameters
		for (int i = 0; i < item.getInSlotParameters().size(); i++) {
			parameterLine.add(getParameter(item,i,num_graph));
		}

		// Get mask
		InSlotMask mask = item.getInSlotMask();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask,num_graph))) {
			maskLine.add(getInputImage(mask,num_graph));
		}

		// Get input images.
		for (InSlotImage im : item.getInSlotImages() ) {
			if (im.getConnectedSlot() == null)
				throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
			inputLine.add(getInputImage(im,num_graph));
		}

		// Get output images.
		LinkedList<OutSlotImage> im_out = item.getOutSlotImages();
		for (int i = 0; i < im_out.size(); i++) {
			outputLine.add(createImageName(i,item.getId(),num_graph));
		}

		String code="";
		LinkedList<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
		for ( String command : commandLine) {
			code +=command+" ";
		}

		// Get output result.
		fp.print(code,indent);
		fp.println("; "+variable(num_graph,item.getId(),0)+"=`"+GraphItem.getCommandStatus()+"`");
	}

	/**
	 * Generates the code for executing an operating system command.
	 * 
	 * @param item the operator to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( SystemItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		LinkedList<String> maskLine = new LinkedList<String>();
		LinkedList<String> parameterLine = new LinkedList<String>();
		LinkedList<String> inputLine = new LinkedList<String>();
		LinkedList<String> outputLine = new LinkedList<String>();

		String codeName = OperatorItem.getCommandPart(item.getBinary());
		if (codeName==null) {
			throw new Exception(Ariane.lang.getString("Generator.error2")+item.getName());
		}

		// Get parameters
		for (int i = 0; i < item.getInSlotParameters().size(); i++) {
			parameterLine.add(getParameter(item,i,num_graph));
		}

		// Get mask
		InSlotMask mask = item.getInSlotMask();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask,num_graph))) {
			maskLine.add(getInputImage(mask,num_graph));
		}

		// Get input images.
		for (InSlotImage im : item.getInSlotImages() ) {
			if (im.getConnectedSlot() == null)
				throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
			inputLine.add(getInputImage(im,num_graph));
		}

		// Get output images.
		LinkedList<OutSlotImage> im_out = item.getOutSlotImages();
		for (int i = 0; i < im_out.size(); i++) {
			outputLine.add(createImageName(i,item.getId(),num_graph));
		}

		String code="";
		LinkedList<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
		for ( String command : commandLine) {
			code +=command+" ";
		}

		// Get output result.
		fp.print(code,indent);
		fp.println("; "+variable(num_graph,item.getId(),0)+"=`"+GraphItem.getCommandStatus()+"`");
	}

	/**
	 * Parses the give formula into shell code.
	 * 
	 * @param formula the function.
	 * @param parameters the list of parameters.
	 * @return the shell code.
	 */
	private String parseFormula( String formula, String[] parameters ) {
		String code="";
		StringTokenizer st = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
		String etape;
		while (st.hasMoreTokens()) {
			etape = st.nextToken();
			if (etape.equals("e") || etape.equals("E")) code += "*10^";
			else if (etape.equals("cos")) code+="c";
			else if (etape.equals("sin")) code+="s";
			else if (etape.equals("ln")) code+="l";
			else if (etape.equals("exp")) code+="e";
			else if (etape.equals("atan")) code+="a";
			else if (etape.equals("$")) {
				int n=Integer.parseInt(st.nextToken());
				code+= parameters[n];
			} else
				code += etape;
		}
		return code;
	}

	/**
	 * Parses the give formula into shell code.
	 * 
	 * @param formula the function.
	 * @param parameters the list of parameters.
	 * @return the shell code.
	 */
	private String setFunction( String formula, String[] parameters ) {
		String code="";
		StringTokenizer st = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
		String etape;
		boolean abs=false;
		boolean max=false;
		boolean min=false;
		boolean tan=false;
		boolean log2=false;
		boolean log10=false;
		boolean acos=false;
		boolean asin=false;

		while (st.hasMoreTokens()) {
			etape = st.nextToken();
			if (etape.equals("abs") && !abs) {
				code+=" define abs(x){; if (x>=0) return(x) else return (-x)} ";
				abs=true;
			} else if (etape.equals("max") && !max) {
				code+=" define max(x,y){; if (x>y) return(x) else return(y)} ";
				max=true;
			} else if (etape.equals("min") && !min) {
				code+=" define min(x,y){; if (x<y) return(x) else return(y)} ";
				min=true;
			} else if (etape.equals("tan") && !tan) {
				code+=" define tan(x){; return (s(x)/c(x)} ";
				tan=true;
			} else if (etape.equals("acos") && !acos) {
				code+=" define acos(x){; if (x==1) return(0) if (x==-1) return (4*a(1)) return (2*a(1)-a(x/sqrt(1-(x^2))))} ";
				acos=true;
			} else if (etape.equals("asin") && !asin) {
				code+=" define asin(x){; if (x==1) return(2*a(1)) if (x==-1) return (-2*a(1)) return (a(x/sqrt(1-(x^2))))} ";
				asin=true;
			} else if (etape.equals("log")) {
				etape = st.nextToken();
				if (etape.equals("2") && !log2) {
					code+=" define log2(x){; return l(x)/l(2)} ";
					log2=true;
				} else if (etape.equals("1") && !log10) {
					code+=" define log10(x){; return l(x)/l(10)} ";
					log10=true;
				}
			}
		}
		return code;
	}

	/**
	 * Generates the code for an Eval item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( EvalItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		String parameters[] = new String[item.getInSlotParameters().size()];
		// Get parameters
		for(int i = 0; i < item.getInSlotParameters().size(); i++) {
			parameters[i]=getParameter(item,i,num_graph);
		}

		fp.println(variable(num_graph,item.getId(),0)+"=" +"$(echo \""
				+ setFunction(item.getFormula(),parameters)
				+ parseFormula(item.getFormula(),parameters)
				+ "\"|bc -l);",indent);
	}

	@Override
	protected void generateItem( FormatItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		String parameters[] = new String[item.getInSlotParameters().size()];
		// Get parameters
		for(int i = 0; i < item.getInSlotParameters().size(); i++) {
			parameters[i]=getParameter(item,i,num_graph);
		}

		String code="";
		StringTokenizer st = new StringTokenizer(item.getFormula(), " $()+'\"", true);
		String etape;
		while (st.hasMoreTokens()) {
			etape = st.nextToken();
			if (etape.equals("+")){
				code+="";
			} else if (etape.equals("quote")) {
			} else if (etape.equals("\"")) {
			} else if (etape.equals("(")) {
				code+="\'";
			} else if (etape.equals(")")) {
				code+="\'";
			} else if (etape.equals("$")) {
				int n=Integer.parseInt(st.nextToken());
				code+= parameters[n];
			} else {
				code += etape;
			}
		}
		fp.println(variable(num_graph,item.getId(),0)+"=\""+code+"\"", indent);
	}

	/**
	 * Generates the code for the control structure Select.
	 * 
	 * @param item the select to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureSelectItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		String value=getParameter(item,0,num_graph);
		fp.println("# Select", indent);
		if (value == null) {
			value = variable(num_graph,item.getId(),0);
			fp.println("read " + value, indent);
			value="$"+value;
		}
		for(int i = 0; i < item.getInSlotImages().size(); i++) {
			fp.println("if [ " + value +" -eq " + i +" ]", indent);
			fp.println("then", indent);
			fp.println("ln -sf "
					+ getInputImage(item.getInSlotImages().get(i),num_graph) +" "
					+ createImageName(0,item.getId(),num_graph), indent+1);
			fp.println(variable(num_graph,item.getId(),0)+"=" + i +";", indent+1);
			fp.println("fi", indent);
		}
	}

	/**
	 * Generates the code for the control structure If.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureIfItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		String parameters[] = new String[2];
		for (int i = 0; i < item.getInSlotParameters().size(); i++) {
			parameters[i]=getNumericalParameter(item, i, num_graph);
		}
		String commandShell = translateLogicalFormula(item.getFormula(), parameters[0], parameters[1]);
		fp.println("if [ $( echo \"" + commandShell + "\" |bc -l) -gt 0 ]", indent);
	}

	/**
	 * Generates the code for the 'then' part.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateThenPart( ControlStructureIfItem item, int num_graph, FileOutputCode fp, int indent) throws Exception {
		fp.println("then",indent);
		fp.print("ln -sf "+getInputImage(item.getInSlotImages().get(0),num_graph)+" "
				+ createImageName(0,item.getId(),num_graph),indent+1);
		fp.println("; "+variable(num_graph,item.getId(),0)+"=0");
	}

	/**
	 * Generates the code for the 'else' part.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateElsePart( ControlStructureIfItem item, int num_graph, FileOutputCode fp, int indent) throws Exception {
		fp.println("else", indent);
		fp.print("ln -sf "+getInputImage(item.getInSlotImages().get(0),num_graph)+" "
				+ createImageName(1,item.getId(),num_graph),indent+1);
		fp.println("; "+variable(num_graph,item.getId(),0)+"=1");
	}

	/**
	 * Generates the code for the end of a if control structure.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateFiPart( ControlStructureIfItem item, int num_graph, FileOutputCode fp, int indent) throws Exception {
		fp.println("fi",indent);
	}

	/**
	 * Generates the code for the control structure merge.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureMergeItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception{
		// input image
		InSlotImage in0;
		in0 = item.getInSlotImages().get(0);
		int id0;
		id0 = in0.getConnectedSlot().getGraphItem().getId();
		fp.print("if [ -n \"$"+variable(num_graph,id0,0)+"\" -a ",indent);
		fp.println("-r "+getInputImage(item.getInSlotImages().get(0),num_graph)+" ]");
		fp.println("then",indent);
		fp.print("cp "
				+ getInputImage(item.getInSlotImages().get(0),num_graph)
				+ " "
				+ createImageName(0,item.getId(),num_graph),indent+1);
		fp.println("; "+variable(num_graph,item.getId(),0)+"=0;");
		fp.println("rm "+getInputImage(item.getInSlotImages().get(0),num_graph), indent+1);
		fp.println("else", indent);
		fp.print("cp "
				+ getInputImage(item.getInSlotImages().get(1),num_graph) + " "
				+ createImageName(0,item.getId(),num_graph),indent+1);
		fp.println("; "+variable(num_graph,item.getId(),0)+"=1;");
		fp.println("rm "+getInputImage(item.getInSlotImages().get(1), num_graph), indent+1);
		fp.println("fi",indent);
	}

	/**
	 * Generates the code for a note. Nothing to do.
	 * 
	 * @param item the note to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 */
	@Override
	protected void generateItem( NoteItem item, int num_graph, FileOutputCode fp, int indent ) { }

	/**
	 * Generates the code for a <code>Filename</code> item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( FileNameItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		if (!item.isArgument()) {
			fp.println(variable(num_graph,item.getId(),0)+"='"+item.getFormula()+"'", indent);
		}
	}

	/**
	 * Generates the code for a <code>Constant</code>item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ConstantItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		if (!item.isArgument() ) {
			String value = item.getFormula().replaceAll("\"","\\\\\"");
			fp.println(variable(num_graph,item.getId(),0)+"=\""+value+"\"",indent);
		}
	}

	/**
	 * Generates the code for a <code>Script</code>item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ScriptItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		String value = item.getFormula().replaceAll("\"","\\\\\"").replaceAll("\n", "\\\\\n");
		fp.println(variable(num_graph,item.getId(),0)+"=\""+value+"\"",indent);
	}

	/**
	 * Generates the code for the specified MacroOperator,
	 * for, or while call.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception in case of incomplete graph.
	 */
	@Override
	protected void generateItem( CompositeItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		String code = "macro"+item.getId()+"_"+item.getGraph().getIdentifier();

		// Get parameters
		for (int i = 0; i < item.getInSlotParameters().size(); i++) {
			code +=" "+getParameter(item,i,num_graph);
		}

		// Get mask
		InSlotMask mask = item.getInSlotMask();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask,num_graph))) {
			code += " " + getInputImage(mask,num_graph);
		}

		// Get input images.
		for( InSlotImage im : item.getInSlotImages() ) {
			if (im.getConnectedSlot() == null)
				throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
			code += " " + getInputImage(im,num_graph);
		}

		// Get output images.
		LinkedList<OutSlotImage> im_out = item.getOutSlotImages();
		for(int i = 0; i < im_out.size(); i++) {
			code += " " +createImageName(i,item.getId(),num_graph);
		}
		// Get output result.
		fp.println(code+"; "+variable(num_graph,item.getId(),0)+"=`"+GraphItem.getCommandStatus()+"`",indent);
	}

	/**
	 * Parses the comment in order to add a sharp after each newlines.
	 * 
	 * @param comment a string that contains the comment.
	 * @return a new string.
	 */
	private String parseComment( String comment ) {
		return comment.replaceAll("\n", "\n# ");
	}

	/**
	 * Generates the code for the specified control structure For.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateSubGraph( ControlStructureForItem item, FileOutputCode fp ) throws Exception {
		SubGraph sb = item.getSubGraph();
		int num_subgraph = sb.getIdentifier();
		ArrayList<GraphItem> litems = buildItemsSequence(sb);

		// Get and discard the virtual operator
		Object virtualOperator=getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.size() == 0) {
			throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
		}

		fp.println("\n#\n# "+item.getNickName());
		for (GraphItem note : sb.getGraphItems()) {
			if (note instanceof NoteItem) {
				fp.println("# "+parseComment(note.getFormula()));
			}
		}
		fp.println("#");

		fp.println("function macro"+item.getId()+"_"+item.getGraph().getIdentifier()+"() {");

		// Get parameters
		fp.println("# Parameters",1);
		fp.println("begin=$1; shift",1);
		fp.println("end=$1; shift",1);
		fp.println("step=$1; shift",1);
		for (int i = 3; i < item.getInSlotParameters().size(); i++) {
			fp.println(variable(num_subgraph,0,i-2)+"=$1; shift",1);
		}

		// Get mask
		fp.println("# Input data files",1);
		InSlotMask mask = item.getInSlotMask();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask,item.getGraph().getIdentifier()))) {
			int rz = sb.getVirtualOperator().getVirtualInSlotMask().getRank();
			fp.println("ln -fs $1 " + createImageName(rz,0,num_subgraph)+"; shift",1);
		}

		// Get input images
		LinkedList<InSlotImage> l= item.getInSlotImages();
		for (int i=0; i<l.size(); i++) {
			InSlotImage isi = l.get(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
			}
			fp.println("cp $1 "+createImageName(i,0,num_subgraph)+"; shift",1);
		}
		fp.println("");
		// Create the output images in case of no iteration
		VirtualItem vo = (VirtualItem)virtualOperator;
		for (int i=0; i<item.getOutSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualOutSlotImages().get(i);
			fp.println("touch "+getInputImage(isi,num_subgraph),1);
		}

		// Initialize the return variable in case of no iteration
		{
			InSlotParameter isp = (InSlotParameter)vo.getVirtualOutSlotResult();
			fp.println(variable(num_subgraph,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank())+"=SUCCESS",1);
		}

		fp.println("");
		fp.println("local i",1);
		fp.println("for ((i=$begin; i<=$end; i+=$step ))",1);
		fp.println("do",1);
		fp.println(variable(num_subgraph,0,0)+"=$i;",2);
		generateSubroutineBody(litems, num_subgraph, fp, 2);

		// Set intermediate images
		for (int i=0; i<item.getInSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualInterSlotImage(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
			}
			String input = getInputImage(isi,num_subgraph);
			String output = createImageName(i,0,num_subgraph);
			if (!input.equals(output)) {
				fp.println("cp "+input+" "+output,2);
			}
		}

		fp.println("done",1);

		fp.println("");
		// Get output images
		for (int i=0; i<item.getOutSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualOutSlotImages().get(i);
			fp.println("ln -fs "+getInputImage(isi,num_subgraph)+" $1; shift",1);
		}
		// Output result
		InSlotParameter isp = (InSlotParameter)vo.getVirtualOutSlotResult();
		if (isp.getConnectedSlot() == null) {
			fp.println("psetstatus SUCCESS",1);
		} else {
			fp.println("psetstatus $"+variable(num_subgraph,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank()),1);
		}
		fp.println("}");

		generateSubroutinesDefinition(litems,fp);
	}

	/**
	 * Generates the code for the specified control structure While.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateSubGraph( ControlStructureWhileItem item, FileOutputCode fp ) throws Exception {
		SubGraph sb = item.getSubGraph();
		int num_subgraph = sb.getIdentifier();
		ArrayList<GraphItem> litems = buildItemsSequence(sb);

		// Get and discard the virtual operator
		Object virtualOperator=getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.size() == 0) {
			throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
		}

		fp.println("\n#\n# "+item.getNickName());
		for (GraphItem note : sb.getGraphItems()) {
			if (note instanceof NoteItem) {
				fp.println("# "+parseComment(note.getFormula()));
			}
		}
		fp.println("#");

		fp.println("function macro"+item.getId()+"_"+item.getGraph().getIdentifier()+"() {");

		// Get parameters
		fp.println("# Parameters",1);
		String parameters[] = new String[item.getInSlotParameters().size()];
		for (int i = 0; i < item.getInSlotParameters().size(); i++) {
			fp.println(variable(num_subgraph,0,i)+"=$1; shift",1);
			parameters[i]="$"+variable(num_subgraph,0,i);
		}

		// Get mask.
		fp.println("# Input data files",1);
		InSlotMask mask = item.getInSlotMask();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask,item.getGraph().getIdentifier()))) {
			int rz = sb.getVirtualOperator().getVirtualInSlotMask().getRank();
			fp.println("ln -fs $1 " + createImageName(rz,0,num_subgraph)+"; shift",1);
		}

		// Get input images.
		LinkedList<InSlotImage> l= item.getInSlotImages();
		for (int i=0; i<l.size(); i++) {
			InSlotImage isi = l.get(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
			}
			fp.println("cp $1 "+createImageName(i,0,num_subgraph)+"; shift",1);
		}

		// Parse the boolean expression.
		String testFormula  = translateLogicalFormula(item.getFormula(),parameters[0],parameters[1]);
		VirtualItem vo = (VirtualItem)virtualOperator;

		fp.println("");
		// Create output images in case of no repetition
		for (int i=0; i<item.getOutSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualOutSlotImages().get(i);
			fp.println("touch "+getInputImage(isi,num_subgraph),1);
		}

		// Initialize the return variable in case of no iteration
		{
			InSlotParameter isp = (InSlotParameter)vo.getVirtualOutSlotResult();
			fp.println(variable(num_subgraph,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank())+"=SUCCESS",1);
		}

		fp.println("");
		fp.println("while [ $( echo \"" + testFormula + "\" |bc -l) -gt 0 ]",1);
		fp.println("do",1);
		generateSubroutineBody(litems,num_subgraph,fp,2);

		// Get intermediate images
		for (int i=0; i<item.getInSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualInterSlotImage(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null)
				throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
			// Discard cp image image (it happens with the output = input)
			String input = getInputImage(isi,num_subgraph);
			String output = createImageName(i,0,num_subgraph);
			if (!input.equals(output)) {
				fp.println("cp "+input+" "+output,2);
			}
		}

		// Get intermediate parameters
		for (int i=0;i<2;i++) {
			InSlotParameter isp = vo.getVirtualInterSlotResult(i);
			if (isp.getConnectedSlot() != null) {
				fp.println(variable(num_subgraph,0,i)+"=$"+variable(num_subgraph,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank()),2);
			}
		}
		fp.println("done",1);
		// Get output images
		for (int i=0; i<item.getOutSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualOutSlotImages().get(i);
			fp.println("ln -sf "+getInputImage(isi,num_subgraph)+" $1; shift",1);
		}

		// Output result.
		InSlotParameter isp = (InSlotParameter)vo.getVirtualOutSlotResult();
		if (isp.getConnectedSlot() == null) {
			fp.println("psetstatus SUCCESS",1);
		} else {
			fp.println("psetstatus $"+variable(num_subgraph,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank()),1);
		}
		fp.println("}");

		generateSubroutinesDefinition(litems,fp);
	}

	/**
	 * Generates the code for the specified MacroOperator body.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception In case of incomplete graph.
	 */
	@Override
	protected void generateSubGraph( MacroOperatorItem item, FileOutputCode fp ) throws Exception {
		SubGraph sb = item.getSubGraph();
		int num_subgraph=sb.getIdentifier();
		ArrayList<GraphItem> litems = buildItemsSequence(sb);

		// Get and discard the virtual operator.
		Object virtualOperator=getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.size() == 0) {
			throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
		}

		fp.println("\n#\n# "+item.getNickName());
		for (GraphItem note : sb.getGraphItems()) {
			if (note instanceof NoteItem) {
				fp.println("# "+parseComment(note.getFormula()));
			}
		}
		fp.println("#");

		fp.println("function macro"+item.getId()+"_"+item.getGraph().getIdentifier()+"() {");

		// Get parameters.
		fp.println("# arguments",1);
		for(int i = 0; i < item.getInSlotParameters().size(); i++) {
			fp.println(variable(num_subgraph,0,i)+"=$1; shift",1);
		}

		// Get mask.
		fp.println("# Input data files",1);
		InSlotMask mask = item.getInSlotMask();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask,item.getGraph().getIdentifier()))) {
			int rz = sb.getVirtualOperator().getVirtualInSlotMask().getRank();
			fp.println("ln -fs $1 " + createImageName(rz,0,num_subgraph)+"; shift",1);
		}

		// Get input images.
		LinkedList<InSlotImage> l= item.getInSlotImages();
		for(int i=0; i<l.size(); i++) {
			InSlotImage isi = l.get(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
			}
			fp.println("ln -fs $1 "+createImageName(i,0,num_subgraph)+"; shift",1);
		}

		fp.println();

		generateSubroutineBody(litems, num_subgraph, fp, 1);

		VirtualItem vo = (VirtualItem)virtualOperator;
		// Get output images.
		for(int i=0; i<item.getOutSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualOutSlotImages().get(i);
			fp.println("ln -fs "+getInputImage(isi,num_subgraph)+" $1; shift",1);
		}

		fp.println("");
		// Output result.
		InSlotParameter isp = (InSlotParameter)vo.getVirtualOutSlotResult();
		if (isp.getConnectedSlot() == null) {
			fp.println("pstetstatus SUCCESS",1);
		} else {
			fp.println("psetstatus $"+variable(num_subgraph,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank()),1);
		}
		fp.println("}");

		generateSubroutinesDefinition(litems,fp);
	}
}