/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.LayoutManager2;
import java.awt.Rectangle;
import java.util.Hashtable;

/**
 * Defines a class that knows how to lay out a graph.
 * A component is added with a location constraints.
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Jan 07, 2007
 * @version Régis CLOUARD - Jul 20, 2007 (fix bug on slots position)
 */
public final class XYLayout implements LayoutManager2 {
	/** The north layout constraint (top of container). */
	public static final int NORTH=0;
	/** The north west layout constraint (top of container). */
	public static final int NORTHWEST=1;
	/** The north east layout constraint (top of container). */
	public static final int NORTHEAST=2; 
	/** The north east north layout constraint (top of container). */
	public static final int SOUTHWEST=3; 
	/** The south west layout constraint (top of container). */
	public static final int SOUTHEAST=4; 
	/** The south est layout constraint (top of container). */
	public static final int EASTSOUTH=5;
	/** The west north layout constraint. */
	public static final int WESTNORTH=6;

	/** A Hash table for indexing components. */
	private final Hashtable<Component,XYConstraint> _info = new Hashtable<Component,XYConstraint>();

	/**
	 * Creates a new XY layout with no size.
	 */
	public XYLayout( ) { }

	/**
	 * @unused.
	 * @param name the name of the component.
	 * @param component the component.
	 */
	@Override
	public final void addLayoutComponent( String name, Component component ) { }

	/**
	 * Adds the specified component to the layout using the specified
	 * constraints.
	 * If the component was previously included, it is changed from the
	 * specified constraints.
	 * 
	 * @param component the component to be added.
	 * @param constraints where/how the component is added to the layout.
	 */
	@Override
	public final void addLayoutComponent( Component component, Object constraints ) {
		if (constraints == null) constraints = new XYConstraint();
		synchronized (component.getTreeLock()) {
			if (constraints instanceof XYConstraint) {
				XYConstraint c = (XYConstraint)constraints;
				_info.put(component, (XYConstraint)constraints);
				if (c.getRank()<0) {
					component.setBounds(c.getX(), c.getY(), c.getWidth(), c.getHeight());
				}
			}
		}
	}

	/**
	 * Removes the specified component from the layout.
	 * 
	 * @param component the component to be removed.
	 */
	@Override
	public final void removeLayoutComponent( Component component ) {
		synchronized (component.getTreeLock()) {
			_info.remove(component);
		}
	}

	/**
	 * Determines the preferred size dimensions for the
	 * specified container, given the components it contains.
	 * Does not take care of insets of the container:
	 * the layout begins at coordinate 0,0.
	 * 
	 * @param container the container in which to do the layout.
	 * @return the preferred dimension.
	 */
	@Override
	public final Dimension preferredLayoutSize( Container container ) {
		synchronized (container.getTreeLock()) {
			Dimension dim = new Dimension(0,0);
			final Dimension containerDim = container.getSize();
			for (int i = 0; i < container.getComponentCount(); i++) {
				Component component = container.getComponent(i);
				if (component.isVisible() && (_info.get(component).getRank()<0)) {
					Rectangle r = getComponentBounds(component, containerDim);
					dim.width  = Math.max(dim.width , r.x + r.width+50);
					dim.height = Math.max(dim.height, r.y + r.height+50);
				}
			}
			dim.width += 50; // an artificial right insets.
			dim.height += 50; // an artificial bottom insets.
			return dim;
		}
	}

	/**
	 * Calculates the minimum size dimensions for the specified
	 * container, given the components it contains.
	 * 
	 * @param container the component to be laid out.
	* @return the minimum layout size.
	 */
	@Override
	public final Dimension minimumLayoutSize( Container container ) {
		synchronized (container.getTreeLock()) {
			Dimension dim = new Dimension(0,0);
			final Dimension containerDim = container.getSize();
			for (int i = 0; i < container.getComponentCount(); i++) {
				Component component = container.getComponent(i);
				if (component.isVisible() && ((_info.get(component)).getRank()<0)) {
					Rectangle r = getComponentBounds(component, containerDim);
					dim.width  = Math.max(dim.width , r.x + r.width);
					dim.height = Math.max(dim.height, r.y + r.height);
				}
			}
			dim.width += 50; // an artificial right insets.
			dim.height += 50; // an artificial bottom insets.
			return dim;
		}
	}

	/**
	 * Returns the maximum dimensions for this layout
	 * given the components in the specified container.
	 * 
	 * @param container the component which needs to be laid out.
	 * @return the maximum layout size.
	 */
	@Override
	public final Dimension maximumLayoutSize( Container container ) {
		return new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE);
	}

	/**
	 * Organizes the components in the container.
	 * Sets the new size of the XYLayout if needed.
	 * Does not take care of insets of the container:
	 * the layout begins at coordinate 0,0.
	 * 
	 * @param container the container to be laid out.
	 */
	@Override
	public final void layoutContainer( Container container ) {
		synchronized (container.getTreeLock()) {
			final Dimension containerDim=container.getSize();
			for (int i = 0 ; i < container.getComponentCount(); i++) {
				Component component = container.getComponent(i);
				if (component.isVisible()) {
					Rectangle r = getComponentBounds(component, containerDim);
					component.setBounds(r.x, r.y, r.width, r.height);
				}
			}
		}
	}

	/**
	 * Returns the alignment along the x axis.
	 * This specifies how the component would like to be aligned
	 * relative to other components.
	 * The value should be a number between 0 and 1 where 0 represents
	 * alignment along the origin, 1 is aligned the more farest
	 * away from the origin, 0.5 is centered, etc.
	 * 
	 * @param container the container.
	 * @return the X layout alignment.
	 */
	@Override
	public final float getLayoutAlignmentX( Container container ) {
		return 0.0f; // Prefer centered alignment.
	}

	/**
	 * Returns the alignment along the y axis.
	 * This specifies how the component would like to be aligned
	 * relative to other components.
	 * The value should be a number between 0 and 1 where 0 represents
	 * alignment along the origin, 1 is aligned the farest
	 * away from the origin, 0.5 is centered, etc.
	 * 
	 * @param container the container.
	 * @return the Y layout alignment.
	 */
	@Override
	public final float getLayoutAlignmentY( Container container) {
		return 0.0f; // Prefer centered alignment.
	}

	/**
	 * Invalidates the layout, indicating that if the layout manager 
	 * has cached information it should be discarded.
	 * 
	 * @param container the container.
	 */
	@Override
	public final void invalidateLayout( Container container ) { }

	/**
	 * Calculates the bounds of the component given its constraints
	 * the current size of the container that holds the component.
	 * 
	 * @param component the component to be laid out.
	 * @param size the current size of the container.
	 * @return the bounds rectangle.
	 */
	private final Rectangle getComponentBounds( Component component, Dimension size ) {
		final int _iconWidth = 12;
		final int _iconHeight = 12;
		final int OFFSET=60;

		XYConstraint constraints = _info.get(component);

		// Constraints by border location.
		if (constraints.getRank() != -1) {
			int rank = constraints.getRank();
			if (constraints.getAlignement() == WESTNORTH)
				return new Rectangle(0, OFFSET + (rank+1)*20, _iconWidth, _iconHeight);
			if (constraints.getAlignement() == NORTHWEST)
				return new Rectangle(OFFSET + (rank+1)*20, 0, _iconWidth, _iconHeight);

			if (constraints.getAlignement() == SOUTHEAST)
				return new Rectangle(400 + (OFFSET+ (rank+1)*20), size.height-_iconHeight,_iconWidth, _iconHeight);

			if (constraints.getAlignement() == SOUTHWEST)
				return new Rectangle(OFFSET + (rank+1)*20,size.height-_iconHeight, _iconWidth, _iconHeight);
			if (constraints.getAlignement() == NORTHEAST)
				return new Rectangle(size.width-_iconWidth, OFFSET+ (rank+1)*20,_iconWidth, _iconHeight);

			if (constraints.getAlignement() == EASTSOUTH)
				return new Rectangle(size.width-_iconWidth, 200 + (OFFSET+(rank+1)*20), _iconWidth, _iconHeight);
		}
		// Constraints by coordinates.
		Rectangle r=component.getBounds();
		Dimension d = component.getPreferredSize();
		r.width = d.width;
		r.height = d.height;
		return r;
	}
}