/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.command.ChangeCompositeStatusEdit;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotMask;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.ui.SubroutineEditor;
import fr.greyc.pantheon.ariane.utils.SubroutineTreeItem;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;

/**
 * An <code>CompositeItem</code> represents an item with
 * an inner subgraph.
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - July 17, 2010 (storable in archive)
 * @version Régis CLOUARD - Jan 04, 2012 (add editable/uneditable routines)
 */
public abstract class CompositeItem extends ControlStructure {
	/** The related subgraph. */
	private SubGraph _subGraph=null;
	/** The methods that depends on the current state. */
	private State _state;

	/**
	 * Creates a new composite composite item with the given
	 * attributes values.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param identifier the identification number (or -1).
	 * @param name the name of the item.
	 */
	public CompositeItem( TreeItem treeItem, Graph graph, int identifier, String name ) {
		super(treeItem, graph, identifier, name);
		_state= new Editable();
		inSlotMask = new InSlotMask(this);
	}

	/**
	 * Creates a new composite item with the given
	 * attributes values.
	 * 
	 * @param graph the parent graph.
	 * @param identifier the identification number (or -1).
	 */
	public CompositeItem( Graph graph, int identifier ) {
		super(graph, identifier);
		_state= new Editable();
		inSlotMask = new InSlotMask(this);
	}

	/**
	 * Creates a new subroutine item (uneditable item)
	 * with the given attributes values.
	 * 
	 * @param graph the parent graph.
	 * @param identifier the identification number (or -1).
	 * @param treeItem the related tree item.
	 */
	public CompositeItem( Graph graph, int identifier, TreeItem treeItem ) {
		super(graph, identifier);
		_treeItem = treeItem;
		inSlotMask = new InSlotMask(this);
		_state = new Uneditable();
	}

	/**
	 * Builds or changes the prototype of the macro-operator
	 * from the given values.
	 * Creates and changes also the related subgraph.
	 * 
	 * @param name the name of the macro-operator.
	 * @param iIm the number of input images.
	 * @param iP the number of input parameters.
	 * @param oIm the number of output images.
	 * @param iiIm the number of intermediate input images.
	 * @param iiP the number of intermediate input parameters.
	 */
	protected void setPrototype( String name, int iIm, int iP, int oIm, int iiIm, int iiP ) {
		adjustSlotsNumber(name, iIm, iP, oIm);
		createSubGraph(name, iIm, iP, oIm);
		draw();
	}

	/**
	 * Creates or removes input and output slots of the macro-operator
	 * from the given number of slots and from the existing number of slots.
	 * 
	 * @param name the name.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	protected void adjustSlotsNumber(String name, int iIm, int iP, int oIm) {
		// The name.
		if (!name.equals(getName())) {
			setNickName(name);
		}

		// Input image slots.
		if (iIm> inSlotImages.size()) { // adds slots
			for (int i=inSlotImages.size(); i<iIm; i++) {
				inSlotImages.add(new InSlotImage(this,i));
			}
		} else if (iIm< inSlotImages.size()) {// removes slots
			for (int i=inSlotImages.size()-1; i>=iIm; i--) {
				InSlotImage slot=inSlotImages.remove(i);
				slot.disconnect();
				slot.dispose();
			}
		}

		// Input parameters slots.
		if (iP> inSlotParameters.size()) { // adds slots
			for (int i=inSlotParameters.size(); i<iP; i++) {
				inSlotParameters.add(new InSlotParameter(this,i));
			}
		} else if (iP< inSlotParameters.size()) {// removes slots
			for (int i=inSlotParameters.size()-1; i>=iP; i--) {
				InSlotParameter slot=inSlotParameters.remove(i);
				slot.disconnect();
				slot.dispose();
			}
		}

		// Output image slots.
		if (oIm> outSlotImages.size()) { // adds slots
			for (int i=outSlotImages.size(); i<oIm; i++) {
				outSlotImages.add(new OutSlotImage(this,i));
			}
		} else if (oIm< outSlotImages.size()) {// removes slots
			for (int i=outSlotImages.size()-1; i>=oIm; i--) {
				OutSlotImage slot=outSlotImages.remove(i);
				slot.disconnectAll();
				slot.dispose();
			}
		}
	}

	/**
	 * Creates the subgraph with the convenient prototype.
	 * 
	 * @param name the name of the macro-operator.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 * @param iiIm the number of intermediate input images.
	 * @param iiP the number of intermediate input parameters.
	 * @return the subgraph.
	 */
	protected final SubGraph createSubGraph( String name, int iIm, int iP, int oIm, int iiIm,int iiP ) {
		_subGraph = new SubGraph(this, name, iIm, iP, oIm, iiIm, iiP);
		return _subGraph;
	}

	/**
	 * Creates the subgraph with the convenient prototype.
	 * 
	 * @param name the name of the macro-operator.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 * @return the subgraph.
	 */
	protected abstract SubGraph createSubGraph( String name, int iIm, int iP, int oIm );

	/** The color of the item. */
	private static final Color _defaultColor=new Color(255,102,204);

	/** The color of the item. */
	private static Color _color=_defaultColor;

	/**
	 * Changes the current color.
	 * 
	 * @param col the new color.
	 */
	public static final void setColor( Color col ) { _color = col; }

	/**
	 * Returns the current color.
	 * 
	 * @return the color.
	 */
	public static final Color getColor( ) { return _color; }

	/**
	 * Returns the default color.
	 * 
	 * @return the color.
	 */
	static public final Color getDefaultColor( ) { return _defaultColor; }

	/**
	 * Returns the background color (same as geColor()).
	 * 
	 * @return the color.
	 */
	@Override
	public final Color getBackgroundColor( ) {
		if (_state==null) {
			return super.getColor();
		} else {
			return _state.getBackgroundColor();
		}
	}

	/**
	 * Sets the item nickname.
	 * 
	 * @param name the new nickname.
	 */
	@Override
	public final void setNickName( String name ) {
		super.setNickName(name);
		Ariane.workspace.renameGraph(_subGraph, name);
	}

	/**
	 * Removes temporary the item from the graph.
	 * Disconnects it from all other item.
	 */
	@Override
	public final void remove() {
		super.remove();
		_subGraph.closeFrame();
	}

	/** 
	 * Disposes definitively the item.
	 * Tries to avoid memory leak.
	 */
	@Override
	public final void dispose() {
		_subGraph.dispose();
		super.dispose();
		_subGraph=null;

	}

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * Returns the related subgraph.
	 * @return the subgraph.
	 */
	@Override
	public final SubGraph getSubGraph() { return _subGraph; }

	/**
	 * Opens the related subgraph and shows its content.
	 */
	protected void viewSubgraph() {
		_subGraph.showContent();
	}

	///// 
	///// M E N U
	/////

	/**
	 * Builds the contextual menu as a Popup Menu.
	 */
	@Override
	protected final JPopupMenu initPopup( ) {
		return _state.initPopup();
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Resets the composite item and its related sub-graph.
	 */
	@Override
	public final void resetAll( ) {
		super.resetAll();
		if (_subGraph!=null)
			_subGraph.resetAll();
	}

	/**
	 * Resets the items.
	 */
	@Override
	protected void reset() {
		super.reset();
		for (int i=1; i<_subGraph.getGraphItems().size(); i++) {
			_subGraph.getGraphItems().get(i).reset();
		}
	}

	/**
	 * @return the virtual operator of the subgraph.
	 */
	protected VirtualItem getVirtualOperator() {
		return _subGraph.getVirtualOperator();
	}

	///// 
	///// I N P U T   /   O U T P U T
	/////

	/**
	 * Saves the composite item as an XML tree.
	 * Saves also the subgraph.
	 * 
	 * @param file the current filename of the project.
	 * @return the XML tree containing the subgraph.
	 * @throws IOException any XML errors.
	 */
	@Override
	public final XMLElement save( File file ) throws IOException {
		XMLElement e = new XMLElement(tagName());
		LinkedList<XMLElement> xmlTree = saveProperties(file);
		if (_treeItem instanceof SubroutineTreeItem && _state instanceof Uneditable) {
			xmlTree.add(new XMLElement("SubroutineReference",_treeItem.getName()));
		}
		e.addChildren(xmlTree);
		e.addChildren(_subGraph.getXML(file));
		return e;
	}

	/**
	 * Loads a composite structure from an XML tree.
	 * 
	 * @param xmlTree the XML tree of data.
	 * @throws IOException any XML errors.
	 */
	@Override
	public final void load( LinkedList<XMLElement> xmlTree ) throws IOException {
		loadProperties(xmlTree);
		boolean subroutine=false;
		XMLElement e;
		Iterator<XMLElement> it = xmlTree.iterator();
		while (it.hasNext()){
			e = it.next();
			if (e.getName().equalsIgnoreCase("SubroutineReference")) {
				subroutine=(_treeItem instanceof SubroutineTreeItem);
			}
		}
		SubGraph subGraph = createSubGraph(getName(),inSlotImages.size(),inSlotParameters.size(), outSlotImages.size());
		if (subroutine) {
			_state = new Uneditable();
			subGraph.setXML(((SubroutineTreeItem)_treeItem).getBody());
		} else {
			subGraph.setXML(xmlTree);
		}
	}

	/**
	 * @return the related tree item in the explorer if any.
	 */
	@Override
	protected TreeItem getRelatedTreeItem( ) {
		return Ariane.workspace.getTree().getTreeItem(getName(), getCategory());
	}

	///// 
	///// P R O P E R T I E S
	/////

	/**
	 * Invoked when the mouse button has been
	 * double clicked (pressed and released) on a graph item.
	 * Shows the dialog box for setting the item's properties.
	 * 
	 * @param e the event attributes.
	 */
	@Override
	protected void onMouseDoubleClicked( MouseEvent e ) {
		_state.onMouseDoubleClicked(e);
	}

	/**
	 * Asks interactively the prototype of the Macro.
	 */
	protected abstract void askProperties();

	protected abstract TreeItem restoreDefaultTreeItem();

	public void changeEditableStatus( boolean editable ) {
		if (editable) {
			_state = new Editable();
			_treeItem = restoreDefaultTreeItem();
		} else {
			_state = new Uneditable();
		}
		draw();
	}

	/**
	 * The state design pattern to manage the composite as
	 * an uneditable subroutine or an editable macro.
	 */
	abstract class State {
		public abstract JPopupMenu initPopup( );
		public abstract void onMouseDoubleClicked( MouseEvent e );
		public abstract Color getBackgroundColor();
	}

	class Uneditable extends State {
		@Override
		public final JPopupMenu initPopup( ) {
			JPopupMenu popupMenu = new JPopupMenu();
			JMenuItem item;

			item = new JMenuItem(Ariane.lang.getString("PopupMenu.execute"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed(ActionEvent e) { getGraph().execute(); }
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
			popupMenu.add(item);

			item = new JMenuItem(Ariane.lang.getString("PopupMenu.reset"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed(ActionEvent e) { resetAll(); }
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
			popupMenu.add(item);

			item = new JMenuItem(Ariane.lang.getString("PopupMenu.remove"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed(ActionEvent e) { delete(); }
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
			popupMenu.add(item);

			popupMenu.addSeparator();
			item = new JMenuItem(Ariane.lang.getString("PopupMenu.convertAsMacro"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed(ActionEvent e) {
					changeEditableStatus(true);
					getGraph().setModified(true);
					getGraph().startActionEdition();
					getGraph().validateAction(new ChangeCompositeStatusEdit(CompositeItem.this, true));
					getGraph().endActionEdition();
				}
			});
			popupMenu.add(item);
			return popupMenu;
		}
		@Override
		public void onMouseDoubleClicked( MouseEvent e ) {
			Ariane.workspace.printWarning(getNickName()+": "+Ariane.lang.getString("CompositeItem.message1"));
		}
		@Override
		public Color getBackgroundColor() { 
			return _color;
		}
	}

	class Editable extends State {
		@Override
		public void onMouseDoubleClicked( MouseEvent e ) {
			viewSubgraph();
		}

		@Override
		public final JPopupMenu initPopup( ) {
			JPopupMenu popupMenu = new JPopupMenu();
			JMenuItem item;

			item = new JMenuItem(Ariane.lang.getString("PopupMenu.execute"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed(ActionEvent e) { getGraph().execute(); }
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
			popupMenu.add(item);

			item = new JMenuItem(Ariane.lang.getString("PopupMenu.reset"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed(ActionEvent e) { resetAll(); }
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
			popupMenu.add(item);

			item = new JMenuItem(Ariane.lang.getString("PopupMenu.remove"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed(ActionEvent e) { delete(); }
			});
			item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
			popupMenu.add(item);

			popupMenu.addSeparator();
			item = new JMenuItem(Ariane.lang.getString("PopupMenu.store"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed(ActionEvent e){
					store();
				}
			});
			popupMenu.add(item);

			popupMenu.addSeparator();
			item = new JMenuItem(Ariane.lang.getString("PopupMenu.openSubgraph"));
			item.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed(ActionEvent e) { viewSubgraph(); }
			});
			popupMenu.add(item);
			return popupMenu;
		}
		public final void store() {
			SubroutineEditor g = new SubroutineEditor(CompositeItem.this);
			g.setVisible(true);
			if (g.isValidated()) {
				changeEditableStatus(false);
				getGraph().setModified(true);
				getGraph().startActionEdition();
				getGraph().validateAction(new ChangeCompositeStatusEdit(CompositeItem.this, false));
				getGraph().endActionEdition();
			}
		}

		@Override
		public Color getBackgroundColor() {
			return ControlStructure.getColor();
		}
	}
}