/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.command.ChangeArgumentValueEdit;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;

/**
 * A <code>FileNameItem</code> implements a filename item
 * other than an image file. Such file is used as a parameter
 * of an other item.
 * 
 * @author Régis CLOUARD - Dec 12, 2008
 * @author Régis CLOUARD - Jan 21, 2011
 * @author Régis CLOUARD - Apr 23, 2011
 * @author Régis CLOUARD - Jul 14, 2011 (fixed bug)
 * @author Régis CLOUARD - Jan 12, 2013 (add file edition facility)
 */
public class FileNameItem extends ControlStructure implements IScriptArgument {
	/** The tag name of the item. */
	public static String TAGNAME="Filename";
	/** Is this item a script argument? */
	private boolean _isArgument=false;
	/** The check box used in the property window. */
	private JCheckBox _checkBox;
	/** The button used to load a file. */
	private JButton _browser;
	/** The date of the last modification of the related file. */
	private long _lastModified=0;

	/**
	 * Creates a <code>FileNameItem</code> with the given
	 * attributes.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the graph item.
	 * @param identifier the identification number (or -1).
	 */
	public FileNameItem( TreeItem treeItem, Graph graph, String name, int identifier ) {
		super(treeItem, graph, identifier,name);
		super.setFormula(""); // IMPERATIVE to signify that it uses formula.
		setComment(getAbstractFileName());
		draw();
	}

	/**
	 * Creates a <code>FileNameItem</code> on the given
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param elements the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public FileNameItem( Graph graph, LinkedList<XMLElement> elements ) throws IOException {
		super(graph, -1);
		super.setFormula(""); // IMPERATIVE to signify that it uses formula.
		load(elements);
		setComment(getAbstractFileName());
		draw();
	}

	/**
	 * @return true if the item is an argument of the
	 * script that can be generated from the graph.
	 */
	@Override
	public boolean isArgument() {
		return _isArgument;
	}

	/**
	 * Sets whether the item is a script argument or not.
	 * 
	 * @param value the new status value.
	 */
	@Override
	public void setIsArgument(boolean value) {
		_isArgument =value;
	}

	/**
	 * Builds or changes the prototype of the control structure
	 * from the given values.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected final void setPrototype( String name, int iIm, int iP, int oIm ) {
		setAbstractName(name);
	}

	/**
	 * @return the list of specific properties (other than common properties).
	 */
	@Override
	public ArrayList<PropertyViewer.PropertyCouple> getSpecificProperties( ) {
		ArrayList<PropertyViewer.PropertyCouple> properties = new ArrayList<PropertyViewer.PropertyCouple>();

		_checkBox = new JCheckBox();
		_checkBox.setSelected(_isArgument);
		_checkBox.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				_isArgument=((JCheckBox)e.getSource()).isSelected();
				getGraph().setModified(true);
				getGraph().startActionEdition();
				getGraph().validateAction(new ChangeArgumentValueEdit(FileNameItem.this, _checkBox, _isArgument));
				getGraph().endActionEdition();
			}
		});
		properties.add(new PropertyViewer.PropertyCouple(Ariane.lang.getString("PropertyViewer.argument"), _checkBox));
		return properties;
	}

	/**
	 * Invoked when the mouse button has been
	 * double clicked (pressed and released) on a graph item.
	 * Shows the dialog box for setting the item's properties.
	 * 
	 * @param e the event attributes.
	 */
	@Override
	protected void onMouseDoubleClicked( MouseEvent e ) {
		if (getFormula().trim().length()==0) {
			_browser.doClick();
		} else {
			String command = PreferenceChooser.get("TextEditor");
			if (command!=null && command.length()>0) {
				try {
					Runtime.getRuntime().exec(command+" "+getFormula());
				} catch (Exception e1) {
					Ariane.workspace.printErrorMessage(Ariane.lang.getString("FileNameItem.error1")+" "+command);
				}
			} else {
				Ariane.workspace.printErrorMessage(Ariane.lang.getString("FileNameItem.error2"));
			}
		}
	}

	/**
	 * Sets the name of the related physical file.
	 * If the name of the file is specified as a relative path 
	 * than the absolute name is built with the directory of the project.
	 * 
	 * @param formula a file name (absolute or relative)
	 * @param directory a home directory for files.
	 */
	@Override
	protected final void setFormula( String formula ) {
		if (formula.length()>0) {
			String current_directory = null;
			if (getGraph().getRoot().getFileName()!=null) {
				current_directory = getGraph().getRoot().getFileName().getParent();
			}
			super.setBinary(DataFile.setPathToFilename(formula, current_directory));
		}
	}

	/**
	 * Returns the formula of the logical or arithmetic expression.
	 * 
	 * @param textFormula the formula.
	 * @return true if the formula is valid, false otherwise.
	 */
	@Override
	public final boolean changeFormula( String textFormula ) {
		if (getFormula()!=null && getFormula().equals(textFormula)) {
			return true;
		}
		setFormula(textFormula);
		setComment(getAbstractFileName());

		return true;
	}

	/**
	 * Returns the name of the file either with relative path if the file is
	 * in the filePath directory or with an absolute path otherwise.
	 * 
	 * @param directory a home directory.
	 * @return the file name.
	 */
	@Override
	public String getFormula( String directory ) {
		return DataFile.getRelativePath(super.getFormula(), directory);
	}

	/**
	 * @return the text associated to the formula editor.
	 */
	@Override
	public String getFormulaLabel( ) {
		return Ariane.lang.getString("PropertyViewer.filename.formula");
	}

	/**
	 * Builds a component convenient to display and change the formula.
	 * 
	 * @return a new component.
	 */
	@Override
	public JComponent getFormulaEditor( ) {
		_browser = new JButton();
		_browser.setText(getFormulaText());
		_browser.addActionListener( new ActionListener(){
			@Override
			public void actionPerformed(ActionEvent e) {
				final String currentFile = getFormula();
				final JFileChooser fileChooser = new JFileChooser();
				if ( currentFile==null || currentFile.trim().length()==0 || !(new File(currentFile).exists())) {
					fileChooser.setCurrentDirectory(new File(Workspace.getCurrentDir()));
				} else {
					fileChooser.setSelectedFile(new File(currentFile));
				}
				fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
				fileChooser.setDialogTitle(Ariane.lang.getString("FileChooser.titleOpenFile"));
				int returnVal = fileChooser.showOpenDialog(Ariane.frame);
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					changeFormula(_browser, fileChooser.getSelectedFile().getAbsolutePath(), true);
				}
			}
		});

		return _browser;
	}

	/**
	 * @return the current file name.
	 */
	@Override
	public String getFormulaText() {
		if (getFormula().length()>0) {
			File file = new File(getFormula());
			return file.getName();
		} else {
			return Ariane.lang.getString("PropertyViewer.browser");
		}
	}

	/**
	 * Resynchronizes the item with the related file.
	 * if the related file has been changed, resets the item.
	 */
	public void synchronizeWithFileSystem() {
		if (getFormula()!=null) {
			File file = new File(getFormula());
			if (_lastModified<file.lastModified()) {
				resetAll();
				_lastModified=file.lastModified();
			}
		}
	}

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public final String tagName( ) {
		return TAGNAME;
	}

	/**
	 * Returns the filename as the nickname of the
	 * item if any, otherwise returns the item class name.
	 * 
	 * @return a name.
	 */
	private final String getAbstractFileName( ) {
		if (getFormula()==null || getFormula().trim().length()==0) {
			return "";
		} else {
			File filename = new File(getFormula());
			return filename.getName();
		}
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes the <code>FileName</code>, means
	 * compute a result from the two parameter values.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			if (getFormula() == null || getFormula().length()==0) {
				throw new ArithmeticException(Ariane.lang.getString("arithmetic.error18"));
			}
			File file = new File(getFormula());
			_lastModified = file.lastModified();
			if (!file.exists()) {
				throw new ArithmeticException(Ariane.lang.getString("ImageFile.error3"));
			}
			getOutSlotResult().setValue(getFormula());
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArithmeticException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (Exception e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}
	}

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected final JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;

		// Action: executer
		item = new JMenuItem(Ariane.lang.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) {
				getGraph().execute();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) {
				resetAll();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		// Edition: delete
		item = new JMenuItem(Ariane.lang.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e){
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		return popupMenu;
	}
}