/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.etc.FileImagePreviewer;
import fr.greyc.pantheon.ariane.graph.item.etc.FileTypeViewer;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Component;
import java.awt.HeadlessException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

/**
 * A <code>LoaderItem</code> manages data files loading
 * within a graph.
 * 
 * @version Régis CLOUARD - Dec 09, 2007 (created from ImageFile)
 * @version Régis CLOUARD - Jun 20, 2010 (enhancement)
 * @version Régis CLOUARD - Jan 19, 2012 (add execution of a folder of images)
 */
public final class LoaderItem extends DataFile {
	/** The tag name of the item. */
	public static final String TAGNAME = "Loader";

	/** The button for loading a new file.  */
	private JButton _jbuttonSelect;
	/** The component that holds the current name. */
	private JTextField _nameEditor = new JTextField(getNickName());
	/** The component that holds the current name. */
	private JTextField _binaryEditor = new JTextField(getBinary());
	/** The date of the last modification of the related file. */
	private long _lastModified=0;

	private static StaticOpenFileChooser _fileChooser = new StaticOpenFileChooser();

	/**
	 * Creates a <code>LoaderItem</code> on the given graph
	 * with the specified properties.
	 * 
	 * @param treeItem the source tree item.
	 * @param graph the graph that holds the item.
	 * @param name the name of the item.
	 * @param identifier the identification number of the item (or -1).
	 * @param category the category name of the item.
	 */
	public LoaderItem( TreeItem treeItem, Graph graph, String name, int identifier, String category ) {
		super(treeItem, graph, name, identifier, null, category, "loader");
		outSlotImages.add(new OutSlotImage(this, 0));
		draw();
	}

	/**
	 * Creates a <code>LoaderItem</code> on the given graph
	 * with the given XML data.
	 * 
	 * @param graph the graph that holds the item.
	 * @param xmlTree an XML tree of data.
	 * @throws IOException XML errors.
	 */
	public LoaderItem( Graph graph,LinkedList<XMLElement> xmlTree ) throws IOException {
		super(graph, xmlTree, "loader");
		draw();
	}

	/**
	 * Returns the address of the related documentation HTML page.
	 * 
	 * @return the complete or relative address.
	 * kept for sake of compatibility.
	 */
	@Override
	public String getDocURL( ) {
		if (_treeItem!=null)
			return _treeItem.getDocURL();
		return "loader.html";
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public String tagName( ) {
		return TAGNAME;
	}

	/**
	 * Sets the name of the related physical file.
	 * If the name of the file is specified as a relative path 
	 * than the absolute name is built with the directory name
	 * of the project.
	 * 
	 * @param binary a file name (absolute or relative)
	 * @param directory a home directory for the binaries.
	 */
	@Override
	public void setBinary( String binary ) {
		String current_directory = null;
		if (getGraph().getRoot().getFileName()!=null) {
			current_directory = getGraph().getRoot().getFileName().getParent();
		}
		super.setBinary(DataFile.setPathToFilename(binary, current_directory));
	}

	/**
	 * Returns the name of the file either with relative path if the file is
	 * in the filePath directory or with an absolute path otherwise.
	 * 
	 * @param filePath a home directory for the files.
	 * @return the file name.
	 */
	@Override
	public String getBinary( String directory ) {
		return getRelativePath(super.getBinary(), directory);
	}

	/**
	 * Resynchronizes the item with the related file.
	 * if the related file has been changed, resets the item.
	 */
	public void synchronizeWithFileSystem() {
		if (getBinary()!=null) {
			File file = new File(getBinary());
			if (_lastModified<file.lastModified()) {
				resetAll();
				_lastModified=file.lastModified();
			}
		}
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes a Loader. It means: loads the file given
	 * by the user in the graph.
	 * If it exists, sets the current state to EXECUTED,
	 * otherwise sets the current state to ERROR.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			_currentFilename = null;
			if (getBinary() == null || getBinary().length()==0)
				throw new DataFileIOException(Ariane.ErrorCode.MISSINGFILE, Ariane.lang.getString("ImageFile.error3"));
			else {
				File file = new File(getBinary());
				_lastModified = file.lastModified();
				if (!file.exists()) {
					throw new DataFileIOException(Ariane.ErrorCode.UNKNOWNFILE,Ariane.lang.getString("ImageFile.error4")+" \""+getBinary()+"\"");
				}
			}
			(outSlotImages.getFirst()).setValue(getBinary());
			_currentFilename = getBinary();
		} catch (DataFileIOException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(e.errorCode, getAbsoluteName() +"("+getId()+"): "+e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}

		// Validate the result.
		changeStatus(Status.EXECUTED);
		return true;
	}

	///// 
	///// M E N U 
	/////

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();
		JMenuItem item;

		item= new JMenuItem(Ariane.lang.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) { getGraph().execute();}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item= new JMenuItem(Ariane.lang.getString("PopupMenu.executeDirectory"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) { getGraph().executeFolder(LoaderItem.this); }
		});
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) { resetAll(); }
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) { delete(); }
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		popupMenu.addSeparator();

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.selectFile"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) { selectFile(); }
		});
		popupMenu.add(item);

		return popupMenu;
	}

	/**
	 * Opens a windows from which the user can select a file.
	 * Keeps the selected directory as the current directory.
	 */
	private void selectFile( ) {
		String path;
		String currentFile=getBinary();

		if ( currentFile==null || currentFile.trim().length()==0 || !(new File(currentFile).exists())) {
			if (_currentLoadingDir == null) {
				_currentLoadingDir = Workspace.getCurrentDir();
			}
			_fileChooser.setCurrentDirectory(new File(_currentLoadingDir));
		} else {
			File f = new File(currentFile);
			_fileChooser.setSelectedFile(f);
		}
		if (_fileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
			try {
				path = ""+_fileChooser.getSelectedFile();
				_currentFilename=null;
				_currentLoadingDir = _fileChooser.getSelectedFile().getParent();
				changeBinaryName(_nameEditor, _binaryEditor, _jbuttonSelect, path, true);
			} catch (Exception e) {
				Ariane.workspace.printErrorMessage(Ariane.lang.getString("ImageFile.error1") + e);
			}
		}
	}

	/**
	 * @return the list of specific properties (other than common properties).
	 */
	@Override
	public ArrayList<PropertyViewer.PropertyCouple> getSpecificProperties( ) {
		ArrayList<PropertyViewer.PropertyCouple> properties = new ArrayList<PropertyViewer.PropertyCouple>();

		String name = "";
		if (getBinary()!=null) {
			File file = new File(getBinary());
			name = file.getName();
		}

		_jbuttonSelect = new JButton();
		if (name.length()==0)
			_jbuttonSelect.setText(Ariane.lang.getString("PropertyViewer.load")+"...");
		else 
			_jbuttonSelect.setText(name);
		_jbuttonSelect.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				selectFile();
			}
		});
		properties.add(new PropertyViewer.PropertyCouple(Ariane.lang.getString("PropertyViewer.fileName"), _jbuttonSelect));
		return properties;
	}

	/**
	 * Invoked when the mouse button has been
	 * double clicked (pressed and released) on a graph item.
	 * Shows the dialog box for setting the item's properties.
	 * 
	 * @param e the event attributes.
	 */
	@Override
	protected void onMouseDoubleClicked( MouseEvent e ) {
		_jbuttonSelect.doClick();
	}

	/**
	 * @return a component appropriate to edit and possibly change the name of the item.
	 */
	@Override
	public JComponent getNameEditor() {
		_nameEditor = new JTextField(getNickName());
		_nameEditor.setEditable(false);
		_nameEditor.setEnabled(false);
		return _nameEditor;
	}

	/**
	 * @return a component appropriate to edit and possibly change the name of the item.
	 */
	@Override
	public JComponent getBinaryEditor() {
		return null;
	}
	/**
	 * Defined to keep the size and location of the 
	 * previous opening of the JFileChooser;
	 */
	private static class StaticOpenFileChooser extends JFileChooser {
		private JDialog _dialog;
		public StaticOpenFileChooser() {
			setFileView(new FileTypeViewer());
			setDialogTitle(Ariane.lang.getString("FileChooser.titleOpenFile"));
			setFileSelectionMode(JFileChooser.FILES_ONLY);
			setAccessory(new FileImagePreviewer(this));
		}

		@Override
		protected JDialog createDialog( Component parent ) throws HeadlessException {
			JDialog dialog = super.createDialog(parent);
			if (_dialog!=null) {
				dialog.setLocation(_dialog.getLocation());
				dialog.setSize(_dialog.getSize());
			}
			_dialog=dialog;
			return dialog;
		}
	}
}