/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.util.StringTokenizer;
import java.util.LinkedList;

import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>NoteItem</code> is a comment item that can be added to a graph.
 * It contains only text and is visualized as a <code>JTextArea</code>.
 *
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 26, 2004 (Prefer formula rather than binary for the text).
 * @version Régis CLOUARD - Nov 22, 2007 (fix bug on print)
 * @version Régis CLOUARD - Mar 30, 2011 (min size)
 */
public final class NoteItem extends GraphItem implements FocusListener {
	/** The tag name of the item. */
	public static final String TAGNAME="Note";

	/** The text to be displayed. */
	private JTextArea _labelNote;

	/**
	 * Creates a <code>NoteItem</code> on the given graph with the given properties.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the item.
	 * @param identifier the identification number (or -1).
	 */
	public NoteItem( TreeItem treeItem, Graph graph, String name, int identifier) {
		super(treeItem, graph,identifier,name,null);
		draw();
	}

	/**
	 * Creates a <code>NoteItem</code> on the given graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public NoteItem( Graph graph, LinkedList<XMLElement> xmlTree ) throws IOException {
		super(graph,-1);
		load(xmlTree);
		if (getBinary() != null) { // KEEP THIS FOR COMPATIBILITY REASONS
			setFormula(getBinary());
			setBinary(null);
		}
		draw();
	}

	/**
	 * Returns the description of the note.
	 * 
	 * @return a string.
	 */
	@Override
	public final String toString( ) {
		return "("+getId()+"):"+TAGNAME;
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public final String tagName( ) {
		return TAGNAME;
	}

	/**
	 * Invoked when the mouse button has been
	 * double clicked (pressed and released) on a component.
	 * Overrides the <code>GraphItem.onMouseDoubleClicked()</code>.
	 * Shows the man page of this item.
	 * 
	 * @param e the event attributes.
	 */
	@Override
	protected final void onMouseDoubleClicked( MouseEvent e ) {
		if (!_labelNote.isEditable())
			showMan();
	}

	/**
	 * Invoked when a JtextArea looses the keyboard focus.
	 * Change the text of the related element in the model.
	 * 
	 * @param e the FocusEvent.
	 */
	@Override
	public final void focusLost( FocusEvent e ) {
		if (e.getSource()==_labelNote) {
			String newFormula=_labelNote.getText();
			// Change formula if it was changed!
			if (getFormula()==null || !getFormula().equals(newFormula)) {
				setFormula(_labelNote.getText());
				getGraph().setModified(true);
			}
		}
		_labelNote.setEditable(false);
		_labelNote.getCaret().setVisible(false);
		_labelNote.setCaretPosition(0);
	}

	/**
	 * Invoked when a component gains the keyboard focus.
	 * 
	 * @param e the FocusEvent.
	 */
	@Override
	public final void focusGained( FocusEvent e ) {
		if (!isSelected()) {
			GraphItem g = NoteItem.this;
			g.getGraph().addItemInSelection(g,false);
		} else {
			_labelNote.setCaretPosition(0);
			_labelNote.moveCaretPosition(_labelNote.getText().length());
		}
	}

	// Notice: theses methods cannot be generalized in the GraphItem class
	// because they use static attributes.

	/** The color of the item. */
	private static final Color _defaultColor=new Color(255,255,51);
	/** The color of the item. */
	private static Color _color=_defaultColor;

	/**
	 * Changes the current color.
	 * 
	 * @param col the new color.
	 */
	public static final void setColor( Color col ) {
		_color = col;
	}

	/**
	 * Returns the current color.
	 * 
	 * @return the color.
	 */
	 public static final Color getColor( ) {
		return _color;
	}

	/**
	 * Returns the default color.
	 * 
	 * @return the color.
	 */
	public static final  Color getDefaultColor( ) {
		return _defaultColor;
	}

	/**
	 * Returns the background color (same as geColor()).
	 * 
	 * @return the color.
	 */
	@Override
	public final Color getBackgroundColor( ) {
		return _color;
	}

	///// 
	///// D I S P L A Y I N G   /   P R I N T I N G
	/////

	/**
	 * Overrides the paint method of GraphItem, because a note
	 * has a different look.
	 */
	@Override
	public final void draw( ) {
		removeAll();
		setBackground(getBackgroundColor());
		setBorder(unselectedBorder);
		JLabel labelName= new JLabel();
		labelName.setFont(PreferenceChooser.defaultFont);
		labelName.setForeground(new Color(0,0,0));
		labelName.setText(getName());
		setToolTipText(toString());
		// layout
		GridBagConstraints c = new GridBagConstraints();
		c.insets = new Insets(2, 2, 3, 2);
		c.gridheight = 1;

		c.gridwidth = 1;
		c.gridy++; c.gridx = 0;
		c.anchor = GridBagConstraints.CENTER;
		add(labelName, c);
		if (_labelNote==null) {
			if (getFormula() == null) {
				_labelNote = new JTextArea("",1,2);
			} else {
				_labelNote = new JTextArea(getFormula(),1,2);
			}
		}
		_labelNote.setFont(PreferenceChooser.defaultFont);
		_labelNote.setForeground(new Color(0,0,0));
		// Mouse events.
		_labelNote.addMouseListener(new MouseAdapter( ) {
			@Override
			public void mouseClicked( MouseEvent e ) {
				if (e.getClickCount()>1)
					NoteItem.this.onMouseDoubleClicked(e);
			}
			@Override
			public void mousePressed( MouseEvent e ) {
				NoteItem.this.onMousePressed(e);
			}
			@Override
			public void mouseReleased( MouseEvent e ) {
				if (isSelected()) {
					_labelNote.setEditable(true);
					_labelNote.getCaret().setVisible(true);
				} else {
					NoteItem.this.onMouseReleased(e);
				}
			}
		});

		_labelNote.addFocusListener(this);
		_labelNote.setEditable(false);
		_labelNote.getCaret().setVisible(false);

		// The lights ramp that shows the current state.
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.gridy++; c.gridx = 0;
		add(_labelNote, c);
	}

	/**
	 * Overrides the default method in order to print components
	 * of the operator in a different way than they are displayed.
	 * 
	 * @param gc the graphic context.
	 */
	@Override
	public final void printComponent( Graphics gc ) {
		super.printComponent(gc);
		FontMetrics fm;
		fm = gc.getFontMetrics();
		int row = fm.getHeight();
		super.centerText(getName(),gc,row,false);
		row += fm.getHeight()+4;
		gc.setFont(PreferenceChooser.defaultFont);
		if (getFormula() != null) {
			StringTokenizer st = new StringTokenizer(getFormula(),"\n");
			while (st.hasMoreTokens()) {
				gc.drawString(st.nextToken(),5,row);
				row += fm.getHeight();
			}
		}
	}

	/**
	 * Executes the note. In fact does nothing.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		return false;
	}

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected final JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;
		item = new JMenuItem(Ariane.lang.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener( ) {
			@Override
			public void actionPerformed( ActionEvent e ){
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);
		return popupMenu;
	}

	/**
	 * Creates a related property viewer.
	 * 
	 * @return a dialog box.
	 */
	@Override
	protected final PropertyViewer createPropertyViewer() {
		return BLANK_PROPERTY_PANEL;
	}
}