/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.etc.ArgumentLineParser;
import fr.greyc.pantheon.ariane.graph.item.etc.OperatorProcessBuilder;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotMask;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotResult;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.LinkedList;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;

/**
 * An <code>OperatorItem</code> is a graph item that encapsulates
 * a Shell command. It uses inputs to produce outputs with
 * specific parameter values.
 * @see GraphItem
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 06, 2004
 * @version Régis CLOUARD - Nov 24, 2007 (add Pandore path to binary operator name)
 * @version Régis CLOUARD - Jun 29, 2009 (add operator output messages)
 */
public final class OperatorItem extends GraphItem {
	/** The tag name of the item. */
	public static final String TAGNAME="Operator";
	public static final String separator = "&";

	/**
	 * Creates an <code>OperatorItem</code> on the given graph with
	 * the given properties. Initializes the mouse listeners,
	 * the popup menu, and display it in the graph.
	 * If the identifier =-1 then creates a unique identifier number.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the item.
	 * @param identifier the identification number (or -1).
	 * @param category the name of the category of the operator.
	 * @param inImages the number of input images.
	 * @param inParams the number of parameters.
	 * @param outImages the number of output images.
	 * @param signature a string of the form : $0:toto=1; $1:titi=10;
	 */
	public OperatorItem( TreeItem treeItem, Graph graph, String name, int identifier, String category, int inImages, int inParams, int outImages, String signature ) {
		super(treeItem, graph, identifier, name, category);
		inSlotMask = new InSlotMask(this);
		outSlotResults.add(new OutSlotResult(this,0));
		for (int i=0; i<inImages; i++) { inSlotImages.add(new InSlotImage(this,i)); }
		for (int i=0; i<outImages; i++) { outSlotImages.add(new OutSlotImage(this, i)); }
		for (int i=0; i<inParams; i++) { inSlotParameters.add(new InSlotParameter(this,i)); }
		setSignature(signature);
		draw();
	}

	/**
	 * Creates an <code>OperatorItem</code> on the given graph
	 * with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public OperatorItem( Graph graph, LinkedList<XMLElement> xmlTree ) throws IOException {
		super(graph,-1);
		// Creates mask and result because they are not saved in the xmlTree.
		inSlotMask = new InSlotMask(this);
		outSlotResults.add(new OutSlotResult(this,0));
		load(xmlTree);
		if (_treeItem!=null) {
			setSignature(_treeItem.getSignature());
		}
		draw();
	}

	/**
	 * Returns the description of the operator: name + parameters value.
	 * 
	 * @return a string description of the operator.
	 */
	@Override
	public final String toString( ) {
		String commandLine="("+getId()+"): " + getName();

		for (int i = 0; i < inSlotParameters.size(); i++){
			InSlotParameter isp = (inSlotParameters.get(i));
			if (isp.getValue() != null)
				commandLine += " " +isp.getValue();
		}
		return commandLine;
	}

	/** 
	 * Returns the tag name.
	 * 
	 * @return the tag name.
	 */
	@Override
	public final String tagName( ) { return TAGNAME; }

	///// 
	///// P R E F E R E N C E S
	///// 

	// Note: theses methods cannot be generalized in the GraphItem class
	// because they use static attributes.

	/** The color of the item. */
	private static final Color _defaultColor=new Color(0,153,255);

	/** The color of the item. */
	private static Color _color=_defaultColor;

	/**
	 * Changes the current color.
	 * 
	 * @param color the new color.
	 */
	public static final void setColor( Color color ) { _color = color; }

	/**
	 * Returns the current color.
	 * 
	 * @return the color.
	 */
	public static final Color getColor( ) { return _color; }

	/**
	 * Returns the default color.
	 * 
	 * @return the color.
	 */
	public static final Color getDefaultColor( ) { return _defaultColor; }

	/**
	 * Returns the background color (same as geColor()).
	 * 
	 * @return the color.
	 */
	@Override
	public final Color getBackgroundColor( ) { return _color; }

	///// 
	///// M E N U
	///// 

	/**
	 * Builds the contextual menu as a Popup Menu.
	 */
	@Override
	protected final JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();
		JMenuItem item;

		// Action: executer
		item = new JMenuItem(Ariane.lang.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener( ) {
			@Override
			public final void actionPerformed( ActionEvent e ) { getGraph().execute(); }
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) {
				resetAll();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e){ delete(); }
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		return popupMenu;
	}

	///// 
	///// P R I N T I N G
	///// 

	/**
	 * Overrides the default method in order to print
	 * the operator in different way that is is normally displayed.
	 * Prints the name of the operator and each parameter value.
	 * 
	 * @param gc the graphics context.
	 */
	@Override
	public final void printComponent( Graphics gc ) {
		super.printComponent(gc);
		FontMetrics fm = gc.getFontMetrics();
		int row = fm.getHeight();

		// The name of the operator.
		super.centerText(getNickName(), gc, row,false);
		// Print each parameters.
		// Print the value only of if the parameter is not connected.
		for (int i = 0; i < inSlotParameters.size(); i++) {
			InSlotParameter param = inSlotParameters.get(i);
			row += (fm.getHeight()+1);
			if (!param.isConnected())
				super.centerText("$"+i+":"+param.getValue(), gc, row,true);
			else
				super.centerText("$"+i+":<linked>", gc, row,true);
		}
	}

	///// 
	///// E X E C U T I O N
	///// 

	/**
	 * Returns the relative name of the specified file
	 * if it is a temporary file otherwise returns
	 * the same file name.
	 * 
	 * @param filename a file name;
	 * @return a relative file name.
	 */
	private final String getRelativeFilePathName( String filename ) {
		String a = Workspace.getTempDir();
		if (filename.startsWith(a))
			return filename.substring(a.length());
		return filename;
	}

	/**
	 * Returns the name of the executable.
	 * 
	 * @return the binary name.
	 */
	public static final String getCommandPart( String binary ) {
		if (binary == null ) return null;
		String[] format =  binary.split(OperatorItem.separator);
		if (format.length>0) {
			return format[0];
		}
		return null;
	}

	public static final String getFormatPart( String binary ) {
		if (binary!=null) {
			int index = binary.indexOf(OperatorItem.separator);
			if (index>0 && index+1<binary.length()) {
				return binary.substring(index+1);
			} else {
				return "[%p] [-m %m] [%i] [%o]";
			}
		}
		return null;
	}

	/**
	 * Returns the command line of the operator.
	 * 
	 * @return the binary name.
	 */
	@Override
	public final String getBinary( ) {
		return (_treeItem==null)? super.getBinary() : _treeItem.getBinary();
	}

	/**
	 * Builds the command line as specified in the command format
	 * and with the arguments. The binary name of the program
	 * is returns with the absolute pathname.
	 * Syntax: format string uses:
	 *  %p for parameter; %p0 for the first parameter
	 *  %m for the mask
	 *  %i for input; %i0 for the first input image
	 *  %o for output; %o0 for the first output image.
	 *  
	 * @param maskLine the mask value.
	 * @param parameterLine the parameter values.
	 * @param outputLine the output values.
	 * @return the command line.
	 */
	public final LinkedList<String> buildAbstractCommandLine( LinkedList<String> maskLine, LinkedList<String> parameterLine, LinkedList<String> inputLine, LinkedList<String> outputLine ) {
		return ArgumentLineParser.parse(getCommandPart(getBinary()), getFormatPart(getBinary()), maskLine, parameterLine, inputLine, outputLine);
	}

	/**
	 * Builds the command line as specified in the command format
	 * and with the arguments. The binary name of the program
	 * is returns with the absolute pathname.
	 *  %p for parameter; %p0 for the first parameter
	 *  %m for the mask
	 *  %i for input; %i0 for the first input image
	 *  %o for output; %o0 for the first output image.
	 *  
	 * @param maskLine the mask value.
	 * @param parameterLine the parameter values.
	 * @param outputLine the output values.
	 * @return the command line.
	 */
	public final LinkedList<String> buildCommandLine( LinkedList<String> maskLine, LinkedList<String> parameterLine, LinkedList<String> inputLine, LinkedList<String> outputLine ) {
		return ArgumentLineParser.parse(GraphItem.getAbsoluteBinary(getCommandPart(getBinary())), getFormatPart(getBinary()), maskLine, parameterLine, inputLine, outputLine);
	}

	/**
	 * Executes the related command. The execution is done
	 * only if the current state is "unexecuted" or "error".
	 * Returns true if the execution is done without fatal error,
	 * where fatal error is in case of:
	 * <ul>
	 * <li>imperative: any errors.
	 * <li>non imperative: only system or parameter unset errors.
	 * </ul>
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			String temp;
			LinkedList<String> maskLine = new LinkedList<String>();
			LinkedList<String> parameterLine = new LinkedList<String>();
			LinkedList<String> inputLine = new LinkedList<String>();
			LinkedList<String> outputLine = new LinkedList<String>();

			// Get the parameter values
			for (InSlotParameter slot : inSlotParameters) {
				if (slot.hasValue(imperative)) {
					temp = slot.getValue();
					parameterLine.add(temp);
				} else {
					throw new MissingParameterException(Ariane.lang.getString("Operator.error1")+slot.getRank());
				}
			}

			// Get the optional mask
			if (inSlotMask.isConnected() && inSlotMask.hasValue(false)) {
				temp = inSlotMask.getValue();
				maskLine.add(temp);
			}

			// Get the input images
			for (InSlotImage slot: inSlotImages) {
				if (slot.hasValue(imperative)) {
					temp = slot.getValue();
					inputLine.add(temp);
				} else {
					throw new MissingInputException(Ariane.lang.getString("Operator.error3")+slot.getRank());
				}
			}

			// Create outputs images as temporary files
			for (OutSlotImage slot : outSlotImages) {
				temp = slot.getTempName();
				outputLine.add(temp);
			}

			LinkedList<String> commandLine = buildCommandLine(maskLine, parameterLine, inputLine, outputLine);
			String consoleMessage=getName();
			for (int i = 1; i< commandLine.size(); i++ ) {
				consoleMessage+=" "+getRelativeFilePathName(commandLine.get(i));
			}
			Ariane.workspace.printSystemMessage(consoleMessage);
			long start = System.currentTimeMillis();

			// Execution of the operator binary
			changeStatus(Status.RUNNING);
			OperatorProcessBuilder pb = new OperatorProcessBuilder(commandLine);
			Process proc1 = pb.start();
			BufferedReader br = new BufferedReader(new InputStreamReader(proc1.getInputStream()));
			String line;
			while ((line = br.readLine()) != null) {
				Ariane.workspace.printOutputMessage(line);
			}
			int exitValue=proc1.waitFor();

			String resultString="";
			if (exitValue==0) {
				// Get the status value
				if (getCommandStatus().trim().length()>0) {
					pb = new OperatorProcessBuilder(GraphItem.getAbsoluteBinary(getCommandStatus()));
					Process proc2 = pb.start();
					InputStream resultStream = proc2.getInputStream();
					proc2.waitFor();

					// Exploit the results
					byte resultByte[] = new byte[resultStream.available()+1];
					resultStream.read(resultByte);
					resultString = new String(resultByte);
					getOutSlotResult().setValue(resultString.trim());
				} else {
					getOutSlotResult().setValue("SUCCESS");
				}
			}
			
			// I mean if an execution fails than the error code !=0. 
			if (exitValue != 0 ) { // '|| resultString.startsWith("FAILURE")) {
				throw new RuntimeException(Ariane.lang.getString("Operator.error2"));
			}
			// Everything is OK -> validate the result of all outputs.
			for (OutSlotImage slot : outSlotImages) {
				slot.setValue();
			}
			long end = System.currentTimeMillis();
			Ariane.workspace.printSystemMessage("("+(end-start)/1000.0f+"s)");
		} catch (MissingParameterException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.MISSINGPARAMETER, getAbsoluteName() + "("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (MissingInputException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.MISSINGPARAMETER, getAbsoluteName() + "("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (RuntimeException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.RUNTIME, getAbsoluteName() + "("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (IOException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.INPUTOUTPUT, getAbsoluteName() + "("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (InterruptedException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.INTERRUPTEDEXECUTION, getAbsoluteName() + "("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}
		changeStatus(Status.EXECUTED);
		return true;
	}

	public static class MissingParameterException extends IllegalArgumentException {
		public MissingParameterException(String s) {
			super(s);
		}
	}

	public static class MissingInputException extends IllegalArgumentException {
		public MissingInputException(String s) {
			super(s);
		}
	}
}