/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.util.LinkedList;

import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;

/**
 * A <code>ScriptItem</code> implements a long text value.
 * 
 * @author Régis CLOUARD - Aug 24, 2012
 */
public class ScriptItem extends ControlStructure implements FocusListener {
	/** The tag name of the item. */
	public static String TAGNAME="Script";

	/** The text to be displayed. */
	private JTextArea _scriptText;

	/**
	 * Creates a <code>ConstantItem</code> with the given
	 * attributes.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the item.
	 * @param identifier the identification number (or -1).
	 */
	public ScriptItem( TreeItem treeItem, Graph graph, String name, int identifier ) {
		super(treeItem, graph, identifier,name);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
		setComment(getFormula());
		draw();
	}

	/**
	 * Creates a <code>ConstantItem</code> on the given
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param elements the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public ScriptItem( Graph graph, LinkedList<XMLElement> elements ) throws IOException {
		super(graph, -1);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
		load(elements);
		setComment(getFormula());
		draw();
	}

	/**
	 * Builds or changes the prototype of the control structure
	 * from the given values.
	 * Note that the number of inputs, outputs, and parameters are always 0.
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected final void setPrototype( String name, int iIm, int iP, int oIm ) {
		setAbstractName(name);
	}

	@Override
	public boolean displayFormula() {
		return false;
	}

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * @return the tag name.
	 */
	@Override
	public final String tagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes the <code>Constant</code>, namely
	 * returns the expression.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			if ( (getFormula()==null) || (getFormula().length()==0) )
				throw new ArithmeticException(Ariane.lang.getString("arithmetic.error19"));
			getOutSlotResult().setValue(getFormula());
			changeStatus(Status.EXECUTED);
			return true;
		} catch (ArithmeticException e){
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}catch(Exception e){
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}
	}

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected final JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;

		// Action: executer
		item = new JMenuItem(Ariane.lang.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) {
				getGraph().execute();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) {
				resetAll();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		// Edition: delete
		item = new JMenuItem(Ariane.lang.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e){
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		return popupMenu;
	}
	
	/**
	 * A hook for customizing the comment editor.
	 * @return a Jlabel.
	 */
	@Override
	protected JComponent getCommentEditor() {
		// Override the class in order to refresh
		// the graph when the text area is modified.
		_scriptText=new JTextArea(){
			@Override
			public void repaint(){
				getGraph().repaint();
			}
		};
		_scriptText.setFont(PreferenceChooser.defaultFont);
		_scriptText.setForeground(new Color(0,0,0));
		_scriptText.setTabSize(3);

		// Mouse events.
		_scriptText.addMouseListener(new MouseAdapter( ) {
			@Override
			public void mouseClicked( MouseEvent e ) {
			}
			@Override
			public void mousePressed( MouseEvent e ) {
				ScriptItem.this.onMousePressed(e);
			}
			@Override
			public void mouseReleased( MouseEvent e ) {
				if (isSelected()) {
					_scriptText.setEditable(true);
					_scriptText.getCaret().setVisible(true);
				} else {
					ScriptItem.this.onMouseReleased(e);
				}
			}
		});

		_scriptText.addFocusListener(this);
		_scriptText.setEditable(false);
		_scriptText.getCaret().setVisible(false);

		return _scriptText;
	}

	/**
	 * Invoked when a JtextArea looses the keyboard focus.
	 * Change the text of the related element in the model.
	 * 
	 * @param e the FocusEvent.
	 */
	@Override
	public final void focusLost( FocusEvent e ) {
		if (e.getSource()==_scriptText) {
			String newFormula=_scriptText.getText();
			// Change formula if it was changed!
			if (getFormula()==null || !getFormula().equals(newFormula)) {
				setFormula(_scriptText.getText());
				getGraph().setModified(true);
			}
		}
		_scriptText.setEditable(false);
		_scriptText.getCaret().setVisible(false);
		_scriptText.setCaretPosition(0);
	}

	/**
	 * Invoked when a component gains the keyboard focus.
	 * 
	 * @param e the FocusEvent.
	 */
	@Override
	public final void focusGained( FocusEvent e ) {
		displayPropertyViewer();
		if (!isSelected()) {
			GraphItem g = ScriptItem.this;
			g.getGraph().addItemInSelection(g,false);
		} else {
			_scriptText.setCaretPosition(0);
			_scriptText.moveCaretPosition(_scriptText.getText().length());
		}
	}
}