/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.OperatorItem.MissingInputException;
import fr.greyc.pantheon.ariane.graph.item.OperatorItem.MissingParameterException;
import fr.greyc.pantheon.ariane.graph.item.etc.ArgumentLineParser;
import fr.greyc.pantheon.ariane.graph.item.etc.OperatorProcessBuilder;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.LinkedList;

import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

/**
 * An <code>SystemItem</code> is a graph item that
 * executes an operating system command. It stores the command line
 * and gets its parameters, inputs and outputs from slots.
 * @see GraphItem
 * 
 * @author Régis CLOUARD - Aug 25, 2012
 */
public final class SystemItem extends ControlStructure {
	/** The tag name of the item. */
	public static String TAGNAME="System";

	/**
	 * Creates a new <code>CommandItem</code> on the specified
	 * graph with the given attribute values.
	 * 
	 * @param treeItem the source tree item.
	 * @param graph the graph that holds the item.
	 * @param name the default name of the item.
	 * @param identifier the identification number (or -1).
	 */
	public SystemItem( TreeItem treeItem, Graph graph, String name, int identifier ) {
		super(treeItem, graph, identifier,name);
		askProperties(null,1,1,1);
		setBinary("");
		draw();
	}

	/**
	 * Creates a <code>CommandItem</code> on the specified
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param elements the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public SystemItem( Graph graph, LinkedList<XMLElement> elements ) throws IOException {
		super(graph, -1);
		load(elements);
		draw();
	}

	/**
	 * Builds the prototype of the command item from the given values.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected final void setPrototype( String name, int iIm, int iP, int oIm ) {
		setAbstractName(name);
		for (int i=0; i < iP; i++) { inSlotParameters.add(new InSlotParameter(this,i)); }
		for (int i=0; i<iIm; i++) { inSlotImages.add(new InSlotImage(this,i)); }
		for (int i=0; i<oIm; i++) { outSlotImages.add(new OutSlotImage(this,i)); }
		draw();
	}

	///// 
	///// P R E F E R E N C E S
	///// 

	// Note: theses methods cannot be generalized in the GraphItem class
	// because they use static attributes.

	/** The color of the item. */
	private static final Color _defaultColor=new Color(0,153,255);

	/** The color of the item. */
	private static Color _color=_defaultColor;

	/**
	 * Changes the current color.
	 * 
	 * @param color the new color.
	 */
	static public void setColor( Color color ) { _color = color; }

	/**
	 * Returns the current color.
	 * 
	 * @return the color.
	 */
	static public Color getColor( ) { return _color; }

	/**
	 * Returns the default color.
	 * 
	 * @return the color.
	 */
	static public Color getDefaultColor( ) { return _defaultColor; }

	/**
	 * Returns the background color (same as geColor()).
	 * 
	 * @return the color.
	 */
	@Override
	public Color getBackgroundColor( ) { return _color; }

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * @return the tag name.
	 */
	@Override
	public final String tagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Builds the command line as specified in the command format
	 * and with the arguments. The binary name of the program
	 * is returns with the absolute pathname.
	 * Syntax: format string uses:
	 *  %p for parameter; %p0 for the first parameter
	 *  %m for the mask
	 *  %i for input; %i0 for the first input image
	 *  %o for output; %o0 for the first output image.
	 *  
	 * @param maskLine the mask value.
	 * @param parameterLine the parameter values.
	 * @param outputLine the output values.
	 * @return the command line.
	 */
	public LinkedList<String> buildAbstractCommandLine( LinkedList<String> maskLine, LinkedList<String> parameterLine, LinkedList<String> inputLine, LinkedList<String> outputLine ) {
		return ArgumentLineParser.parse("", getBinary(), maskLine, parameterLine, inputLine, outputLine);
	}

	/**
	 * Executes the <code>ScriptItem</code>, means
	 * compute a result from the arithmetic formula.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			String temp;
			LinkedList<String> parameterLine = new LinkedList<String>();
			LinkedList<String> inputLine = new LinkedList<String>();
			LinkedList<String> outputLine = new LinkedList<String>();

			// Get the parameter values
			for (InSlotParameter slot : inSlotParameters) {
				if (slot.hasValue(imperative)) {
					temp = slot.getValue();
					parameterLine.add(temp);
				} else {
					throw new OperatorItem.MissingParameterException(Ariane.lang.getString("Operator.error1")+slot.getRank());
				}
			}

			// Get the input images
			for (InSlotImage slot: inSlotImages) {
				if (slot.hasValue(imperative)) {
					temp = slot.getValue();
					inputLine.add(temp);
				} else {
					throw new OperatorItem.MissingInputException(Ariane.lang.getString("Operator.error3")+slot.getRank());
				}
			}

			// Create outputs images as temporary files
			for (OutSlotImage slot : outSlotImages) {
				temp = slot.getTempName();
				outputLine.add(temp);
			}
			LinkedList<String> maskLine = new LinkedList<String>();

			LinkedList<String> commandLine = ArgumentLineParser.parse("", getBinary(), maskLine, parameterLine, inputLine, outputLine );
			String consoleMessage="";
			for (int i = 0; i< commandLine.size(); i++ ) {
				consoleMessage+=" "+commandLine.get(i);
			}
			Ariane.workspace.printSystemMessage(consoleMessage);

			// Execution of the operator binary
			changeStatus(Status.RUNNING);
			OperatorProcessBuilder pb = new OperatorProcessBuilder(commandLine);
			Process proc1 = pb.start();
			BufferedReader br = new BufferedReader(new InputStreamReader(proc1.getInputStream()));
			String line;
			while ((line = br.readLine()) != null) {
				Ariane.workspace.printOutputMessage(line);
			}
			int exitValue=proc1.waitFor();

			String resultString="";
			if (exitValue==0) {
				// Get the status value
				if (getCommandStatus().trim().length()>0) {
					pb = new OperatorProcessBuilder(GraphItem.getAbsoluteBinary(getCommandStatus()));
					Process proc2 = pb.start();
					InputStream resultStream = proc2.getInputStream();
					proc2.waitFor();

					// Exploit the results
					byte resultByte[] = new byte[resultStream.available()+1];
					resultStream.read(resultByte);
					resultString = new String(resultByte);
					getOutSlotResult().setValue(resultString.trim());
				} else {
					getOutSlotResult().setValue("SUCCESS");
				}
			}
			// I mean if an execution fails than the error code !=0. 
			if (exitValue != 0 ) { // '|| resultString.startsWith("FAILURE")) {
				throw new RuntimeException(Ariane.lang.getString("Operator.error2"));
			}
			// Everything is OK -> validate the result of all outputs.
			for (OutSlotImage slot : outSlotImages) {
				slot.setValue();
			}
		} catch (MissingInterpreterNameException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.MISSINGPARAMETER, getAbsoluteName() + "("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (MissingParameterException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.MISSINGPARAMETER, getAbsoluteName() + "("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (MissingInputException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.MISSINGPARAMETER, getAbsoluteName() + "("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (RuntimeException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.RUNTIME, getAbsoluteName() + "("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (IOException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.INPUTOUTPUT, getAbsoluteName() + "("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (InterruptedException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.INTERRUPTEDEXECUTION, getAbsoluteName() + "("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}
		changeStatus(Status.EXECUTED);
		return true;
	}

	/**
	 * Thrown to indicate that an input data file is missing.
	 */
	public static class MissingInterpreterNameException extends RuntimeException {
		/**
		 * Creates a new exception with the specified diagnostic message.
		 * 
		 * @param s the detail message.
		 */
		public MissingInterpreterNameException( String s ) {
			super(s);
		}
	}

	///// 
	///// P R O P E R T I E S
	/////

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected final JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;

		// Action: executer
		item = new JMenuItem(Ariane.lang.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				getGraph().execute();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				resetAll();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		// Edition: delete
		item = new JMenuItem(Ariane.lang.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e){
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		return popupMenu;
	}

	/**
	 * @return a component appropriate to edit and possibly change the name of the item.
	 */
	@Override
	public JComponent getBinaryEditor() {
		if (getBinary() != null) {
			JTextField c = new JTextField(getBinary());
			c.addFocusListener(new FocusListener() {
				@Override
				public final void focusLost(FocusEvent e) {
					changeBinary((JTextField)e.getSource(), ((JTextField)e.getSource()).getText(), true);
				}
				@Override
				public final void focusGained(FocusEvent e){ }
			});
			c.setEnabled(true);
			c.setEditable(true);
			return c;
		}
		return null;
	}
}