/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.command.ChangeViewerCommandEdit;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

/**
 * A <code>ViewerItem</code> manages displaying
 * of a data file in a graph.
 * 
 * @version Régis CLOUARD - Jun 19, 2010
 */
public final class ViewerItem extends DataFile {
	/** The tag name of the item. */
	public static final String TAGNAME = "Viewer";
	/** The textfield that display the viewer command */
	private JTextField _commandTextfield = new JTextField(10);

	/**
	 * Creates a <code>ViewerItem</code> on the specified graph
	 * with the specified properties.
	 * 
	 * @param treeItem the source tree item.
	 * @param graph the graph that holds the item.
	 * @param name the name of the item.
	 * @param identifier the identification number of the item (or -1).
	 * @param category the category name of the operator.
	 */
	public ViewerItem( TreeItem treeItem, Graph graph, String name, int identifier, String category ) {
		super(treeItem, graph, name, identifier, null, category, "viewer");
		inSlotImages.add(new InSlotImage(this,0));
		outSlotImages.add(new OutSlotImage(this,0));
		draw();
	}

	/**
	 * Creates a <code>ViewerItem</code> on the specified graph
	 * with the given XML data.
	 * 
	 * @param graph the graph that holds the item.
	 * @param xmlTree an XML tree of data.
	 * @throws IOException XML errors.
	 */
	public ViewerItem( Graph graph,LinkedList<XMLElement> xmlTree ) throws IOException {
		super(graph, xmlTree, "viewer");
		draw();
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public final String tagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * @return the absolute name of the viewer binary.
	 */
	@Override
	public final String getViewerCommand(){
		if (getBinary() == null)
			return super.getViewerCommand();
		else
			return getBinary();
	}

	/**
	 * @return the name if the viewer binary.
	 */
	public final String getViewerCommandName() {
		File file = new File(getViewerCommand());
		return file.getName();
	}

	/**
	 * Executes a Viewer. It means: sets the input file into
	 * the related physical file.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			_currentFilename = null;
			InSlotImage isi = (inSlotImages.getFirst());
			if (isi.hasValue(imperative)) {
				_currentFilename=isi.getValue();
				(outSlotImages.getFirst()).setValue(_currentFilename);
			} else
				throw new MissingInputException(Ariane.ErrorCode.MISSINGINPUT, Ariane.lang.getString("ImageFile.error5"));

			if (inSlotImages.size() > 0)
				displayImage();
		} catch (MissingInputException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(e.errorCode, getAbsoluteName() +"("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (Exception e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}

		// Validate the result.
		changeStatus(Status.EXECUTED);
		return true;
	}

	///// 
	///// M E N U 
	/////

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected final JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item= new JMenuItem(Ariane.lang.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) { getGraph().execute();}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) { resetAll(); }
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) { delete(); }
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);
		return popupMenu;
	}

	/**
	 * Change the viewer command.
	 * @param newCommand the new command name.
	 * 	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void changeCommandName( String newCommand, boolean historical ) {
		String oldCommand = getBinary();
		if (newCommand!=null && newCommand.length()>0 && !newCommand.equals(getCommandVisu())) {
			setBinary(newCommand);
		} else {
			setBinary(null);
		}
		_commandTextfield.setText((getBinary()!=null)?getBinary():getCommandVisu());
		updatePropertyViewer();
		getGraph().setModified(true);
		resetAll();
		if (historical) {
			getGraph().startActionEdition();
			getGraph().validateAction(new ChangeViewerCommandEdit(this, newCommand, oldCommand));
			getGraph().endActionEdition();
		}
	}

	/**
	 * @return the list of specific properties (other than common properties).
	 */
	@Override
	public final ArrayList<PropertyViewer.PropertyCouple> getSpecificProperties( ) {
		ArrayList<PropertyViewer.PropertyCouple> properties = new ArrayList<PropertyViewer.PropertyCouple>();
		_commandTextfield.setText((getBinary()!=null)?getBinary():getCommandVisu());
		_commandTextfield.addFocusListener(new FocusListener() {
			@Override
			public void focusLost(FocusEvent e) {
				changeCommandName(((JTextField)(e.getSource())).getText().trim(),true);
			}
			@Override
			public void focusGained(FocusEvent e){ }
		});
		properties.add(new PropertyViewer.PropertyCouple(Ariane.lang.getString("PropertyViewer.commandName"), _commandTextfield));
		return properties;
	}
	
	/**
	 * Invoked when the mouse button has been
	 * double clicked (pressed and released) on a graph item.
	 * Shows the dialog box for setting the item's properties.
	 * 
	 * @param e the event attributes.
	 */
	@Override
	protected final void onMouseDoubleClicked( MouseEvent e ) {
		if (getExecutionStatus() != Status.EXECUTED) return;
		try {
			displayImage();
		} catch (Exception ex) {
			Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+"): "+ ex.getLocalizedMessage());
		}
	}
}