/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import java.awt.Color;
import java.io.File;
import java.io.IOException;
import java.util.LinkedList;

import javax.swing.JPopupMenu;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotResult;
import fr.greyc.pantheon.ariane.graph.slot.Slot;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>VirtualItem</code> is used to reference
 * inputs and outputs of a <code>SubGraph</code>.
 * It is invisible as it has no graphical representation.
 *
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Regis CLOUARD - Jan 07, 2007
 */
public class VirtualItem extends GraphItem {
	/** Number of input images. */
	private final int inImages;
	/** Number of output images. */
	private final int outImages;
	/** Number of intermediates input images. (for reentry) */
	private final int interImages;
	/** Number of input parameters. */
	private final int inParams;
	/** Number of intermediates input parameters (for reentry) */
	private final int interParams;

	/**
	 * Creates a VirtualOperator with the given attributes.
	 * Initializes the mouse listeners, the popup menu
	 * and displays the content in the graph.
	 * 
	 * @param graph the parent graph.
	 * @param name the name of the macro-operator.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 * @param iiIm the number of intermediate input images.
	 * @param iiP the number of intermediate input parameters.
	 */
	public VirtualItem( SubGraph graph, String name, int iIm, int iP, int oIm, int iiIm, int iiP ) {
		super(null,graph, 0,name,null);       // IMPERATIVE = 0.
		inImages = iIm;
		outImages = oIm;
		inParams = iP;
		interImages = iiIm;
		interParams = iiP;
		for (int i = 0; i < inImages; i++) { 
			OutSlotImage o = new OutSlotImage(this, i);
			o.setToolTipText(graph.getSuperGraphItem().getNickName()+Ariane.lang.getString("VirtualOperator.input")+i);
			outSlotImages.add(o);
		}
		{
			OutSlotImage o = new OutSlotImage(this, inImages);
			o.setToolTipText(graph.getSuperGraphItem().getNickName()+Ariane.lang.getString("VirtualOperator.mask"));
			outSlotImages.add(o);
		}
		for (int i = 0; i < outImages; i++) {
			InSlotImage o = new InSlotImage(this,i);
			o.setToolTipText(graph.getSuperGraphItem().getNickName()+Ariane.lang.getString("VirtualOperator.output")+i);
			inSlotImages.add(o);
		}
		for (int i = outImages; i < outImages +interImages; i++) {
			InSlotImage o = new InSlotImage(this,i);
			o.setToolTipText(graph.getSuperGraphItem().getNickName()+Ariane.lang.getString("VirtualOperator.reinput")+(i-outImages));
			inSlotImages.add(o);
		}
		{
			InSlotParameter o = new InSlotParameter(this,0);
			o.setToolTipText(graph.getSuperGraphItem().getNickName()+Ariane.lang.getString("VirtualOperator.result"));
			inSlotParameters.add(o);
		}
		for (int i = 1; i < 1+interParams; i++) {
			InSlotParameter o = new InSlotParameter(this,i);
			o.setToolTipText(graph.getSuperGraphItem().getNickName()+Ariane.lang.getString("VirtualOperator.reparameter")+(i+(inParams-interParams-1)));
			inSlotParameters.add(o);
		}
		for (int i = 0; i < inParams; i++) {
			OutSlotResult o = new OutSlotResult(this,i);
			o.setToolTipText(graph.getSuperGraphItem().getNickName()+Ariane.lang.getString("VirtualOperator.parameter")+i);
			outSlotResults.add(o);
		}
		setFocusable(false);
	}

	/**
	 * Returns the identifier number of a VirtualOperator
	 * which is always 0.
	 * 
	 * @return the identifier number.
	 */
	@Override
	public final int getId( ){
		return 0;
	}

	/**
	 * Returns the list of inputs images.
	 * 
	 * @return the LinkedList of images.
	 */
	public final LinkedList<OutSlotImage> getVirtualInSlotImages( ) {
		return outSlotImages;
	}

	/**
	 * Returns the nth image in the inputs images list.
	 * 
	 * @param i the index number of the image.
	 * @return the image.
	 */
	public final OutSlotImage getVirtualInSlotImage( int i ) {
		return outSlotImages.get(i);
	}

	/**
	 * Returns the list of output images.
	 * 
	 * @return the LinkedList of images.
	 */
	public final LinkedList<InSlotImage> getVirtualOutSlotImages( ) {
		return inSlotImages;
	}

	/**
	 * Returns the nth output image slot.
	 * 
	 * @param i the index number of the image.
	 * @return the out slot image.
	 */
	public final InSlotImage getVirtualOutSlotImage( int i ) {
		return inSlotImages.get(i);
	}

	/**
	 * Returns the list of all input parameters.
	 * 
	 * @return the linked list of input images.
	 */
	public final LinkedList<OutSlotResult> getVirtualInSlotParameters( ) {
		return outSlotResults;
	}

	/**
	 * Returns the nth input parameter.
	 * 
	 * @param i the index number of the parameter.
	 * @return the parameter slot.
	 */
	public final OutSlotResult getVirtualInSlotParameter( int i ) {
		return outSlotResults.get(i);
	}

	/**
	 * Returns the nth intermediate parameter.
	 * => InSlotParameter +1.
	 * 
	 * @param i the index number of the result.
	 * @return the parameter slot.
	 */
	public final InSlotParameter getVirtualInterSlotResult( int i ) {
		return inSlotParameters.get(1+i);
	}

	/**
	 * Return the image mask slot.
	 * 
	 * @return the mask slot.
	 */
	public final Slot getVirtualInSlotMask( ) {
		return outSlotImages.getLast();
	}

	/**
	 * Returns the result slot.
	 * 
	 * @return the result slot.
	 */
	public final Slot getVirtualOutSlotResult( ) {
		return inSlotParameters.getFirst();
	}

	/**
	 * Returns the nth image in the intermediate input images list.
	 * 
	 * @param i the index number of the image.
	 * @return the image slot.
	 */
	public final InSlotImage getVirtualInterSlotImage( int i ) {
		return inSlotImages.get(i+outImages);
	}

	/**
	 * Resets the item and all its forwarded connected items
	 * and its parent.
	 */
	@Override
	public final void resetAll( ) {
		if (getExecutionStatus() != Status.UNEXECUTED) {

			// Overrides the graphitem reset in order to
			// avoid resetting the inter images and inter parameters
			// during a loop since a loop need to reset the
			// contents of the subgraph for each iterations.
			reset();
			if (outSlotImages != null){
				for (OutSlotImage slot: outSlotImages) {
					slot.reset();
				}
			}
			if (outSlotResults.size() > 0){
				getOutSlotResult().reset();
			}
			// Its parent
			if (getGraph().getSuperGraphItem()!=null) {
				getGraph().getSuperGraphItem().resetAll();
			}
		}
	}

	/**
	 * Executes the related subgraph and returns the
	 * status of the execution (always true).
	 * 
	 * @param imperative is FAILURE is considered as an error?
	 * @return true if the virtual operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		changeStatus(Status.EXECUTED);
		return true;
	}

	/**
	 * Saves a VirtualOperator as an XML tree.
	 * Does nothing!
	 * 
	 * @param file the current filename of the project.
	 * @return the XML tree containing the subgraph.
	 * @throws IOException XML errors.
	 */
	@Override
	public final XMLElement save( File file) throws IOException {
		return null;
	}

	/**
	 * Never paint this item.
	 */
	@Override
	public final void draw( ) { }

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public final String tagName( ) {
		return null;
	}

	/**
	 * Returns the background color.
	 */
	@Override
	public final Color getBackgroundColor( ) {
		return Graph.getBackgroundColor();
	}

	/**
	 * Creates a related property viewer.
	 * 
	 * @return a dialog box.
	 */
	@Override
	protected final PropertyViewer createPropertyViewer() {
		return BLANK_PROPERTY_PANEL;
	}

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected final JPopupMenu initPopup( ) { return null; }
}