/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item.etc;

import fr.greyc.pantheon.ariane.Ariane;

/**
 * A <code>ArithmeticExpression</code> is a calculus formula that includes
 * arithmetic operators, constant values and parameters values ($0, $1, etc.)
 *
 * @author Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Jan 23, 2007 (add: modulo operator)
 * @version Frédéric LAIR - Jun 28, 2007 (add: more than two parameters)
 */
public class ArithmeticExpression {
	/** The rough formula. */
	private String _initialExpression;

	/** The coded formula. */
	private byte[] _codedExpression;

	/** The stack of operations. */
	private double[] _stack;

	/** The table of constants used in the formula. */
	private double[] _constantStack;

	/** The codes of known arithmetic operations. */
	private static final byte PLUS = - 1 ;
	/** The codes of known arithmetic operations. */
	private static final byte MINUS = - 2 ;
	/** The codes of known arithmetic operations. */
	private static final byte TIMES = - 3 ;
	/** The codes of known arithmetic operations. */
	private static final byte DIVIDE = - 4 ;
	/** The codes of known arithmetic operations. */
	private static final byte MODULO = - 5 ;
	/** The codes of known arithmetic operations. */
	private static final byte POWER = - 6 ;
	/** The codes of known arithmetic operations. */
	private static final byte MIN = - 11 ;
	/** The codes of known arithmetic operations. */
	private static final byte MAX = - 12 ;
	/** The codes of known arithmetic operations. */
	private static final byte SIN = - 21 ;
	/** The codes of known arithmetic operations. */
	private static final byte COS = - 22 ;
	/** The codes of known arithmetic operations. */
	private static final byte TAN = - 23 ;
	/** The codes of known arithmetic operations. */
	private static final byte ASIN = - 24 ;
	/** The codes of known arithmetic operations. */
	private static final byte ACOS = - 25 ;
	/** The codes of known arithmetic operations. */
	private static final byte ATAN = - 26 ;
	/** The codes of known arithmetic operations. */
	private static final byte EXP = - 27 ;
	/** The codes of known arithmetic operations. */
	private static final byte LN = - 28 ;
	/** The codes of known arithmetic operations. */
	private static final byte LOG10 = - 29 ;
	/** The codes of known arithmetic operations. */
	private static final byte LOG2 = - 30 ;
	/** The codes of known arithmetic operations. */
	private static final byte ABS = - 31 ;
	/** The codes of known arithmetic operations. */
	private static final byte SQRT = - 32 ;
	/** The codes of known arithmetic operations. */
	private static final byte UNARYMINUS = - 33 ;
	// Warning: -34 -> infinite is used for variables. 
	/** The code for all variables: first -34, then -35, etc. */
	private byte[] VARIABLES;

	/** The number of used parameter. */
	private int _nbParam;

	/** The codes of known unary functions. */
	private String[] _functionNamesUnary = {
			"sin", "cos", "tan", "asin",
			"acos", "atan", "exp",
			"ln", "log10", "log2", "abs", "sqrt" };

	/** The codes of known binary functions. */
	private String[] _functionNamesBinary = { "min","max" };

	/* Internal variables. */
	/** */
	private int _pos;
	/** */
	private int _constantStackIndex;
	/** */
	private int _codeSize;

	/**
	 * Creates an arithmetic expression from a formula.
	 * In case of error, generates an exception.
	 * 
	 * @param formula the rough formula.
	 * @param nbParam the number of parameters.
	 */
	public ArithmeticExpression( String formula, int nbParam ) {
		// Empty formula
		if (formula == null || formula.trim().length()==0)
			return;

		_nbParam = nbParam;
		VARIABLES = new byte[_nbParam];
		for(int i =0; i < _nbParam; i++) {
			VARIABLES[i] = (byte) (- 34 -i);
		}
		_initialExpression = formula;
		_codedExpression = new byte[formula.length()];
		_constantStack = new double[formula.length()];

		// Convert formula to computable code
		_pos = 0;
		_constantStackIndex = 0;
		_codeSize = 0;
		parseFormula();
		skip();
		if (next() != 0)
			error(Ariane.lang.getString("arithmetic.error1"));
		int stackSize = computeStackUsage();
		_stack = new double[stackSize];
		byte[] c = new byte[_codeSize];
		System.arraycopy(_codedExpression,0,c,0,_codeSize);
		_codedExpression = c;
		double[] A = new double[_constantStackIndex];
		System.arraycopy(_constantStack,0,A,0,_constantStackIndex);
		_constantStack = A;
	}

	/**
	 * Evaluates the current expression from the specified
	 * values of the two parameters.
	 * 
	 * @param val the list of value.
	 * @return the result.
	 * @throws ArithmeticException in case of malformed expression.
	 */
	public double evaluate( double[] val ) throws ArithmeticException {
		try {
			int top = 0;
			boolean flag = false; // used for while condition
			int var = 0;
			for (int i = 0; i < _codeSize; i++) {
				var = 0;
				flag = false;
				if (_codedExpression[i] >= 0){ // constants
					_stack[top++] = _constantStack[_codedExpression[i]];
				} else if (_codedExpression[i] >= POWER) { // binary operators
					double y = _stack[--top];
					double x = _stack[--top];
					double ans = Double.NaN;
					switch (_codedExpression[i]) {
					case PLUS: ans = x + y;  break;
					case MINUS: ans = x - y;  break;
					case TIMES: ans = x * y;  break;
					case DIVIDE: ans = x / y;  break;
					case MODULO: ans = x % y;  break;
					case POWER: ans = Math.pow(x,y);  break;
					}
					if (Double.isNaN(ans))
						return ans;
					_stack[top++] = ans;
				} else {
					// parameters
					while ((flag == false) && (var < _nbParam)){
						if (_codedExpression[i] == VARIABLES[var]){
							flag = true;
							_stack[top++] = val[var];
						}
						var++;
					}
					// otherwise
					if (!flag) {
						double y,x = _stack[--top];
						double ans = Double.NaN;
						switch (_codedExpression[i]) {
						case MIN: y= _stack[--top]; ans = (x<y)? x:y; break;
						case MAX: y= _stack[--top]; ans = (x>y)? x:y;  break;
						case SIN: ans = Math.sin(x);  break;
						case COS: ans = Math.cos(x);  break;
						case TAN: ans = Math.tan(x);  break;
						case ASIN: if (Math.abs(x) <= 1.0) ans = Math.asin(x);  break;
						case ACOS: if (Math.abs(x) <= 1.0) ans = Math.acos(x);  break;
						case ATAN: ans = Math.atan(x);  break;
						case EXP: ans = Math.exp(x);  break;
						case LN: if (x > 0.0) ans = Math.log(x);  break;
						case LOG2: if (x > 0.0) ans = Math.log(x)/Math.log(2);  break;
						case LOG10: if (x > 0.0) ans = Math.log(x)/Math.log(10);  break;
						case ABS: ans = Math.abs(x);  break;
						case SQRT: if (x >= 0.0) ans = Math.sqrt(x);  break;
						case UNARYMINUS: ans = -x; break;
						}
						if (Double.isNaN(ans))
							return ans;
						_stack[top++] = ans;
					}
				}
			}
		} catch (Exception e) {
//			e.printStackTrace();
			throw new ArithmeticException(e.getLocalizedMessage());
		}
		if (Double.isInfinite(_stack[0]))
			throw new ArithmeticException(Ariane.lang.getString("arithmetic.error17"));
		else
			return _stack[0];
	}

	/**
	 * Generates a exception error with the specified message.
	 * 
	 * @param message the message to be displayed.
	 */
	private void error( String message ) {
		throw new IllegalArgumentException(Ariane.lang.getString("arithmetic.error0.1") +" "+ message+" " + Ariane.lang.getString("arithmetic.error0.2") +" "+ _pos + ")");
	}

	/**
	 * Calculates the maximum size of the stack.
	 * 
	 * @return the size.
	 */
	private int computeStackUsage( ) {
		int s = 0;
		int max = 0;
		for (int i = 0; i < _codeSize; i++) {
			if ( (_codedExpression[i] >= 0) || (isVariable(i))) {
				s++;
				if (s > max)
					max = s;
			} else if (_codedExpression[i] >= POWER)
				s--;
		}
		return max;
	}

	/**
	 * tells if the component i from the coded expression is a parameter or not.
	 * 
	 * @param i the rank of the item we want to check.
	 * @return true if it is a parameter, false if not.
	 */
	private boolean isVariable(int i){
		return ((_codedExpression[i] <= -34) && (_codedExpression[i] > -34 - _nbParam));
	}

	/**
	 * @return the next current character in the formula.
	 */
	private char next( ) {
		return (_pos >= _initialExpression.length())? 0 : _initialExpression.charAt(_pos);
	}

	/**
	 * Skips white spaces.
	 */
	private void skip( ) {
		while(Character.isWhitespace(next()))
			_pos++;
	}

	/**
	 * Parses a mathematical expression such as:
	 * [+|-]Term[+|-]Term
	 */
	private void parseFormula( ) {
		boolean negative = false;
		skip();
		// Signe plus ou moins au debut de l'expression
		if (next() == '+' || next() == '-') {
			negative = (next() == '-');
			_pos++;
			skip();
		}
		parseTerm();
		if (negative)
			_codedExpression[_codeSize++] = UNARYMINUS;
		skip();
		while (next() == '+' || next() == '-') {
			char op = next();
			_pos++;
			parseTerm();
			_codedExpression[_codeSize++] = (op == '+')? PLUS : MINUS;
			skip();
		}
	}

	/**
	 * Parses an expression such as :
	 * Primary [*|/|^] Primary.
	 */
	private void parseTerm( ) {
		parsePrimary();
		skip();
		while (next() == '*' || next() == '/' || next() == '^' || next() == '%') {
			char op = next();
			_pos++;
			parsePrimary();
			if (op == '*')
				_codedExpression[_codeSize++] = TIMES;
			else if (op == '/')
				_codedExpression[_codeSize++] = DIVIDE;
			else if (op == '^')
				_codedExpression[_codeSize++] = POWER;
			else if (op == '%')
				_codedExpression[_codeSize++] = MODULO;
			skip();
		}
	}

	/**
	 * Parses an expression such as:
	 * [VARIABLE1 | VARIABLE2| FUNCTION | NUMBER | (FORMULA) ]
	 */
	private void parsePrimary( ) {
		skip();
		char ch = next();
		if (ch == '$') {
			_pos++;
			if ( Integer.valueOf(next())<(48+_nbParam) && Integer.valueOf(next())>=48){
				_codedExpression[_codeSize++] = VARIABLES[-48+Integer.valueOf(next())];
			} else //rajouter des conditions. Du genre while
			error(Ariane.lang.getString("arithmetic.error2")+" "+ch+next());
			_pos++;
		} else if (Character.isLetter(ch))
			parseFunction();
		else if (Character.isDigit(ch) || ch == '.')
			parseNumber();
		else if (ch == '(') {
			_pos++;
			parseFormula();
			skip();
			if (next() != ')')
				error(Ariane.lang.getString("arithmetic.error3"));
			_pos++;
		} else if (ch == ')')
			error(Ariane.lang.getString("arithmetic.error4"));
		else if (ch == '+' || ch == '-' || ch == '*' || ch == '/' || ch == '^')
			error(Ariane.lang.getString("arithmetic.error5.1") + ch + Ariane.lang.getString("arithmetic.error5.2"));
		else if (ch == 0)
			error(Ariane.lang.getString("arithmetic.error6"));
		else
			error(Ariane.lang.getString("arithmetic.error7.1")+ch + Ariane.lang.getString("arithmetic.error7.2"));
	}

	/**
	 * Parses a word that must be a function name.
	 */
	private void parseFunction( ) {
		String w = "";
		while (Character.isLetterOrDigit(next())) {
			w += next();
			_pos++;
		}
		w = w.toLowerCase();
		for (int i = 0; i < _functionNamesUnary.length; i++) {
			if (w.equals(_functionNamesUnary[i])) {
				skip();
				if (next() != '(')
					error(Ariane.lang.getString("arithmetic.error8.1`")+ w + Ariane.lang.getString("arithmetic.error8.2"));
				_pos++;
				parseFormula();
				skip();
				if (next() != ')')
					error(Ariane.lang.getString("arithmetic.error9")+ w + "'.");
				_pos++;
				_codedExpression[_codeSize++] = (byte)(SIN - i);
				return;
			}
		}
		for (int i = 0; i < _functionNamesBinary.length; i++) {
			if (w.equals(_functionNamesBinary[i])) {
				skip();
				if (next() != '(')
					error(Ariane.lang.getString("arithmetic.error10.1")+w+Ariane.lang.getString("arithmetic.error10.2"));
				_pos++;
				parseFormula();
				skip();
				if (next() != ',')
					error(Ariane.lang.getString("arithmetic.error11") + w + "'.");
				_pos++;
				parseFormula();
				skip();
				if (next() != ')')
					error(Ariane.lang.getString("arithmetic.error12") + w + "'.");
				_pos++;
				_codedExpression[_codeSize++] = (byte)(MIN - i);
				return;
			}
		}
		error(Ariane.lang.getString("arithmetic.error13")+ w );
	}

	/**
	 * Parses a numerical value which must be a real value.
	 */
	private void parseNumber( ) {
		String w = "";
		while (Character.isDigit(next())) {
			w += next();
			_pos++;
		}
		if (next() == '.') {
			w += next();
			_pos++;
			while (Character.isDigit(next())) {
				w += next();
				_pos++;
			}
		}
		if (w.equals("."))
			error(Ariane.lang.getString("arithmetic.error14"));
		if (next() == 'E' || next() == 'e') {
			w += next();
			_pos++;
			if (next() == '+' || next() == '-') {
				w += next();
				_pos++;
			}
			if (! Character.isDigit(next()))
				error(Ariane.lang.getString("arithmetic.error15"));
			while (Character.isDigit(next())) {
				w += next();
				_pos++;
			}
		}
		double d = Double.NaN;
		try {
			d = Double.valueOf(w).doubleValue();
		}
		catch (Exception e) {
		}
		if (Double.isNaN(d))
			error(Ariane.lang.getString("arithmetic.error16.1")+w+Ariane.lang.getString("arithmetic.error16.2"));
		_codedExpression[_codeSize++] = (byte)_constantStackIndex;
		_constantStack[_constantStackIndex++] = d;
	}

	/**
	 * @return the string representation of the expression.
	 */
	@Override
	public String toString () {
		return _initialExpression;
	}
}