/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item.etc;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.RandomAccessFile;

import javax.swing.ImageIcon;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.border.BevelBorder;
import javax.swing.border.EmptyBorder;

import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.utils.ImageDecoder;

/**
 * An implementation of image previewer that is used in FileChooser.
 * Allows the visualization in thumbnail image for supported image format,
 * such as GIF or JPEG or PANDORE.
 *
 * @author Xavier Roche, Yann Philippot 1998
 * @version Régis CLOUARD - Jan 01, 2003
 * @version Régis CLOUARD - Nov 27, 2011 (optimization for big and 3D images)
 */
public class FileImagePreviewer extends JPanel implements PropertyChangeListener {
	private final int MAXSIZE=120;
	private int _previewLength = 300;

	private File _selectedFile;
	private ImageIcon _thumbnail;
	private JTextArea _textArea;

	/**
	 * Allows file preview for known file format (png, jpg, pan,..).
	 * 
	 * @param fc the specified file chooser.
	 */
	public FileImagePreviewer( JFileChooser fc ) {
		super(new BorderLayout());
		setPreferredSize(new Dimension(200, 50));
		setBorder(new BevelBorder(BevelBorder.LOWERED));
		fc.addPropertyChangeListener(this);

		_textArea = new JTextArea();
		_textArea.setEditable(false);
		_textArea.setBorder(new EmptyBorder(5,5,5,5));
		_textArea.setFont(PreferenceChooser.fontDialogPLAIN8);
	}

	/**
	 * Triggered when an item is selected from the FileChooser.
	 * 
	 * @param e the related event.
	 */
	@Override
	public void propertyChange( PropertyChangeEvent e ) {
		if (e.getPropertyName().equals(JFileChooser.SELECTED_FILE_CHANGED_PROPERTY)) {
			_selectedFile = (File) e.getNewValue();
			removeAll();
			if (_selectedFile!=null) {
				if (isText(_selectedFile, false)) {
					loadText();
					add(_textArea, BorderLayout.CENTER);
				} else {
					loadImage();
					add(new JLabel(_thumbnail), BorderLayout.CENTER);
				}
			}
			revalidate();
			repaint();
		}
	}

	/**
	 * Loads data from the related file and builds an
	 * icon if the file contains supported image data
	 * (such as, gif, jpeg or PANDORE).
	 */
	private void loadImage( ) {
		if (_selectedFile != null) {
			_thumbnail = ImageDecoder.getThumbnail(_selectedFile.getPath(), MAXSIZE);
		}
	}

	private void loadText() {
		if (_selectedFile != null) {
			String preview = "";
			try {
				FileInputStream fin = new FileInputStream(_selectedFile);
				byte[] data = new byte[_previewLength];
				int bytesRead = 0;
				for (int i=0; i< _previewLength; i++) {
					int b = fin.read();
					if (b == -1) break;
					bytesRead++;
					data[i] = (byte) b;
				}
				preview = new String(data, 0, bytesRead);
				fin.close();
				_textArea.setText(preview);
			} catch (IOException ex) {
				
			}
		}
	}

	/**
	 * Test if a file contains only Latin-1 characters.
	 *
	 * @param file the file to test
	 * @param acceptUTF8encoding  if <ff>true</ff>, a file in UTF-8 encoding will be accepted if it contains only
	 * characters existing in the Latin-1 table.
	 * @return
	 * @throws IOException
	 */
	public static boolean isText(final File file, boolean acceptUTF8encoding) {
		if(!file.exists())
			return false;

		final int BUFFER_SIZE = 1024;
		boolean isText = true;
		byte[] buffer = new byte[BUFFER_SIZE];
		try {
			final RandomAccessFile fis = new RandomAccessFile(file, "r");
			try {
				fis.seek(0);
				final int read = fis.read(buffer);
				int lastByteTranslated = 0;
				for (int i = 0; i < read && isText; i++) {
					final byte b = buffer[i];
					int ub = b & (0xff);  // unsigned
					int utf8value = lastByteTranslated + ub;
					lastByteTranslated = (ub) << 8;

					if ( 	   ub == 0x09 /*(tab)*/
							|| ub == 0x0A /*(line feed)*/
							|| ub == 0x0C /*(form feed)*/
							|| ub == 0x0D /*(carriage return)*/
							|| (ub >= 0x20 && ub <= 0x7E) /* Letters, Numbers and other "normal synbols" */
							|| (ub >= 0xA0 && ub <= 0xEE) /* Symbols of Latin-1 */
							|| (acceptUTF8encoding && (utf8value >= 0x2E2E && utf8value <= 0xC3BF)) /* Latin-1 in UTF-8 encoding */) {
						// ok
					} else {
						isText = false;
						break;
					}
				}
			} finally {
				try {
					fis.close();
				} catch (final Throwable th) {
				}
			}

			return isText;
		} catch ( IOException e) {
			return false;
		}
	}
}