/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item.etc;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.ResourceBundle;

import javax.swing.AbstractCellEditor;
import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;
import javax.swing.text.JTextComponent;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;

/**
 * A <code>PropertyViewer</code> provides 
 * a dialog box to display graph item's properties.
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Sep 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Jun 10, 2010 (new design)
 * @version Régis CLOUARD - May 10, 2011
 * @version Régis CLOUARD -Jul 07, 2012 (Put button panel below the properties)
 */
public class PropertyViewer extends JPanel {
	/** The result value. */
	private JLabel _resultLabel;

	/** The list of all output result buttons. */
	ArrayList<JButton> _inputButtons = new ArrayList<JButton>();

	/** The list of all output result buttons. */
	ArrayList<JButton> _outputButtons = new ArrayList<JButton>();

	/** The default row height. */
	private final int DEFAULT_ROW_HEIGHT=25;

	/** The language used to display the properties. */
	private ResourceBundle _lang;

	/** The list of properties as a pair (name, value). */
	private ArrayList<PropertyCouple> _properties = new ArrayList<PropertyCouple>();

	private static Border unselectedBorder;
	private static Border selectedBorder;

	/**
	 * Creates an empty frame.
	 * Used for note for example.
	 */
	public PropertyViewer( ) { setOpaque(false); }

	/**
	 * Creates a convenient dialog box
	 * for the specified item.
	 * 
	 * @param lang the language of display.
	 * @param item the related graph item
	 */
	public PropertyViewer( ResourceBundle lang, final GraphItem item ) {
		_lang = lang;
		setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PropertyViewer.title")+" ("+item.getId()+")"));
		buildPropertyList(item);

		JTable table = new CustomTraversableTable(new PropertyTableModel());
		table.getTableHeader().setFont(PreferenceChooser.fontDialogPLAIN10);
		table.setFont(PreferenceChooser.fontDialogPLAIN10);
		table.getColumnModel().getColumn(1).setCellRenderer(new PropertyCellRenderer());
		table.getColumnModel().getColumn(1).setCellEditor(new PropertyCellEditor());
		selectedBorder =  UIManager.getBorder("Table.focusCellHighlightBorder") ;
		unselectedBorder = new EmptyBorder(1, 2, 1, 2);

		JPanel buttonPanel = new JPanel(new GridLayout(1,2));
		JButton buttonOK;
		buttonOK = new JButton(Ariane.lang.getString("Graph.Menu.execute")+" (ctrl-R)");
		buttonOK.setFont(PreferenceChooser.fontDialogPLAIN10);
		buttonOK.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) {
				item.getGraph().execute();
			}
		});
		buttonPanel.add(buttonOK);
		
		JButton buttonReset= new JButton();
		buttonReset = new JButton(Ariane.lang.getString("Graph.Menu.reset"));
		buttonReset.setFont(PreferenceChooser.fontDialogPLAIN10);
		buttonReset.addActionListener(new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) {
				item.resetAll();
			}
		});
		buttonPanel.add(buttonReset);

		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		JScrollPane scrollPane = new JScrollPane(table);
		Dimension d;
		d = table.getPreferredSize();
		d.width=10000; d.height += 25;
		scrollPane.setMaximumSize(d);
		scrollPane.getViewport().setBackground(Color.WHITE);
		add(scrollPane);

		d = buttonPanel.getPreferredSize();
		d.width=10000;
		buttonPanel.setMaximumSize(d);
		add(buttonPanel);
		
//		// TODO
//		JPanel panel = new JPanel(new BorderLayout());
//		d = panel.getPreferredSize();
//		d.width=10000;
//		panel.setMaximumSize(d);
//		panel.add(new JTextArea());
//		add(panel);
	}

	/**
	 * @return the current language used to display the properties.
	 */
	public final ResourceBundle getLanguage() {
		return _lang;
	}

	/**
	 * This class is a simple holder for a property pair
	 * (a title and an editor for the value).
	 */
	public static final class PropertyCouple {
		/** The title of the property. */
		public String title;
		/** The value of the property. */
		public JComponent value;

		/**
		 * Create a property pair title value.
		 * @param t a title name.
		 * @param v a component for editing the value.
		 */
		public PropertyCouple( String t, JComponent v) {
			title=t;
			v.setFont(PreferenceChooser.fontDialogPLAIN10);
			value=v;
		}
	}

	/**
	 * Builds the list of properties as an array list.
	 * Template Method Design Pattern
	 * 
	 * @param item the related graph item.
	 */
	private final void buildPropertyList( final GraphItem item ) {
		_properties.add(new PropertyCouple(Ariane.lang.getString("PropertyViewer.name"), item.getNameEditor()));
		if (item.getCategory()!=null) {
			JLabel l = new JLabel(item.getCategory());
			l.setEnabled(false);
			_properties.add(new PropertyCouple(Ariane.lang.getString("PropertyViewer.category"), l));
		}
		JComponent editor = item.getBinaryEditor();
		if (editor!=null) {
			_properties.add(new PropertyCouple(Ariane.lang.getString("PropertyViewer.binary"), editor));
		}
		{
			JButton jbuttonman;
			jbuttonman= new JButton(Ariane.lang.getString("PropertyViewer.manual"));
			jbuttonman.setFont(PreferenceChooser.fontDialogPLAIN10);
			jbuttonman.addActionListener(new ActionListener() {
				@Override
				public final void actionPerformed(ActionEvent e) {
					item.showMan();
				}
			});
			_properties.add(new PropertyCouple(Ariane.lang.getString("PropertyViewer.documentation"), jbuttonman));
		}
		if (item.getFormula()!=null && item.displayFormula()) {
			_properties.add(new PropertyCouple(item.getFormulaLabel(), item.getFormulaEditor()));
		}
		ArrayList<PropertyCouple> additionalProperties = item.getSpecificProperties( );
		if (additionalProperties!=null) _properties.addAll(additionalProperties);

		// Parameters.
		for (int i = 0; i < item.getInSlotParameters().size(); i++) {
			InSlotParameter isp = (item.getInSlotParameters().get(i));
			_properties.add(new PropertyCouple(item.getInSlotParameters().get(i).getName() ,isp.getParameterEditor()));
		}
		// Input images.
		for (int i = 0; i < item.getInSlotImages().size(); i++) {
			JButton button=((item.getInSlotImages().get(i))).getButton();
			if (item.getExecutionStatus() != GraphItem.Status.EXECUTED || (!((item.getInSlotImages().get(i))).isValidate())) {
				button.setEnabled(false);
			}
			_inputButtons.add(button);
			_properties.add(new PropertyCouple(Ariane.lang.getString("PropertyViewer.input")+" #" + i,button));
		}
		// Output images.
		for (int i = 0; i < item.getOutSlotImages().size(); i++) {
			JButton button=((item.getOutSlotImages().get(i))).getButton();
			if (item.getExecutionStatus() != GraphItem.Status.EXECUTED || (!((item.getOutSlotImages().get(i))).isValidate())) {
				button.setEnabled(false);
			}
			_outputButtons.add(button);
			_properties.add(new PropertyCouple(Ariane.lang.getString("PropertyViewer.output")+" #" + i+" (ctrl-"+i+")",button));
		}
		// Result
		if (item.getOutSlotResults().size() > 0) {
			if (item.getExecutionStatus() != GraphItem.Status.EXECUTED) {
				_resultLabel = new JLabel(Ariane.lang.getString("PropertyViewer.result.unbound"));
			} else {
				_resultLabel = new JLabel(item.getOutSlotResult().getValue());
			}
			_resultLabel.setEnabled(false);
			_properties.add(new PropertyCouple(Ariane.lang.getString("PropertyViewer.result"),_resultLabel));
		}
	}

	/**
	 * Updates the contents of the panel.
	 * 
	 * @param item the related item.
	 */
	public final void update( GraphItem item ) {
		boolean updated=false;
		if (_resultLabel!=null) {
			updated=true;
			if (item.getExecutionStatus() != GraphItem.Status.EXECUTED) {
				_resultLabel.setText(Ariane.lang.getString("PropertyViewer.result.unbound"));
			} else {
				_resultLabel.setText(item.getOutSlotResult().getValue());
			}
		}
		if (_inputButtons.size()>0) {
			for (int i = 0; i < item.getInSlotImages().size(); i++) {
				updated=true;
				if (item.getExecutionStatus() != GraphItem.Status.EXECUTED || (!((item.getInSlotImages().get(i))).isValidate())) {
					_inputButtons.get(i).setEnabled(false);
				} else {
					_inputButtons.get(i).setEnabled(true);
				}
			}
		}
		if (_outputButtons.size()>0) {
			for (int i = 0; i < item.getOutSlotImages().size(); i++) {
				updated=true;
				if (item.getExecutionStatus() != GraphItem.Status.EXECUTED || (!((item.getOutSlotImages().get(i))).isValidate())) {
					_outputButtons.get(i).setEnabled(false);
				} else {
					_outputButtons.get(i).setEnabled(true);
				}
			}
		}
		if (updated) {
			revalidate();
			repaint();
		}
	}

	/**
	 * The property table model.
	 *
	 */
	final class PropertyTableModel extends AbstractTableModel {
		/** The name of the table columns. */
		private final String[] _columnNames = {"PropertyViewer.table.title1","PropertyViewer.table.title2"};

		/**
		 * Gets the number of columns for this table.
		 * 
		 * @return the number of columns.
		 */
		@Override
		public int getColumnCount( ) {
			return _columnNames.length;
		}

		/**
		 * Gets the number of rows in the table.
		 * 
		 * @return the number of rows.
		 */
		@Override
		public int getRowCount( ) {
			return _properties.size();
		}

		/**
		 * Gets the name of the specified column.
		 * 
		 * @param col the index of the column.
		 * @return the name of the column.
		 */
		@Override
		public String getColumnName( int col ) {
			return Ariane.lang.getString(_columnNames[col]);
		}

		/**
		 * Returns whether the selected cell is editable.
		 * 
		 * @param row the row of the selected cell.
		 * @param col the column of the selected cell.
		 * @return true if the cell is editable, false otherwise.
		 */
		@Override
		public boolean isCellEditable( int row, int col ) {
			return col==1;
		}

		/**
		 * Gets the value of the table in the specified cell.
		 * 
		 * @param row the row.
		 * @param col the column.
		 * @return the cell content.
		 */
		@Override
		public Object getValueAt( int row, int col ) {
			switch(col) {
			case 0: return _properties.get(row).title;
			case 1: return _properties.get(row).value;
			}
			return null;
		}
	}

	/**
	 * A cell renderer.
	 */
	final class PropertyCellRenderer implements TableCellRenderer {
		/**
		 * {@inheritDoc}
		 */
		@Override
		public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
			JComponent c = _properties.get(row).value;
			if (isSelected) {
				c.setBorder(selectedBorder);
			} else {
				c.setBorder(unselectedBorder);
			}
			return c;
		}
	}

	/**
	 * A cell editor.
	 *
	 */
	final class PropertyCellEditor extends AbstractCellEditor implements ActionListener, TableCellEditor {
		/**
		 * {@inheritDoc}
		 */
		@Override
		public void actionPerformed( ActionEvent e ) {
			stopCellEditing();
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Object getCellEditorValue( ) {
			return null;
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Component getTableCellEditorComponent( JTable table, Object value, boolean isSelected, int row, int column) {
			return _properties.get(row).value;
		}
	}

	/**
	 * Redefinition of the table to change the cell traversal policy.
	 */
	final class CustomTraversableTable extends JTable {
		public CustomTraversableTable( PropertyTableModel model ) {
			super(model);
			putClientProperty("terminateEditOnFocusLost", Boolean.TRUE);
			getColumnModel().getColumn(0).setPreferredWidth(100);
			getColumnModel().getColumn(1).setPreferredWidth(150);
			setShowGrid(true);
			setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			setAutoResizeMode(JTable.AUTO_RESIZE_LAST_COLUMN);
			setRowHeight(DEFAULT_ROW_HEIGHT);
			setCellSelectionEnabled(true);
			getSelectionModel().addListSelectionListener(new ListSelectionListener() {
				@Override
				public void valueChanged(ListSelectionEvent e){
					if (!e.getValueIsAdjusting()) {
						int row = getSelectedRow();
						int column = getSelectedColumn();
						Object c = getModel().getValueAt(row,column);
						if ( c instanceof JTextComponent) {
							JTextComponent jf = (JTextComponent)c;
							editCellAt(row,column);
							jf.selectAll();
							jf.requestFocusInWindow();
							jf.setBorder(selectedBorder);
						}
					}
				}
			});
		}

		/**
		 * Changes the next cell to be edited. Only use the second column.
		 */
		@Override
		public void changeSelection(int rowIndex, int columnIndex, boolean toggle, boolean extend ) {
			if (getModel().isCellEditable(rowIndex, columnIndex)) {
				super.changeSelection(rowIndex, columnIndex, toggle, extend);
			} else {
				int curRow = getEditingRow();
				int curColumn = getEditingColumn();
				if (curRow == -1) curRow = getSelectedRow();
				if (curColumn == -1) curColumn = getSelectedColumn();
				int nRows = getRowCount();

				int row = rowIndex;
				int direction;
				if (curRow<rowIndex || (curRow == nRows-1 && rowIndex==0)) {
					direction = 1;
					row--;
				} else if (curRow<rowIndex) {
					direction = -1;
					row++;
				} else if (curColumn<columnIndex){
					direction = 1;
				} else {
					direction = -1;
				}
				do {
					row+=direction;
					if (row >= nRows) {
						row =0;
					} else if (row <0) {
						row = nRows-1;
					}
					if (isCellEditable(row, 1)) {
						break;
					}
				} while (row!=rowIndex);
				super.changeSelection(row, 1, toggle, extend);
			}
		}
	}
}