/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.slot;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.Point;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;

import javax.swing.ImageIcon;
import javax.swing.JButton;

/**
 * The class <code>InSlotImage</code> defines input image connection slot.
 * It can only be connected to an <code>OutImageSlot</code>.
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Frédéric LAIR - May 23, 2007
 */
public final class InSlotImage extends InSlot {
	/** The connected rendering of the slot. */
	transient private static ImageIcon _connectedIcon=null;

	/** The disconnected rendering of the slot. */
	transient private static ImageIcon _disconnectedIcon=null;

	/**
	 * Creates an input image slot on the given graph item.
	 * 
	 * @param g the graph item that will hold the slot.
	 * @param rank the rank in the list of input slot of the graph item.
	 */
	public InSlotImage( GraphItem g, int rank ) {
		super(g, rank);

		// Load connected and disconnected icons.
		if (_connectedIcon == null) {
			ClassLoader classLoader = getClass().getClassLoader();
			_connectedIcon = new ImageIcon(classLoader.getResource(Ariane.iconDir+"SlotDataOn.png"));
			_disconnectedIcon = new ImageIcon(classLoader.getResource(Ariane.iconDir+"SlotDataOff.png"));
		}
		setIcon(_disconnectedIcon);
		dropTarget = new DropTarget(this, DnDConstants.ACTION_LINK, this, true);
	}

	/**
	 * @return a new link.
	 */
	@Override
	protected Link createLink() {
		return new DataLink(this);
	}

	/**
	 * Returns a textual description of the slot.
	 * For debugging purpose only.
	 * 
	 * @return a string.
	 */
	@Override
	public String toString( ) { return "InSlotImage #" +getRank(); }

	/**
	 * Returns the icon for connected rendering.
	 * 
	 * @return an icon.
	 */
	@Override
	protected ImageIcon getConnectedIcon( ) { return _connectedIcon; } 

	/**
	 * Returns the icon for disconnected rendering.
	 * 
	 * @return an icon.
	 */
	@Override
	protected ImageIcon getDisconnectedIcon( ) { return _disconnectedIcon; }

	/**
	 * Returns the anchor coordinates of the slot.
	 * 
	 * @return anchor point.
	 */
	@Override
	public Point getAnchor( ) {
		return(new Point((int)(getLocation().getX()-3 + getGraphItem().getLocation().getX() ),
						 (int)(getLocation().getY()+getHeight()/2+ getGraphItem().getLocation().getY())));
	}

	/**
	 * Returns the current color of that link type.
	 * 
	 * @return the user defined color.
	 */
	@Override
	protected Color getLinkColor( ) {
		return Graph.getImageLinkColor();
	}

	/**
	 * Tries to get the image name of the connected slot.
	 * The value (a file) exists if:
	 * <ul>
	 * <li>the slot is connected;
	 * <li>the result of the connected slot is validate.
	 * </ul>
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the result is validate.
	 */
	public boolean hasValue( boolean imperative ) {
		// The slot is connected.
		if (isConnected()) {
			return getConnectedSlot().hasValue(imperative);
		} else
			return false;
	}

	/**
	 * Returns the name of the image, which is the name
	 * the value of the connected slot.
	 * 
	 * @return an image name.
	 */
	public String getValue( ) {
		return ((OutSlotImage)getConnectedSlot()).getValue();
	}

	/**
	 * Creates the button that allows to display the result.
	 * 
	 * @return the JButton.
	 */
	public JButton getButton( ) {
		JButton button = new JButton(Ariane.lang.getString("OutSlotImage.display"));
		button.addActionListener(new ActionListener( ) {
			@Override
			public void actionPerformed( ActionEvent e ) {
				((OutSlotImage)getConnectedSlot()).showImage();
			}
		});
		return button;
	}

	/**
	 * Determines if the slot has a validated result.
	 * 
	 * @return true if the result is valid.
	 */
	public boolean isValidate( ) {
		// The slot is connected.
		if (isConnected()) {
			return getConnectedSlot().isValidate();
		} else
			return false;
	}

	/**
	 * Saves the slot with its values as an XML tree.
	 * 
	 * @return the root of the XML tree.
	 * @throws IOException XML problem.
	 */
	@Override
	public XMLElement save( ) throws IOException {
		if (isConnected()) {
			XMLElement e = new XMLElement("Connection");
			e.addChildren(new XMLElement("EndItemId",""+getGraphItem().getId()));
			e.addChildren(new XMLElement("EndSlotType","InSlotImage"));
			e.addChildren(new XMLElement("EndSlotIndex",""+getRank()));
			getConnectedSlot().save(e);
			if (getLink().getPath().size() != 0) {
				e.addChildren(new XMLElement("UserDefined", ""+getLink().isUserDefined()));
				XMLElement pa = new XMLElement("LinkPath");
				for (int i = 0; i<getLink().getPath().size(); i++) {
					XMLElement po = new XMLElement("LinkPoint");
					po.addChildren(new XMLElement("IdPoint", Integer.toString(i)));
					po.addChildren(new XMLElement("X-coordinate", Integer.toString((int)getLink().getPath().get(i).getX())));
					po.addChildren(new XMLElement("Y-coordinate", Integer.toString((int)getLink().getPath().get(i).getY())));
					pa.addChildren(po);
				}
				e.addChildren(pa);
			}
			return e;
		}
		return null;
	}
}