/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */
package fr.greyc.pantheon.ariane.graph.slot;

import java.awt.Color;
import java.awt.Point;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.io.IOException;

import javax.swing.ImageIcon;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * The class <code>InSlotMask</code> defines an input mask connection slot.
 * It can only be connected to an <code>OutImageSlot</code>.
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 25, 2004
 * @version Frédéric LAIR - May 23, 2007
 */
public final class InSlotMask extends InSlot {
	/** The connected rendering of the slot. */
	transient private static ImageIcon _connectedIcon=null;

	/** The disconnected rendering of the slot. */
	transient private static ImageIcon _disconnectedIcon=null;

	/**
	 * Creates an input mask slot on the given graph item.
	 * 
	 * @param g the graph item that will hold the slot.
	 */
	public InSlotMask( GraphItem g ) {
		super(g,0);

		// Load connected and disconnected icons.
		if (_connectedIcon == null) {
			ClassLoader classLoader = getClass().getClassLoader();
			_connectedIcon = new ImageIcon(classLoader.getResource(Ariane.iconDir+"SlotMaskOn.png"));
			_disconnectedIcon = new ImageIcon(classLoader.getResource(Ariane.iconDir+"SlotMaskOff.png"));
		}
		setIcon(_disconnectedIcon);
		dropTarget = new DropTarget(this , DnDConstants.ACTION_LINK, this, true);
	}

	/**
	 * @return a new link.
	 */
	@Override
	protected Link createLink() {
		return new MaskLink(this);
	}

	/**
	 * Returns a textual description of the slot.
	 * For debugging purpose only.
	 * 
	 * @return a string.
	 */
	@Override
	public String toString( ) {
		return "InSlotMask";
	}

	/**
	 * Returns the icon for connected rendering.
	 * 
	 * @return an icon.
	 */
	@Override
	protected ImageIcon getConnectedIcon( ) {
		return _connectedIcon;
	} 

	/**
	 * Returns the icon for disconnected rendering.
	 * 
	 * @return an icon.
	 */
	@Override
	protected ImageIcon getDisconnectedIcon( ) {
		return _disconnectedIcon;
	}

	/**
	 * Returns the anchor coordinates of the slot.
	 * 
	 * @return anchor point.
	 */
	@Override
	public Point getAnchor( ) {
		return( new Point((int) ( getLocation().getX() + getWidth()/2+getGraphItem().getLocation().getX() ),
				(int) (getLocation().getY() + getGraphItem().getLocation().getY())));
	}

	/**
	 * Returns the current color of that link type.
	 * 
	 * @return the user defined color.
	 */
	@Override
	protected Color getLinkColor( ) {
		return Graph.getMaskLinkColor();
	} 

	///// 
	/////  R E S U L T
	/////

	/**
	 * Tries to get the image name of the connected slot.
	 * Result exists if:
	 * <ul>
	 * <li>the slot is connected;
	 * <li>the result of the connected slot is validate.
	 * </ul>
	 * 
	 * @param imperative if true need to be executed first.
	 * @return true if the result is validate.
	 */
	public boolean hasValue( boolean imperative ) {
		// The slot is connected.
		if (isConnected()) {
			return ((OutSlotImage)getConnectedSlot()).hasValue(imperative);
		} else
			return false;
	}

	/**
	 * Returns the name of the mask which is the name
	 * of the connected <b>OutSlotImage</b>.
	 * 
	 * @return the filename.
	 */
	public String getValue( ) {
		return ((OutSlotImage)getConnectedSlot()).getValue();
	}

	/**
	 * Saves the slot with its values as an XML tree.
	 * 
	 * @return the root of the XML tree.
	 * @throws IOException	XML problem.
	 */
	@Override
	public XMLElement save( ) throws IOException {
		if (isConnected()) {
			XMLElement e = new XMLElement("Connection");
			e.addChildren(new XMLElement("EndItemId",Integer.toString(getGraphItem().getId())));
			e.addChildren(new XMLElement("EndSlotType","InSlotMask"));
			e.addChildren(new XMLElement("EndSlotIndex",Integer.toString(getRank())));
			getConnectedSlot().save(e);
			if (getLink().getPath().size() != 0){
				e.addChildren(new XMLElement("UserDefined", ""+getLink().isUserDefined()));
				XMLElement pa = new XMLElement("LinkPath");
				for (int i = 0; i<getLink().getPath().size(); i++) {
					XMLElement po = new XMLElement("LinkPoint");
					po.addChildren(new XMLElement("IdPoint", Integer.toString(i)));
					po.addChildren(new XMLElement("X-coordinate",Integer.toString((int)getLink().getPath().get(i).getX())));
					po.addChildren(new XMLElement("Y-coordinate",Integer.toString((int)getLink().getPath().get(i).getY())));
					pa.addChildren(po);
				}
				e.addChildren(pa);
			}
			return e;
		}
		return null;
	}
}