/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.slot;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragSource;
import java.awt.dnd.DropTargetDropEvent;
import java.io.IOException;
import java.util.ArrayList;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * The class <code>OutSlot</code> defines an output connection slot.
 * Several <code>InSlot</code> can be connected to an output slot.
 * Two types of output slots are available:
 * <ul>
 * <li>@see OutSlotImage
 * <li>@see outSlotResult
 * </ul>
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD  - Oct 16, 2001
 */
public abstract class OutSlot extends Slot {
	/** Is the result of the slot validate? */
	private boolean _validate = false;
	
	/** The list of connected input slots. */
	 private transient ArrayList<InSlot> _connectedSlots = null;

	/** The DataFlavor as a Transferable object. */
	public static DataFlavor[] flavors ={ new DataFlavor("application/x-java-serialized-object-out-slot", "OutSlot Flavor")};

	/**
	 * Creates an output slot on the given graph item.
	 * 
	 * @param g the graph item.
	 * @param rank the rank in the list of input slot of the graph item.
	 */
	public OutSlot( GraphItem g, int rank ) {
		super(g, rank);
		initDrop();
		_connectedSlots = new ArrayList<InSlot>();
	}

	///// 
	///// D  R A G   a n d   D R O P
	/////

	/**
	 * Initializes the slot as a droppable component.
	 */
	protected final void initDrop( ) {
		dragSource = DragSource.getDefaultDragSource();
		dragSource.createDefaultDragGestureRecognizer(this, DnDConstants.ACTION_MOVE, this);
	}

	/**
	 * Gets the transfered object.
	 * 
	 * @return the list of recognized flavors.
	 */
	@Override
	public final DataFlavor[] getTransferDataFlavors( ) {
		return flavors;
	}

	/** 
	 * Is the object a transferable object?
	 * 
	 * @param flavor the flavor to be tested.
	 * @return true if the flavor is transferable.
	 */
	@Override
	public final boolean isDataFlavorSupported( DataFlavor flavor ) { 
		return flavor.equals(flavors[0]);
	}

	/**
	 * Gets the transfered data.
	 * 
	 * @param flavor the flavor of the transfered object.
	 * @return the data.
	 * @throws UnsupportedFlavorException unrecognized flavors.
	 * @throws IOException in case of IO errors.
	 */
	@Override
	public final Object getTransferData( DataFlavor flavor ) throws UnsupportedFlavorException, IOException {
		if (isDataFlavorSupported(flavor))
			return this;
		else
			throw new UnsupportedFlavorException(flavor);
	}

	/**
	 * The DropTarget intercepts drop() notifications before the
	 * registered DropTargetListener gets them.
	 * Connects the two slots if the connection is possible:
	 * <ul>
	 * <li>drag and drop slots are different ;
	 * <li>slots are compatibles.
	 * </ul>
	 * 
	 * @param e the DropTargetDropEvent.
	 */
	@Override
	public final void drop( DropTargetDropEvent e ) {
		setNormalBorder();
		try {
			if (e.isDataFlavorSupported(InSlot.flavors[0])) {// Vient d'un InSlot.
				InSlot inSlot = (InSlot)(e.getTransferable().getTransferData(InSlot.flavors[0]));
				boolean connect = false;
				// check if the slot are in the same graph!
				if (getGraph() != (inSlot).getGraph())
					throw new Exception(Ariane.lang.getString("Slot.error3"));
				// check slots consistancy.
				if (((this instanceof OutSlotImage) && (inSlot instanceof InSlotImage))
						||
						((this instanceof OutSlotImage) && (inSlot instanceof InSlotMask))
						||
						((this instanceof OutSlotResult) && (inSlot instanceof InSlotParameter))) {
					connect = true;
				}
				if (connect) {
					getGraph().startActionEdition();
					inSlot.connectFrom(this);
					getGraph().endActionEdition();
					getGraph().unSelectAll();
					e.acceptDrop(DnDConstants.ACTION_MOVE);
					e.dropComplete(true);
				} else
					throw new Exception(Ariane.lang.getString("Slot.error2"));
			} else
				throw new Exception(Ariane.lang.getString("Slot.error2"));
		} catch (UnsupportedFlavorException ex) {
			e.rejectDrop();
//			ex.printStackTrace();
			e.dropComplete(false);
			Ariane.workspace.printErrorMessage(Ariane.lang.getString("Slot.error1")+ ex.getLocalizedMessage());
		} catch (Exception ex) {
			e.rejectDrop();
//			ex.printStackTrace();
			e.dropComplete(false);
			Ariane.workspace.printErrorMessage(Ariane.lang.getString("Slot.error1")+ ex.getLocalizedMessage());
		}
	}

	///// 
	///// C O N N E C T I O N   /   D I S C O N N E C T I O N 
	/////

	/**
	 * @return the list of connected slots.
	 */
	public final ArrayList<InSlot> getConnectedSlots( ) {
		return _connectedSlots;
	}

	/**
	 * Adds the slot in the list of connected slots.
	 * @param slot the slot to be connected.
	 */
	public final void setConnectedSlots( InSlot slot ) {
		_connectedSlots.add(slot);
		setIcon(getConnectedIcon());
	}

	/**
	 * Determines if the slot is connected.
	 * 
	 * @return true if the slot is already in the list of connected slots.
	 */
	@Override
	public final boolean isConnected( ) {
		return (_connectedSlots.size() > 0);
	}

	/**
	 * Disconnected the given slot.
	 * 
	 * @param inSlot the input slot to be disconnected.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void disconnect( InSlot inSlot, boolean historical ) {
		inSlot.setConnectedSlot(null,historical);
		inSlot.disposeLink();
		if (_connectedSlots.remove(inSlot)) {
			if (_connectedSlots.size()==0)
				setIcon(getDisconnectedIcon());
		}
	}

	/**
	 * Disconnects all the connected slots.
	 * Implementation: calls the disconnection method of each slot.
	 */
	public final void disconnectAll( ) {
		while (_connectedSlots.size()>0)
			disconnect(_connectedSlots.get(0),true);
	}

	/**
	 * Connects the given slot with the slot.
	 * <br>
	 * If the given slot is already connected to the slot does nothing.
	 * Else, disconnects the given slot if needed
	 * and adds the slot to the list of connected slots.
	 * 
	 * @param toSlot the input slot to be connected.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void connectTo( InSlot toSlot, boolean historical ) {
		// already in the list?
		if (_connectedSlots.indexOf(toSlot) < 0) {
			toSlot.disconnect();
			setConnectedSlots(toSlot);
			toSlot.setConnectedSlot(this,historical);
		}
	}

	/**
	 * Determines if the slot has a validated result.
	 * 
	 * @return true if the result is valid.
	 */
	public final boolean isValidate( ) {
		return _validate;
	}

	/**
	 * Sets the result to validate.
	 */
	protected final void setValidated( ) {
		_validate = true;
	}

	@Override
	public final void setSelected( boolean isSelected ) {
		if (getGraphItem() instanceof VirtualItem) return;
		if (isConnected()) {
			super.setSelected(isSelected);
		}
	}

	public final void reset( ) {
		_validate = false;
	}

	/**
	 * Determines if the slot has a validate result,
	 * if needed executes the related forward operators
	 * to create output images.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the slot has a result.
	 */
	public final boolean hasValue( boolean imperative ) {
		return getGraphItem().execute(imperative) && isValidate();
	}

	/**
	 * Saves the slot with its values as an XML tree.
	 * 
	 * @param e the XML parent in the tree.
	 * @return the root of the XML tree.
	 * @throws IOException is case of output errors.
	 */
	public abstract XMLElement save( XMLElement e ) throws IOException;
}