/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.slot;

import java.awt.Point;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.io.IOException;

import javax.swing.ImageIcon;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * The class <code>OutSlotResult</code> defines result slot.
 * <br>Result value can be :
 * <ul>
 * <li>SUCCESS
 * <li>FAILURE
 * <li>a float
 * <li>a string.
 * </ul>
 * <br>It can be connected to several <code>InSlotParameter</code>.
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 24, 2004
 */
public final class OutSlotResult extends OutSlot {
	/** The result value  */
	private String _result=null;

	/** The connected rendering of the slot. */
	transient private static ImageIcon _connectedIcon=null;

	/** The disconnected rendering of the slot. */
	transient private static ImageIcon _disconnectedIcon=null;

	/**
	 * Creates an output result slot on the given graph item.
	 * 
	 * @param g the graph item that will hold the slot.
	 * @param rank the rank in the input slots list of the graph item.
	 */
	public OutSlotResult( GraphItem g, int rank ) {
		super(g,rank);

		// Load connected and disconnected icons.
		if (_disconnectedIcon == null) {
			ClassLoader classLoader = getClass().getClassLoader();
			_disconnectedIcon = new ImageIcon(classLoader.getResource(Ariane.iconDir+"SlotValueOff.png"));
			_connectedIcon = new ImageIcon(classLoader.getResource(Ariane.iconDir+"SlotValueOn.png"));
		}
		setIcon(_disconnectedIcon);
		dropTarget = new DropTarget(this, DnDConstants.ACTION_LINK, this, true);
	}
	
	/**
	 * Returns a textual description of the slot.
	 * 
	 * @return a string.
	 */
	@Override
	public String toString( ) {
		return "OutSlotResult: " + getValue();
	}

	/**
	 * Returns the icon for connected rendering.
	 * 
	 * @return an icon.
	 */
	@Override
	protected ImageIcon getConnectedIcon( ) {
		return _connectedIcon;
	}

	/**
	 * Returns the icon for disconnected rendering.
	 * 
	 * @return an icon.
	 */
	@Override
	protected ImageIcon getDisconnectedIcon( ) {
		return _disconnectedIcon;
	}

	/**
	 * Returns the anchor coordinates of the slot.
	 * 
	 * @return anchor point.
	 */
	@Override
	public Point getAnchor( ) {
		return( new Point((int)(getLocation().getX()+getWidth()/2+ getGraphItem().getLocation().getX() ),
				(int)(getLocation().getY() + getHeight()+ getGraphItem().getLocation().getY())));
	}

	/**
	 * Sets the given result as the new result.
	 * For boolean value (isNUmeric = FALSE),
	 * value 0 corresponds to FAILURE and 1 to SUCCESS.
	 * 
	 * @param res the string that contains the result.
	 */
	public void setValue( String res ) {
		_result = res;
		setValidated();
	}

	/**
	 * Returns the current value.
	 * 
	 * @return a String.
	 */
	public String getValue() {
		return _result;
	}

	/**
	 * Saves the slot with its values as an XML tree.
	 * 
	 * @param e the XML parent in the tree. 
	 * @return the root of the XML tree.
	 */
	@Override
	public XMLElement save( XMLElement e ) throws IOException {
		e.addChildren(new XMLElement("StartItemId", Integer.toString(getGraphItem().getId())));
		e.addChildren(new XMLElement("StartSlotType", "OutSlotResult"));
		e.addChildren(new XMLElement("StartSlotIndex", Integer.toString(getRank())));
		return e;
	}
}