/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.slot;

import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.datatransfer.Transferable;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.border.LineBorder;

/**
 * The class <code>Slot</code> defines connection slot between 2 graph items.
 * The two derived classes <code>InSlot</code> and <code>OutSlot</code>
 * distinguish the input from output slots. A slot is graphically represented
 * an icon (using JLabel). A slot is both a draggable and a droppable component.
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 10, 2004
 * @version Frédéric LAIR - May 23, 2007
 */
public abstract class Slot extends JLabel implements Transferable, DropTargetListener, DragSourceListener, DragGestureListener {
	/** The rank of the slot in the slots list of the graph item. */
	private final int _rank;

	/** The graph item that supports the slot. */
	private transient GraphItem _graphItem;

	/** The drag site. */
	protected transient DragSource dragSource = null;

	/** The drop site. */
	protected transient DropTarget dropTarget = null;

	private final static int lineWidth=1;

	/** Border for a selected looking of the component. */
	private transient static final LineBorder _activeBorder = (LineBorder)BorderFactory.createLineBorder(Color.RED,lineWidth);

	/** Border for a normal looking of the component.*/
	private LineBorder _normalBorder = null;

	/** Flag that indicates if the slot is currently selected. */
	private boolean _selected = false;

	/** The cursor for dragging motions. */
	private transient final static Cursor SLOTLINK_CURSOR = new Cursor(Cursor.CROSSHAIR_CURSOR);

	/**
	 * Creates a slot on the given <i>GraphItem</i>.
	 * 
	 * @param g the related graph item.
	 * @param rank the rank of the slot in the list of graph's slots.
	 */
	public Slot( GraphItem g, int rank ) {
		super();
		_rank = rank;
		_graphItem= g;
		// Create the normal border (depends on the related graphitem).
		_normalBorder = (LineBorder)BorderFactory.createLineBorder(g.getBackgroundColor(),lineWidth);
		setNormalBorder();
		setFocusable(false);
	}

	/**
	 * Disposes the slot (definitively).
	 * Tries to avoid memory leak.
	 * <b>IT IS SUPPOSED TO BE DISCONNECTED!</b>
	 */
	public final void dispose( ) {
		_graphItem=null;
		_normalBorder=null;
		dragSource=null;
		dropTarget=null;
	}

	/**
	 * Invoked when a slot detects a drag initiating gesture.
	 * Initiates the drag action for the user.
	 * 
	 * @param event describes the gesture that has just occurred.
	 */
	@Override
	public final void dragGestureRecognized( DragGestureEvent event ) {
		dragSource.startDrag(event,SLOTLINK_CURSOR, this, this);
	}

	/**
	 * Invoked to signify that the Drag and Drop operation is completed.
	 * Sets the border to normal looking.
	 * 
	 * @param event the DragSourceDropEvent
	 */
	@Override
	public final void dragDropEnd( DragSourceDropEvent event ) {
		setSelected(false);
		if (!isConnected()) {
			setIcon(getDisconnectedIcon());
		}
	}

	/**
	 * Called when the cursor's hotspot enters a platform-dependent drop site.
	 * Changes the cursor rendering.
	 * 
	 * @param event the DragSourceDragEvent.
	 */
	@Override
	public final void dragEnter( DragSourceDragEvent event ) {
		setIcon(getConnectedIcon());
		setSelected(true);
	}

	/**
	 * @return true if it is selected.
	 */
	protected final boolean isSelected() {
		return _selected;
	}

	/**
	 * Selects the slot.
	 * @param isSelected true if it is selected.
	 */
	public void setSelected( boolean isSelected ) {
		if (isSelected) {
			_selected = true;
			setActiveBorder();
		} else {
			_selected = false;
			setNormalBorder();
		}
	}

	/**
	 * Called when the cursor's hotspot exits a platform-dependent drop site.
	 * Do nothing.
	 * 
	 * @param event the DragSourceEvent.
	 */
	@Override
	public final void dragExit( DragSourceEvent event ) { }

	/**
	 * Called when the cursor's hotspot moves over a platform-dependent
	 * drop site (a Graph or a Slot).
	 * 
	 * @param event the DragSourceDragEvent.
	 */
	@Override
	public final void dragOver( DragSourceDragEvent event ) { }

	/**
	 * Called when the user has modified the drop gesture.
	 * 
	 * @param event the DragSourceDragEvent.
	 */
	@Override
	public final void dropActionChanged( DragSourceDragEvent event ) {}

	/**
	 * Called while a drag operation is ongoing, when the mouse pointer
	 * enters the operable part of the drop site for the DropTarget
	 * registered with this listener.
	 * Highlights the slot.
	 * 
	 * @param event the DropTargetDragEvent.
	 */
	@Override
	public final void dragEnter(DropTargetDragEvent event) {
		// Accepts the drop action for a move action.
		event.acceptDrag(DnDConstants.ACTION_MOVE);
		setActiveBorder();
	}

	/**
	 * Called while a drag operation is ongoing, when the mouse pointer
	 * has exited the operable part of the drop site for the DropTarget
	 * registered with this listener.
	 * Unhighlights the slot if it is not the source of the DnD operation.
	 * 
	 * @param event the DropTargetEvent.
	 */
	@Override
	public final void dragExit( DropTargetEvent event ) {
		if (!_selected) {
			setNormalBorder();
		}
	}

	/**
	 * Called when a drag operation is ongoing, while the mouse pointer is
	 * still over the operable part of the drop site for the DropTarget
	 * registered with this listener.
	 * 
	 * @param e the DropTargetDragEvent.
	 */
	@Override
	public final void dragOver( DropTargetDragEvent e ) { }

	/**
	 * Called if the user has modified the current drop gesture.
	 * 
	 * @param e the DropTargetDragEvent.
	 */
	@Override
	public final void dropActionChanged( DropTargetDragEvent e ) { }

	/**
	 * Called when the drag operation has terminated with a drop on
	 * the operable part of the drop site for the DropTarget registered
	 * with this listener.
	 * 
	 * @param e the DropTargetDropEvent.
	 */
	@Override
	public abstract void drop( DropTargetDropEvent e );

	/**
	 * Sets the border with active looking.
	 */
	private final void setActiveBorder( ) {
		setBorder(_activeBorder);
	}

	/**
	 * Sets the border with normal looking.
	 */
	protected final void setNormalBorder( ) {
		setBorder(_normalBorder);
	}

	public final void updateBackgroundColor( Color color ) {
		_normalBorder = (LineBorder)BorderFactory.createLineBorder(color,lineWidth);
		if (getBorder() != _activeBorder) {
			setBorder(_normalBorder);
		}
	}

	/**
	 * @return the rank of the slot in the slot list of the graph item.
	 */
	public final int getRank( ) {
		return _rank;
	}

	/**
	 * Return the icon for connected looking of the slot.
	 * 
	 * @return an icon.
	 */
	protected abstract ImageIcon getConnectedIcon( );

	/**
	 * Returns the icon for disconnected looking of the slot.
	 * 
	 * @return an icon.
	 */
	protected abstract ImageIcon getDisconnectedIcon( );

	/**
	 * Determines if the slot is connected.
	 * 
	 * @return true if the item is connected.
	 */
	protected abstract boolean isConnected( );

	/**
	 * Returns the graph item that supports the slot.
	 * 
	 * @return a reference to the graph item.
	 */
	public final GraphItem getGraphItem( ) {
		return _graphItem;
	}

	/**
	 * Returns the related graph.
	 * 
	 * @return a reference to the graph.
	 */
	public final Graph getGraph( ) {
		return _graphItem.getGraph();
	}

	/**
	 * overrides the default print method to print
	 * a filled rectangle.
	 * 
	 * @param gc the graph context.
	 */
	@Override
	public final void print( Graphics gc ) {
		gc.fillRect(0,0,getWidth(),getHeight());
	}

	/**
	 * Returns the first anchor coordinates of the slot.
	 * 
	 * @return anchor point.
	 */
	abstract protected Point getAnchor( );

	/**
	 * Loads a set of connections from an XML tree.
	 * 
	 * @param graph the graph that supports the connections.
	 * @param items the XML tree.
	 * @throws IOException input problems.
	 */
	public static final void load( Graph graph, LinkedList<XMLElement> items ) throws IOException {
		if (items == null) // Empty set.
			return;
		for (XMLElement el : items) {
			if (el.getName().equalsIgnoreCase("Connection")  ||
					el.getName().equalsIgnoreCase("Connexion") ) {
				int idTo=-1, idFrom=-1, rankFrom=-1,rankTo=-1;
				String typeTo = null;
				String typeFrom = null;
				String parameterValue = null;
				ArrayList<Point> path = new ArrayList<Point>();
				Point p = new Point();
				int pos = 0;
				boolean userDefined = false;
				for ( XMLElement elc : el.getChildren()) {;
				if (elc.getName().equalsIgnoreCase("EndItemId") ||
						elc.getName().equalsIgnoreCase("IdElementArrivee"))
					idTo = Integer.parseInt(elc.getValue());
				else if (elc.getName().equalsIgnoreCase("StartItemId") ||
						elc.getName().equalsIgnoreCase("IdElementDepart"))
					idFrom = Integer.parseInt(elc.getValue());
				else if (elc.getName().equalsIgnoreCase("EndSlotType") ||
						elc.getName().equalsIgnoreCase("TypeSlotArrivee"))
					typeTo = elc.getValue();
				else if (elc.getName().equalsIgnoreCase("StartSlotType") ||
						elc.getName().equalsIgnoreCase("TypeSlotDepart"))
					typeFrom = elc.getValue();
				else if (elc.getName().equalsIgnoreCase("StartSlotIndex") ||
						elc.getName().equalsIgnoreCase("NumeroSlotDepart"))
					rankFrom = Integer.parseInt(elc.getValue());
				else if (elc.getName().equalsIgnoreCase("EndSlotIndex") ||
						elc.getName().equalsIgnoreCase("NumeroSlotArrivee"))
					rankTo = Integer.parseInt(elc.getValue());
				else if (elc.getName().equalsIgnoreCase("ParameterValue"))
					parameterValue=elc.getValue();
				else if (elc.getName().equalsIgnoreCase("UserDefined"))
					userDefined = elc.getValue().equalsIgnoreCase("true");
				else if (elc.getName().equalsIgnoreCase("LinkPath")){
					if(elc.getChildren()!=null){
						for (XMLElement pt : elc.getChildren()){
							p = new Point();
							if (pt.getName().equalsIgnoreCase("LinkPoint")){
								for (XMLElement elP : pt.getChildren()) {
									if (elP.getName().equalsIgnoreCase("IdPoint"))
										pos = Integer.parseInt(elP.getValue());
									else if (elP.getName().equalsIgnoreCase("X-coordinate"))
										p.x= Integer.parseInt(elP.getValue());
									else if (elP.getName().equalsIgnoreCase("Y-coordinate"))
										p.y= Integer.parseInt(elP.getValue());
								}
								if (pos > 0)
									path.add(pos,p);
								else 
									path.add(p);
							}
						}
					}
				}
				}
				if (typeTo == null)
					throw new IOException("Unknown connection type");
				if (typeTo.equalsIgnoreCase("InSlotParameter")) { // Parameters
					GraphItem gTo=graph.getGraphItem(idTo);
					if (idFrom > -1){ // Connection parameter -> result
						GraphItem gFrom=graph.getGraphItem(idFrom);
						if ((gFrom != null) && (gTo != null)) {
							OutSlot sFrom = (OutSlot)gFrom.getSlot(typeFrom,rankFrom);
							InSlot sTo = (InSlot)gTo.getSlot(typeTo,rankTo);
							if ((sFrom != null) && (sTo != null)){
								sFrom.connectTo(sTo,true);
								Link l = sTo.createLink();
								l.setPath(path,userDefined);
								sTo.setLink(l);
							}
						}
					}
					if (gTo != null) {  // At least one parameter value.
						InSlotParameter par = gTo.getInSlotParameters().get(rankTo);
						if (par != null)
							par.setValue(parameterValue);
					}
				} else if (typeTo.equalsIgnoreCase("InSlotMask")  // Input or mask images.
						|| typeTo.equalsIgnoreCase("InSlotImage")) {
					GraphItem gFrom=graph.getGraphItem(idFrom);
					GraphItem gTo=graph.getGraphItem(idTo);

					if ((gFrom != null) && (gTo != null)) {
						InSlot sTo = (InSlot)gTo.getSlot(typeTo,rankTo);
						OutSlot sFrom = (OutSlot)gFrom.getSlot(typeFrom,rankFrom);
						if ((sFrom != null) && (sTo != null)){
							sFrom.connectTo(sTo,true);
							Link l = sTo.createLink();
							l.setPath(path,userDefined);
							sTo.setLink(l);
						}
					}
				}
			}
		}
	}
}
