/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;

import javax.swing.JDialog;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTree;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.utils.GDialog;
import fr.greyc.pantheon.ariane.utils.MutableTreeItem;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;
import fr.greyc.pantheon.ariane.xml.XMLFileOutputStream;

/**
 * The class <code>OperatorTreeEditor</code> defines a dialog box
 * where the user can modify the personal operators tree.
 * 
 * @version Régis CLOUARD - Jul 20, 2010
 */
public final class OperatorTreeEditor extends TreeEditor {

	/**
	 * Creates a dialog box that manages the sub tree
	 * of personal operators.
	 * 
	 * @param workspace the related workspace.
	 * @param rootNode the subroutine tree.
	 */
	public OperatorTreeEditor( Workspace workspace, MutableTreeItem rootNode ) {
		super(workspace, Ariane.lang.getString("OperatorTreeEditor.title"), rootNode);
	}

	/**
	 * Triggered when a node is selected in the tree.
	 * Enables/Disables buttons with regard to the selected node.
	 * 
	 * @param path the path to the selected node.
	 */
	@Override
	protected void nodeSelected( TreePath path ) {
		_newButton.setEnabled(path!=null);
		_editButton.setEnabled(path!=null && path.getPathCount()>1);
		_deleteButton.setEnabled(path!=null && path.getPathCount()>1);
	}

	/**
	 * Returns a new instance of the item editor.
	 * 
	 * @param dialog the parent.
	 * @param currentNode the current Node;
	 * @param treeModel the tree model.
	 * @return the dialog.
	 */
	@Override
	protected JDialog getItemEditor( GDialog dialog, MutableTreeItem currentNode, DefaultTreeModel treeModel ) {
		return new OperatorEditor(dialog, currentNode, treeModel);
	}

	/**
	 * Creates and displays a popup menu
	 * for the current selected category in the tree.
	 * 
	 * @param tree the current tree.
	 * @param x the x coordinate position to popup the menu.
	 * @param y the y coordinate position to popup the menu.
	 */
	@Override
	protected void openPopupMenuCategory( final JTree tree, int x, int y ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item = new JMenuItem(Ariane.lang.getString("OperatorTreeEditor.addOperator"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				addAction(tree.getSelectionPath());
			}
		});
		popupMenu.add(item);
		item = new JMenuItem(Ariane.lang.getString("TreeEditor.editCategory"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				modifyAction(tree.getSelectionPath());
			}
		});
		popupMenu.add(item);
		item = new JMenuItem(Ariane.lang.getString("TreeEditor.deleteCategory"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				deleteAction(tree.getSelectionPath());
			}
		});
		popupMenu.add(item);
		popupMenu.show(tree, x, y);
	}

	/**
	 * Creates and displays a popup menu
	 * for the current selected operator in the tree.
	 * 
	 * @param tree the current tree.
	 * @param x the x coordinate position to popup the menu.
	 * @param y the y coordinate position to popup the menu.
	 */
	@Override
	protected void openPopupMenuItem( final JTree tree, int x, int y ) {
		JPopupMenu popupMenu = new JPopupMenu();
		JMenuItem item = new JMenuItem(Ariane.lang.getString("OperatorTreeEditor.addOperator"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				addAction(tree.getSelectionPath().getParentPath());
			}
		});
		popupMenu.add(item);
		item = new JMenuItem(Ariane.lang.getString("TreeEditor.modify"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				modifyAction(tree.getSelectionPath());
			}
		});
		popupMenu.add(item);
		item = new JMenuItem(Ariane.lang.getString("TreeEditor.delete"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				deleteAction(tree.getSelectionPath());
			}
		});
		popupMenu.add(item);
		popupMenu.show(tree, x, y);
	}

	/**
	 * Stores the specified subtree in the current explorer$
	 * tree and in the related file.
	 * 
	 * @param node the root node of the modified subtree.
	 * @return true if all preferences has been saved.
	 */
	@Override
	protected final boolean store( MutableTreeItem node ) {
		Ariane.workspace.getTree().setUserOperatorTree(node);

		File f = new File(Ariane.getPersonalOperatorFilename());
		File d = f.getParentFile();
		try {
			d.mkdirs();
			XMLFileOutputStream out = new XMLFileOutputStream(f,null);
			out.write(saveOperators());
			out.close();
		} catch (Exception e) {
			Ariane.workspace.printErrorMessage(Ariane.lang.getString("PreferenceChooser.error.saved")+" "+Ariane.getPreferencesFilename()+" "+e.getLocalizedMessage());
			return false;
		}
		return true;
	}

	/**
	 * Saves personal operators in an XML tree.
	 * 
	 * @return the XML tree.
	 * @throws IOException in case of IO problems.
	 */
	private static final XMLElement saveOperators() throws IOException {
		XMLElement e = null;
		LinkedList<TreeItem> operatorList=Ariane.workspace.getTree().getUserOperatorList();
		if (operatorList.size() > 0) {
			e = new XMLElement("OperatorLibrary");
			Iterator<TreeItem> it=operatorList.iterator();
			while (it.hasNext())
				e.addChildren(it.next().save());
		}
		return e;
	}
}