/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Locale;
import java.util.Map;
import java.util.StringTokenizer;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.SpinnerModel;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingUtilities;
import javax.swing.ToolTipManager;
import javax.swing.border.EtchedBorder;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructure;
import fr.greyc.pantheon.ariane.graph.item.DataFile;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.utils.GDialog;
import fr.greyc.pantheon.ariane.xml.XMLElement;
import fr.greyc.pantheon.ariane.xml.XMLFileOutputStream;

/**
 * The class <code>Preferences</code> stores the current values
 * for all user's preferences and defines the dialog box
 * that allows the user to change this values.
 * <br>
 * This class sounds like the @link{java.util.prefs.Preferences Preferences class},
 * but the persistence is ensured by our own methods.
 * Each preference is handled using the static get and put methods:
 * <pre>
 * Object value = PreferencesChooser.get("property_name",default_value);
 * PreferencesChooser.set("property_name", value);
 * </pre>
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 30, 2006 (revision)
 * @version Frédéric LAIR - Jun 28, 2007 (add selected Links color)
 * @version Régis CLOUARD - Jun 10, 2010 (improve rendering)
 * @version Régis CLOUARD - Jan 21, 2011 (use map to store preferences)
 * @version Régis CLOUARD - Jan 12, 2013 (add browser and text editor)
 */
public final class PreferenceChooser extends GDialog {
	/** The empty character for the item comment. */
	public static final String EMPTY_STRING="\t";

	/** The temporary file extension. Keep this as variable for sake of efficiency. */
	public static String TEMPORARY_FILE_EXT= ".tmp";

	/** The temporary file extension. Keep this as variable for sake of efficiency. */
	public static String TEMPORARY_FILE_PREFIX= "data";

	public static Locale locale;

	public static Font defaultFont;

	private static int _defaultFontSize;

	/** Stores the last opened tab index. */
	private static int _lastestSelectedTab=0;

	/** The table of colors of the graphical components. */
	private Color[] _colorArray;

	/** The number of colors*/
	private final int _itemCount = 9;

	/** The table of legends of the graphical components. */
	private String[] _colorStringArray;

	/** The current workspace. */
	private Workspace _workspace=null;

	/** The name of the current language. */
	private String _language;

	/** The name of the visualization program. */
	private String _visuCommand;

	/** The name of the status command. */
	private String _statusCommand;

	/** The name of the temporary directory. */
	private String _tempDir;
	/** The file directory. */
	private String _fileDirectory;
	/** The operator binary path. */
	private String _binaryDirectory;

	/** The checkbox for tip options. */
	private final JCheckBox _tipc = new JCheckBox();

	/** The  checkbox for desktop browser. */
	private final JTextField _browser = new JTextField(10);
	private final JTextField _editor = new JTextField(10);

	/** The undo history size. */
	private JSpinner _undoSizeSpinner;

	/** Stores all the preferences. */
	private static Map<String, Object> _preferences = new HashMap<String, Object>();

	/**
	 * Stores the specified value with the specified key
	 * in this preference node.
	 * 
	 * @param key the name of the property.
	 * @param value the property value.
	 */
	public static final void putBoolean ( String key, boolean value) {
		_preferences.put(key, value);
	}

	/**
	 * Returns the preference value for the specified property name.
	 * 
	 * @param key the name of the property.
	 * @return the property value.
	 */
	public static final boolean getBoolean( String key ) {
		return (Boolean)_preferences.get(key);
	}

	/**
	 * Stores the specified value with the specified key
	 * in this preference node.
	 * 
	 * @param key the name of the property.
	 * @param value the property value.
	 */
	public static final void put( String key, String value) {
		_preferences.put(key, value);
	}

	/**
	 * Returns the preference value for the specified property name.
	 * 
	 * @param key the name of the property.
	 * @return the property name.
	 */
	public static final String get( String key ) {
		return (String)_preferences.get(key);
	}

	/**
	 * Stores the specified value with the specified key
	 * in this preference node.
	 * 
	 * @param key the name of the property.
	 * @param value the property value.the dimension
	 */
	public static final void putDimension( String key, Dimension value ) {
		_preferences.put(key, value);
	}

	/**
	 * Returns the preference value for the specified property name.
	 * 
	 * @param key the name of the property.
	 * @return the property name.
	 */
	public static final Dimension getDimension( String key ) {
		return (Dimension)_preferences.get(key);
	}

	/**
	 * Stores the specified value with the specified key
	 * in this preference node.
	 * 
	 * @param key the name of the property.
	 * @param value the coordinates.
	 */
	public static final void putPoint( String key, Point value ) {
		_preferences.put(key, value);
	}

	/**
	 * Returns the preference value for the specified property name.
	 * 
	 * @param key the name of the property.
	 * @return the property name.
	 */
	public static final Point getPoint( String key ) {
		return (Point)_preferences.get(key);
	}

	/**
	 * Stores the specified value with the specified key
	 * in this preference node.
	 * 
	 * @param key the name of the property.
	 * @param value an integer value.
	 */
	public static final void putInt( String key, int value ) {
		_preferences.put(key, value);
	}

	/**
	 * Returns the preference value for the specified property name.
	 * 
	 * @param key the name of the property.
	 * @return the property name.
	 */
	public static final int getInt( String key ) {
		return (Integer)_preferences.get(key);
	}

	public static final String createTemporaryName( int graphIndex, int itemIndex, int slotIndex ) {
		String name= TEMPORARY_FILE_PREFIX
		+ "_"
		+ graphIndex
		+ "_" + itemIndex
		+ "_" + slotIndex + PreferenceChooser.TEMPORARY_FILE_EXT;
		return name;
	}

	/**
	 * Creates a dialog for setting new preferences.
	 * 
	 * @param workspace the related workspace.
	 */
	public PreferenceChooser( Workspace workspace ) {
		super(Ariane.frame, Ariane.lang.getString("PreferenceChooser"), false);
		_workspace = workspace;
		setResizable(false);
		displayContent();
		pack();
		setLocationRelativeTo(Ariane.frame);
		addWindowListener(new WindowAdapter() {
			// Redirects the window closing event to cancelAction();
			@Override
			public final void windowClosing(WindowEvent e){
				cancelAction();
			}
		});
	}

	/**
	 * Saves all the preferences in the given filename.
	 * 
	 * @return true if all preferences has been saved.
	 */
	public static final boolean save() {
		File f = new File(Ariane.getPreferencesFilename());
		File d = f.getParentFile();
		try {
			d.mkdirs();
			XMLFileOutputStream out = new XMLFileOutputStream(f,null);
			out.write(savePreferences());
			out.write(saveRecentFiles());
			out.close();
		} catch (Exception ex) {
			Ariane.workspace.printSystemMessage(Ariane.lang.getString("PreferenceChooser.error.saved")+" "+Ariane.getPreferencesFilename());
			return false;
		}
		return true;
	}

	/**
	 * Loads all preferences from the given XML tree.
	 * 
	 * @param xmlTree the XML tree of preferences.
	 */
	public static final void load( LinkedList<XMLElement> xmlTree ) {
		XMLElement e;

		// Discards tag <ARIANE> from the xmlTree.
		xmlTree = ((xmlTree.getFirst())).getChildren();
		Iterator<XMLElement> it = xmlTree.iterator();

		try {
			while (it.hasNext()) {
				e = it.next();
				if (e.getName().equalsIgnoreCase("Preferences")) {
					loadPreferences(e.getChildren());
				} else if (e.getName().equalsIgnoreCase("RecentFiles")) {
					loadRecentFiles(e.getChildren());
				}
			}
		} catch (Exception ex) {
			//			ex.printStackTrace();
		}
	}

	///// 
	///// F O N T
	///// 

	public static Font fontDialogPLAIN6;
	public static Font fontDialogPLAIN8;
	public static Font fontDialogPLAIN9;
	public static Font fontDialogPLAIN11;
	public static Font fontDialogPLAIN10;
	public static Font fontDialogPLAIN12;
	public static Font fontDialogPLAIN16;
	
	/**
	 * Loads the various fonts.
	 * Sets component with the right font.
	 */
	static private final void loadFonts( ) {
		int screenRes = Toolkit.getDefaultToolkit().getScreenResolution();
		int fontSize = (int)Math.round(12.0 * screenRes / 72.0);
		Font f1 = new Font("Arial", Font.PLAIN, fontSize);
		fontDialogPLAIN6=f1.deriveFont(6f);
		fontDialogPLAIN8=f1.deriveFont(8f);
		fontDialogPLAIN9=f1.deriveFont(9f);
		fontDialogPLAIN10=f1.deriveFont(10.0f);
		fontDialogPLAIN11=f1.deriveFont(11.0f);
		fontDialogPLAIN12=f1.deriveFont(12f);
		fontDialogPLAIN16=f1.deriveFont(16f);
	}

	/**
	 * Sets the default preferences.
	 * If the default language is not represented into
	 * the system, then English is the default language.
	 */
	public static final void initPreferences( ) {
		loadFonts( );
		putPoint("Frame.Position", new Point(-1,-1));
		putDimension("Frame.Size", new Dimension(1000,660));
		putInt("SplitPane.V1.Ratio", 232);
		putInt("SplitPane.V2.Ratio", 584);
		putInt("SplitPane.H1.Ratio", 436);
		putInt("SplitPane.H2.Ratio", 433);
		putInt("CodeGenerator.Preference", 0);
		putBoolean("UseOperatorLibrary", true);
		putBoolean("UseToolTip", true);
		putInt("UndoHistorySize", 200);
		locale = Locale.getDefault();
		if (locale.getDisplayLanguage().equals(Locale.FRENCH.getDisplayLanguage())) {
			setLocale("Français");
		} else {
			setLocale("English");
		}
		setFontSize(10);
	}

	/**
	 * Initializes contents of the dialog.
	 */
	private final void initializePreferences() {
		_colorArray = new Color[_itemCount];
		_colorStringArray = new String[_itemCount];
		int i=0;
		_colorArray[i] = NoteItem.getColor();
		_colorStringArray[i++] = Ariane.lang.getString("PreferenceChooser.note");

		_colorArray[i] = DataFile.getColor();
		_colorStringArray[i++] = Ariane.lang.getString("PreferenceChooser.datafile");

		_colorArray[i] = OperatorItem.getColor();
		_colorStringArray[i++] = Ariane.lang.getString("PreferenceChooser.operator");

		_colorArray[i] = ControlStructure.getColor();
		_colorStringArray[i++] = Ariane.lang.getString("PreferenceChooser.controlStructure");

		_colorArray[i] = CompositeItem.getColor();
		_colorStringArray[i++] = Ariane.lang.getString("PreferenceChooser.compositeItem");

		_colorArray[i] = Graph.getBackgroundColor();
		_colorStringArray[i++] = Ariane.lang.getString("PreferenceChooser.background");

		_colorArray[i] = Graph.getImageLinkColor();
		_colorStringArray[i++] = Ariane.lang.getString("PreferenceChooser.imagelinks");

		_colorArray[i] = Graph.getValueLinkColor();
		_colorStringArray[i++] = Ariane.lang.getString("PreferenceChooser.numericlinks");

		_colorArray[i] = Graph.getMaskLinkColor();
		_colorStringArray[i++] = Ariane.lang.getString("PreferenceChooser.masklinks");

		_language=get("Language");
		_defaultFontSize=getInt("FontSize");
		_visuCommand = GraphItem.getCommandVisu();
		_statusCommand = GraphItem.getCommandStatus();
		_tempDir = Workspace.getTempDirBasename();
		_fileDirectory = get("FileDirectory");
		_binaryDirectory= get("BinaryDirectory");
		_tipc.setSelected(getBoolean("UseToolTip"));
		_browser.setText(get("WebBrowser"));
		_editor.setText(get("TextEditor"));
	}

	/**
	 * Sets the new user's preferences values as the
	 * current preferences.
	 */
	protected final void changePreferences() {
		int i=0;
		NoteItem.setColor(_colorArray[i++]);
		DataFile.setColor(_colorArray[i++]);
		OperatorItem.setColor(_colorArray[i++]);
		ControlStructure.setColor(_colorArray[i++]);
		CompositeItem.setColor(_colorArray[i++]);

		Graph.setBackgroundColor(_colorArray[i++]);
		Graph.setImageLinkColor(_colorArray[i++]);
		Graph.setValueLinkColor(_colorArray[i++]);
		Graph.setMaskLinkColor(_colorArray[i++]);
		changeLanguage(_language);
		setFontSize(_defaultFontSize);
		GraphItem.setCommandVisu(_visuCommand);
		GraphItem.setCommandStatus(_statusCommand);
		Workspace.setTempDir(_tempDir);
		put("FileDirectory", _fileDirectory);
		put("BinaryDirectory", _binaryDirectory);
		putInt("UndoHistorySize", (Integer)_undoSizeSpinner.getValue());
		setToolTip(_tipc.isSelected());
		put("WebBrowser", _browser.getText().trim());
		put("TextEditor", _editor.getText().trim());
	}

	/**
	 * Sets the new state value of the tool tips option.
	 * 
	 * @param state the state.
	 */
	private static final void setToolTip( boolean state ) {
		putBoolean("UseToolTip",state);
		ToolTipManager.sharedInstance().setEnabled(state);
	}

	/**
	 * Changes the current language and changes all texts 
	 * in the User Interface.
	 * 
	 * @param language the new language (Français, English).
	 */
	private static final void changeLanguage( String language ) {
		setLocale(language);
		Ariane.workspace.changeLanguage();
	}

	/**
	 * Changes the current language and changes all texts 
	 * in the User Interface.
	 * 
	 * @param language the new language (Français, English).
	 */
	private static final void setFontSize( int fontSize ) {
		putInt("FontSize",fontSize);
		_defaultFontSize= fontSize;
		switch (fontSize) {
		case 8: defaultFont = fontDialogPLAIN8; break;
		case 9: defaultFont = fontDialogPLAIN9; break;
		case 10: defaultFont = fontDialogPLAIN10; break;
		case 11: defaultFont = fontDialogPLAIN11; break;
		case 12: defaultFont = fontDialogPLAIN12; break;
		default: defaultFont = fontDialogPLAIN10; break;
		}
	}

	/**
	 * Sets the current language to the specified language.
	 * Loads related resources from the resource files.
	 * 
	 * @param language the name of the language.
	 */
	private static final void setLocale( String language ) {
		put("Language",language);
		if (language.equals("Français")) {
			locale = new Locale("fr","FR");
		} else if (language.equals("English")) {
			locale = new Locale("en","US");
		}
		Locale.setDefault(locale);
		Ariane.setLocale(locale);
	}

	/** The tabbed pane that supports the preferences elements. */
	private final JTabbedPane _prefsTabbedPane = new JTabbedPane();

	/**
	 * Initializes the content of the dialog.
	 */
	private final void displayContent() {
		initializePreferences();

		// Creation of the TabbedPane
		//
		// TAB 1 - General.
		//
		{
			Box panel=Box.createVerticalBox();
			JPanel panel1=new JPanel();
			panel.add(panel1);
			panel1.setLayout(new FlowLayout(FlowLayout.LEFT));
			panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PreferenceChooser.general.language")));

			// List of known languages.
			panel1.add(new JLabel(Ariane.lang.getString("PreferenceChooser.general.language.label")));
			String[] languages={"Français","English"};
			JComboBox list = new JComboBox(languages);
			list.setSelectedItem(_language);
			list.addActionListener( new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					JComboBox cb = (JComboBox)e.getSource();
					_language = (String)cb.getSelectedItem();
				}
			});
			panel1.add(list);

			JPanel panel2 = new JPanel(new FlowLayout(FlowLayout.LEFT));
			panel.add( panel2);
			panel2.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PreferenceChooser.general.tooltip")));
			panel2.add(_tipc);
			panel2.add(new JLabel(Ariane.lang.getString("PreferenceChooser.general.tooltip.text")));


			JPanel panel3 = new JPanel(new FlowLayout(FlowLayout.LEFT));
			panel.add(panel3);
			panel3.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(EtchedBorder.RAISED),Ariane.lang.getString("PreferenceChooser.undoHistoryTitle")));
			SpinnerModel spinnerModel;
			spinnerModel = new SpinnerNumberModel(getInt("UndoHistorySize"),10,500,1);
			_undoSizeSpinner = new JSpinner(spinnerModel);
			panel3.add(new JLabel(Ariane.lang.getString("PreferenceChooser.undoHistoryText")));
			panel3.add(_undoSizeSpinner);

			// Browser
			JPanel panel4 = new JPanel(new FlowLayout(FlowLayout.LEFT));
			panel.add( panel4);
			panel4.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PreferenceChooser.general.browser")));
			panel4.add(new JLabel(Ariane.lang.getString("PreferenceChooser.general.browser.text")));
			panel4.add(_browser);
			JButton browser1 = new JButton(Ariane.lang.getString("Ariane.browse"));
			panel4.add(browser1);
			browser1.addActionListener( new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					final JFileChooser fc = new JFileChooser();
					fc.setDialogType(JFileChooser.OPEN_DIALOG);
					fc.setDialogTitle(Ariane.lang.getString("PreferenceChooser.general.selectProgram"));
					fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
					fc.setCurrentDirectory(new File("."));
					int returnVal = fc.showOpenDialog(PreferenceChooser.this);
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						_browser.setText(fc.getSelectedFile().getAbsolutePath());
					}
				}
			});

			// Editor
			JPanel panel5 = new JPanel(new FlowLayout(FlowLayout.LEFT));
			panel.add(panel5);
			panel5.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PreferenceChooser.general.editor")));
			panel5.add(new JLabel(Ariane.lang.getString("PreferenceChooser.general.editor.text")));
			panel5.add(_editor);
			JButton browser2 = new JButton(Ariane.lang.getString("Ariane.browse"));
			panel5.add(browser2);
			browser2.addActionListener( new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					final JFileChooser fc = new JFileChooser();
					fc.setDialogType(JFileChooser.OPEN_DIALOG);
					fc.setDialogTitle(Ariane.lang.getString("PreferenceChooser.general.selectProgram"));
					fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
					fc.setCurrentDirectory(new File("."));
					int returnVal = fc.showOpenDialog(PreferenceChooser.this);
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						_editor.setText(fc.getSelectedFile().getAbsolutePath());
					}
				}
			});

			panel.add(Box.createGlue());
			_prefsTabbedPane.addTab(Ariane.lang.getString("PreferenceChooser.general.title"), panel);
		}
		//
		// TAB 2- Runtime PreferenceChooser.
		//
		{
			Box panel=Box.createVerticalBox();

			// -> VISU command
			JPanel panel1 = new JPanel(new FlowLayout(FlowLayout.LEFT));
			panel.add( panel1);
			panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PreferenceChooser.runtime.visuCommand")));
			final JTextField textfield2 = new JTextField(20);
			panel1.add(new JLabel(Ariane.lang.getString("PreferenceChooser.runtime.CommandName")));

			textfield2.setText(_visuCommand);
			textfield2.addFocusListener(new FocusAdapter() {
				@Override
				public void focusLost(FocusEvent e) {
					_visuCommand = ((JTextField)e.getSource()).getText();
				}
			});
			panel1.add(textfield2);

			// STATUS command
			JPanel panel2 = new JPanel(new FlowLayout(FlowLayout.LEFT));
			panel.add( panel2);
			panel2.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PreferenceChooser.runtime.statusCommand")));
			panel2.add(new JLabel(Ariane.lang.getString("PreferenceChooser.runtime.CommandName")));
			final JTextField textfield3 = new JTextField(20);
			textfield3.setText(_statusCommand);
			textfield3.addFocusListener(new FocusAdapter() {
				@Override
				public void focusLost(FocusEvent e) {
					_statusCommand = (((JTextField)e.getSource()).getText());
				}
			});
			panel2.add(textfield3);

			_prefsTabbedPane.addTab(Ariane.lang.getString("PreferenceChooser.runtime.title"), panel);
		}

		//
		// TAB 3: The graph PreferenceChooser.
		//
		{
			Box panel=Box.createVerticalBox();

			// List of known languages.
			JPanel panel1=new JPanel();
			panel.add(panel1);
			panel1.setLayout(new FlowLayout(FlowLayout.LEFT));
			panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PreferenceChooser.graph.font")));

			panel1.add(new JLabel(Ariane.lang.getString("PreferenceChooser.graphe.font.size")));
			String[] fontSize={"8","9", "10", "11", "12"};
			JComboBox list = new JComboBox(fontSize);
			list.setSelectedItem(""+_defaultFontSize);
			list.addActionListener( new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					JComboBox cb = (JComboBox)e.getSource();
					_defaultFontSize= Integer.parseInt((String)cb.getSelectedItem());
				}
			});
			panel1.add(list);
			panel.add(panel1);
			
			for (int i = 5; i < _itemCount; i++) {
				JPanel panel2 = new JPanel(new FlowLayout(FlowLayout.LEFT));
				panel.add(panel2);
				panel2.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), _colorStringArray[i]));
				panel2.add(new BindingColor(Ariane.lang.getString("PreferenceChooser.backgroundColor"), i));
			}

			_prefsTabbedPane.addTab(Ariane.lang.getString("PreferenceChooser.graph.title"), panel);
		}

		//
		// TAB 4- Item color.
		//
		{
			Box panel=Box.createVerticalBox();
			for (int i = 0; i < 5; i++) {
				JPanel panel1 = new JPanel(new FlowLayout(FlowLayout.LEFT));
				panel.add(panel1);
				panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), _colorStringArray[i]));
				panel1.add(new BindingColor(Ariane.lang.getString("PreferenceChooser.backgroundColor"), i));
			}

			_prefsTabbedPane.addTab(Ariane.lang.getString("PreferenceChooser.item.title"), panel);
		}

		//
		// TAB 5- The working paths.
		//
		{
			Box panel=Box.createVerticalBox();
			{
				// Temporary paths
				JPanel panel3=new JPanel();
				panel.add(panel3);
				panel3.setLayout(new FlowLayout(FlowLayout.LEFT));
				panel3.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PreferenceChooser.path.temporaryDirectory")));
				JButton button = new JButton(Ariane.lang.getString("Ariane.browse"));
				final JTextField textfield3 = new JTextField(20);
				button.addActionListener( new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent e) {
						final JFileChooser fc = new JFileChooser();
						fc.setDialogType(JFileChooser.OPEN_DIALOG);
						fc.setDialogTitle(Ariane.lang.getString("FileChooser.titleSelectDir"));
						fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
						fc.setCurrentDirectory(new File(Workspace.getTempDirBasename()));
						int returnVal = fc.showOpenDialog(PreferenceChooser.this);
						if (returnVal == JFileChooser.APPROVE_OPTION) {
							// New directory
							_tempDir = fc.getSelectedFile().getAbsolutePath();
							textfield3.setText(_tempDir);
						}
					}
				});
				textfield3.setText(_tempDir);
				textfield3.addFocusListener(new FocusAdapter() {
					@Override
					public void focusLost(FocusEvent e) {
						_tempDir = ((JTextField)e.getSource()).getText();
					}
				});
				panel3.add(textfield3);
				panel3.add(button);
			}
			// -> Temporary file extension
			{
				JPanel panel2 = new JPanel(new FlowLayout(FlowLayout.LEFT));
				panel.add( panel2);
				panel2.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PreferenceChooser.runtime.TemporaryFileExtension")));
				JTextField textfield4 = new JTextField(20);
				panel2.add(new JLabel(Ariane.lang.getString("PreferenceChooser.runtime.fileExtension")));

				textfield4.setText(TEMPORARY_FILE_EXT);
				textfield4.addFocusListener(new FocusAdapter() {
					@Override
					public void focusLost(FocusEvent e) {
						TEMPORARY_FILE_EXT = ((JTextField)e.getSource()).getText();
					}
				});
				panel2.add(textfield4);
				panel.add(Box.createGlue());

				JPanel panel1 = new JPanel(new FlowLayout(FlowLayout.LEFT));
				panel.add( panel1);
				panel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PreferenceChooser.environment.title")));
				// -> Remove temporary files
				panel1.add(new JLabel(Ariane.lang.getString("PreferenceChooser.environment.cleaning")));
				JButton button = new JButton(Ariane.lang.getString("PreferenceChooser.environment.erasing"));
				button.addActionListener( new ActionListener() {
					// Erases all temporary files.
					@Override
					public void actionPerformed(ActionEvent e) {
						int j = Workspace.deleteTemporaryFile();
						if (j>0) {
							JOptionPane.showMessageDialog(PreferenceChooser.this,
									j+Ariane.lang.getString("PreferenceChooser.message1"),
									Ariane.lang.getString("notice"),
									JOptionPane.INFORMATION_MESSAGE);
						}
					}
				} );
				panel1.add(button);
				panel.add(Box.createGlue());
			}
			{
				// File directory
				JPanel panel3=new JPanel();
				panel.add(panel3);
				panel3.setLayout(new FlowLayout(FlowLayout.LEFT));
				panel3.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PreferenceChooser.path.fileDirectory")));
				JButton button = new JButton(Ariane.lang.getString("Ariane.browse"));
				final JTextField textfield3 = new JTextField(20);
				button.addActionListener( new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent e) {
						final JFileChooser fc = new JFileChooser();
						fc.setDialogType(JFileChooser.OPEN_DIALOG);
						fc.setDialogTitle(Ariane.lang.getString("FileChooser.titleSelectDir"));
						fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
						if (_fileDirectory!=null )
							fc.setCurrentDirectory(new File(_fileDirectory));
						int returnVal = fc.showOpenDialog(PreferenceChooser.this);
						if (returnVal == JFileChooser.APPROVE_OPTION) {
							_fileDirectory = fc.getSelectedFile().getAbsolutePath();
							textfield3.setText(_fileDirectory);
						}
					}
				});
				textfield3.setText(_fileDirectory);
				textfield3.addFocusListener(new FocusAdapter() {
					@Override
					public void focusLost(FocusEvent e) {
						_fileDirectory = ((JTextField)e.getSource()).getText();
						if (_fileDirectory.trim().length()==0) _fileDirectory=null;
					}
				});
				panel3.add(textfield3);
				panel3.add(button);
			}
			{
				// Personal operator paths
				JPanel panel3=new JPanel();
				panel.add(panel3);
				panel3.setLayout(new FlowLayout(FlowLayout.LEFT));
				panel3.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("PreferenceChooser.path.binaryDirectory")));
				JButton button = new JButton(Ariane.lang.getString("Ariane.browse"));
				final JTextField textfield3 = new JTextField(20);
				button.addActionListener( new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent e) {
						final JFileChooser fc = new JFileChooser();
						fc.setDialogType(JFileChooser.OPEN_DIALOG);
						fc.setDialogTitle(Ariane.lang.getString("FileChooser.titleSelectDir"));
						fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
						if (_binaryDirectory!=null)
							fc.setCurrentDirectory(new File(_binaryDirectory));
						int returnVal = fc.showOpenDialog(PreferenceChooser.this);
						if (returnVal == JFileChooser.APPROVE_OPTION) {
							_binaryDirectory= fc.getSelectedFile().getAbsolutePath();
							textfield3.setText(_binaryDirectory);
						}
					}
				});
				textfield3.setText(_binaryDirectory);
				textfield3.addFocusListener(new FocusAdapter() {
					@Override
					public void focusLost(FocusEvent e) {
						_binaryDirectory= ((JTextField)e.getSource()).getText();
						if (_binaryDirectory.trim().length()==0) _binaryDirectory=null;
					}
				});
				panel3.add(textfield3);
				panel3.add(button);
			}
			_prefsTabbedPane.addTab(Ariane.lang.getString("PreferenceChooser.paths"), panel);
			_prefsTabbedPane.setSelectedIndex(_lastestSelectedTab);
		}

		// Buttons panel
		final JPanel buttonPanel = new JPanel();
		JButton applyButton = new JButton(Ariane.lang.getString("PreferenceChooser.applyText"),
				new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"gtk-apply.png")));
		applyButton.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) {
				buttonPanel.requestFocusInWindow();
				// Tip: handle event before this event.
				SwingUtilities.invokeLater(new Runnable() {
					@Override
					public void run() {
						applyAction();
					}
				});
			}
		} );

		// cancel button
		JButton cancelButton = new JButton();
		cancelButton.setAction(new AbstractAction(Ariane.lang.getString("TreeEditor.cancel"),
				new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"gtk-cancel.png"))) {
			@Override
			public final void actionPerformed(ActionEvent e) {
				cancelAction();
			}
		});

		JButton helpButton = new JButton();
		helpButton.setAction(new AbstractAction(null,new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"help.png"))) {
			@Override
			public final void actionPerformed(ActionEvent e) {
				Ariane.workspace.displayMan("Ariane","manual.html", "#tab"+_prefsTabbedPane.getSelectedIndex());
			}
		});
		buttonPanel.add(helpButton);
		buttonPanel.add(Box.createHorizontalStrut(80));
		buttonPanel.add(applyButton);
		buttonPanel.add(cancelButton);
		setDefaultButton(applyButton);
		setCancelButton(cancelButton);

		getContentPane().setLayout(new BorderLayout());
		getContentPane().add(buttonPanel, BorderLayout.SOUTH);
		getContentPane().add(_prefsTabbedPane, BorderLayout.CENTER);
		pack();
	}

	/**
	 * Closes the preferences dialog box.
	 */
	private final void cancelAction() {
		setVisible(false);
		dispose();
	}

	/**
	 * Applies and saves the chosen preferences,
	 * and closes the dialog box.
	 */
	private final void applyAction() {
		changePreferences();
		_workspace.repaintAllGraphs();
		save();
		setVisible(false);
		_lastestSelectedTab=_prefsTabbedPane.getSelectedIndex();
		dispose();
	}

	/**
	 * Codes a color as a triplet: r,g,b.
	 * 
	 * @param color the color to be coded.
	 * @return the color encoded as a string.
	 */
	private static final String code( Color color ) {
		return color.getRed()+","+color.getGreen()+","+color.getBlue();
	}

	/**
	 * Decodes a color as a string to a <code>Color</code>.
	 * Color is given as a triplet r,g,b.
	 * 
	 * @param color the string to be uncoded.
	 * @return the <code>Color</code>.
	 */
	private static final Color decode( String color ) {
		StringTokenizer st = new StringTokenizer(color,",");

		int r = Integer.parseInt(st.nextToken());
		int v = Integer.parseInt(st.nextToken());
		int b = Integer.parseInt(st.nextToken());
		return new Color(r,v,b);
	}

	/**
	 * Saves the preferences as an XML tree.
	 * 
	 * @return the XML tree.
	 * @throws IOException in case of IO problems.
	 */
	private static final XMLElement savePreferences( ) throws IOException {
		XMLElement e = new XMLElement("Preferences");
		e.addChildren(new XMLElement("CurrentDir",Workspace.getCurrentDir()));
		if (!Workspace.getDefaultTempDir().equals(Workspace.getTempDirBasename())) {
			e.addChildren(new XMLElement("TempDir",Workspace.getTempDirBasename()));
		}
		e.addChildren(new XMLElement("PandoreDir",Ariane.getLibraryPath()));
		if (GraphItem.getCommandVisu().trim().length()>0) {
			e.addChildren(new XMLElement("Command.Visu",GraphItem.getCommandVisu()));
		}
		if (GraphItem.getCommandStatus().trim().length()>0) {
			e.addChildren(new XMLElement("Command.Status",GraphItem.getCommandStatus()));
		}
		if (!NoteItem.getDefaultColor().equals(NoteItem.getColor()))
			e.addChildren(new XMLElement("Note.Color",code(NoteItem.getColor())));
		if (!OperatorItem.getDefaultColor().equals(OperatorItem.getColor()))
			e.addChildren(new XMLElement("Operator.Color",code(OperatorItem.getColor())));
		if (!DataFile.getDefaultColor().equals(DataFile.getColor()))
			e.addChildren(new XMLElement("Datafile.Color",code(DataFile.getColor())));
		if (!ControlStructure.getDefaultColor().equals(ControlStructure.getColor()))
			e.addChildren(new XMLElement("Control.Color",code(ControlStructure.getColor())));
		if (!CompositeItem.getDefaultColor().equals(CompositeItem.getColor()))
			e.addChildren(new XMLElement("Module.Color",code(CompositeItem.getColor())));
		if (!Graph.getDefaultBackgroundColor().equals(Graph.getBackgroundColor()))
			e.addChildren(new XMLElement("Graph.Background.Color",code(Graph.getBackgroundColor())));
		if (!Graph.getDefaultImageLinkColor().equals(Graph.getImageLinkColor()))
			e.addChildren(new XMLElement("Image.Link.Color",code(Graph.getImageLinkColor())));
		if (!Graph.getDefaultValueLinkColor().equals(Graph.getValueLinkColor()))
			e.addChildren(new XMLElement("Value.Link.Color",code(Graph.getValueLinkColor())));
		if (!Graph.getDefaultMaskLinkColor().equals(Graph.getMaskLinkColor()))
			e.addChildren(new XMLElement("Mask.Link.Color",code(Graph.getMaskLinkColor())));
		e.addChildren(new XMLElement("Language",get("Language")));
		e.addChildren(new XMLElement("FontSize",""+getInt("FontSize")));
		e.addChildren(new XMLElement("ToolTip",""+getBoolean("UseToolTip")));
		String web = get("WebBrowser");
		if (web!=null &&web.trim().length()>0)
			e.addChildren(new XMLElement("WebBrowser",""+web));
		String editor = get("TextEditor");
		if (editor!=null &&editor.trim().length()>0)
			e.addChildren(new XMLElement("TextEditor",""+editor));
		e.addChildren(new XMLElement("Frame.Width",""+Ariane.frame.getSize().width));
		e.addChildren(new XMLElement("Frame.Height",""+Ariane.frame.getSize().height));
		e.addChildren(new XMLElement("Frame.X",""+Ariane.frame.getLocation().x));
		e.addChildren(new XMLElement("Frame.Y",""+Ariane.frame.getLocation().y));
		e.addChildren(new XMLElement("SplitH1.ratio",""+Ariane.workspace.getSplitH1()));
		e.addChildren(new XMLElement("SplitH2.ratio",""+Ariane.workspace.getSplitH2()));
		e.addChildren(new XMLElement("SplitV1.ratio",""+Ariane.workspace.getSplitV1()));
		e.addChildren(new XMLElement("SplitV2.ratio",""+Ariane.workspace.getSplitV2()));
		e.addChildren(new XMLElement("UseOperatorLibrary",""+getBoolean("UseOperatorLibrary")));
		e.addChildren(new XMLElement("TemporaryFileExtension",TEMPORARY_FILE_EXT));
		e.addChildren(new XMLElement("UndoHistorySize",""+getInt("UndoHistorySize")));
		e.addChildren(new XMLElement("CodeGenerator.Preference",""+getInt("CodeGenerator.Preference")));
		if (get("FileDirectory")!=null) {
			e.addChildren(new XMLElement("FileDirectory",""+get("FileDirectory")));
		}
		if (get("BinaryDirectory")!=null) {
			e.addChildren(new XMLElement("BinaryDirectory",""+get("BinaryDirectory")));
		}
		return e;
	}

	/**
	 * Saves the recent files as an XML tree.
	 * 
	 * @return the XML tree.
	 * @throws IOException in case of IO problems.
	 */
	private static final XMLElement saveRecentFiles() throws IOException {
		XMLElement e = new XMLElement("RecentFiles");
		for (String filename : Workspace.getRecentProjectFilenames()) {
			e.addChildren(new XMLElement("File",filename));
		}
		return e;
	}

	/**
	 * Sets all preferences in the current project.
	 * 
	 * @param xmlTree the XML tree of preferences.
	 * @throws IOException in case of IO problems.
	 */
	private static final void loadPreferences( LinkedList<XMLElement> xmlTree ) throws IOException {
		XMLElement e;
		Iterator<XMLElement> it = xmlTree.iterator();
		while(it.hasNext()) {
			e = it.next();
			// A l'interieur de <Preferences>
			if (e.getName().equalsIgnoreCase("CurrentDir"))
				Workspace.setCurrentDir(e.getValue());
			else if (e.getName().equalsIgnoreCase("TempDir"))
				Workspace.setTempDir(e.getValue());
			else if (e.getName().equalsIgnoreCase("PandoreDir"))
				Ariane.setLibraryPath(e.getValue());
			else if (e.getName().equalsIgnoreCase("Command.Visu"))
				GraphItem.setCommandVisu(e.getValue());
			else if (e.getName().equalsIgnoreCase("Command.Status"))
				GraphItem.setCommandStatus(e.getValue());
			else if (e.getName().equalsIgnoreCase("Note.Color"))
				NoteItem.setColor(decode(e.getValue()));
			else if (e.getName().equalsIgnoreCase("Operator.Color"))
				OperatorItem.setColor(decode(e.getValue()));
			else if (e.getName().equalsIgnoreCase("Datafile.Color"))
				DataFile.setColor(decode(e.getValue()));
			else if (e.getName().equalsIgnoreCase("Control.Color"))
				ControlStructure.setColor(decode(e.getValue()));
			else if (e.getName().equalsIgnoreCase("Composite.Color"))
				CompositeItem.setColor(decode(e.getValue()));
			else if (e.getName().equalsIgnoreCase("Graph.Background.Color"))
				Graph.setBackgroundColor(decode(e.getValue()));
			else if (e.getName().equalsIgnoreCase("Image.Link.Color"))
				Graph.setImageLinkColor(decode(e.getValue()));
			else if (e.getName().equalsIgnoreCase("Value.Link.Color")) {
				Graph.setValueLinkColor(decode(e.getValue()));
			} else if (e.getName().equalsIgnoreCase("Mask.Link.Color")) {
				Graph.setMaskLinkColor(decode(e.getValue()));
			} else if (e.getName().equalsIgnoreCase("Language")) {
				setLocale(e.getValue());
			} else if (e.getName().equalsIgnoreCase("FontSize")) {
				int size = Integer.parseInt(e.getValue());
				setFontSize(size);
			} else if (e.getName().equalsIgnoreCase("ToolTip")) {
				setToolTip(Boolean.parseBoolean(e.getValue()));
			} else if (e.getName().equalsIgnoreCase("WebBrowser")) {
				put("WebBrowser", e.getValue());
			} else if (e.getName().equalsIgnoreCase("TextEditor")) {
				put("TextEditor", e.getValue());
			} else if (e.getName().equalsIgnoreCase("Frame.Width")) {
				Dimension d= getDimension("Frame.Size");
				d.width = Integer.parseInt(e.getValue());
				putDimension("Frame.Size", d);
			} else if (e.getName().equalsIgnoreCase("Frame.Height")) {
				Dimension d= getDimension("Frame.Size");
				d.height = Integer.parseInt(e.getValue());
				putDimension("Frame.Size", d);
			} else if (e.getName().equalsIgnoreCase("Frame.X")) {
				Point p= getPoint("Frame.Position");
				p.x = Integer.parseInt(e.getValue());
				putPoint("Frame.Position", p);
			} else if (e.getName().equalsIgnoreCase("Frame.Y")) {
				Point p = getPoint("Frame.Position");
				p.y = Integer.parseInt(e.getValue());
				putPoint("Frame.Position", p);
			} else if (e.getName().equalsIgnoreCase("SplitH1.ratio")) {
				putInt("SplitPane.H1.Ratio", Integer.parseInt(e.getValue()));
			} else if (e.getName().equalsIgnoreCase("SplitH2.ratio")) {
				putInt("SplitPane.H2.Ratio", Integer.parseInt(e.getValue()));
			} else if (e.getName().equalsIgnoreCase("SplitV1.ratio")) {
				putInt("SplitPane.V1.Ratio", Integer.parseInt(e.getValue()));
			} else if (e.getName().equalsIgnoreCase("SplitV2.ratio")) {
				putInt("SplitPane.V2.Ratio", Integer.parseInt(e.getValue()));
			} else if (e.getName().equalsIgnoreCase("UseOperatorLibrary")) {
				putBoolean("UseOperatorLibrary", Boolean.parseBoolean(e.getValue()));
			} else if (e.getName().equalsIgnoreCase("TemporaryFileExtension")) {
				TEMPORARY_FILE_EXT=e.getValue();
			} else if (e.getName().equalsIgnoreCase("UndoHistorySize")) {
				putInt("UndoHistorySize", Integer.parseInt(e.getValue()));
			} else if (e.getName().equalsIgnoreCase("CodeGenerator.Preference")) {
				putInt("CodeGenerator.Preference", Integer.parseInt(e.getValue()));
			} else if (e.getName().equalsIgnoreCase("FileDirectory")) {
				e.addChildren(new XMLElement("FileDirectory",""+get("FileDirectory")));
				put("FileDirectory", e.getValue());
			} else if (e.getName().equalsIgnoreCase("BinaryDirectory")) {
				e.addChildren(new XMLElement("BinaryDirectory",""+get("BinaryDirectory")));
				put("BinaryDirectory", e.getValue());
			}
		}
	}

	/**
	 * Gets the list of recent files.
	 * 
	 * @param xmlTree the XML tree of preferences.
	 * @throws IOException in case of IO problems.
	 */
	private static final void loadRecentFiles( LinkedList<XMLElement> xmlTree ) throws IOException {
		XMLElement e;
		Iterator<XMLElement> it = xmlTree.iterator();
		while(it.hasNext()) {
			e = it.next();
			if (e.getName().equalsIgnoreCase("File")) {
				Workspace.getRecentProjectFilenames().add(e.getValue());
			}
		}
	}

	/**
	 * The class <code>BindingColor</code> is an inner class
	 * that represents a pair (item, color).
	 * It provides a pane of controls designed to allow a user
	 * to choose a color for a given item.
	 */
	private final class BindingColor extends JPanel {
		/**  The index of the color of the item. */
		private final int _numColor;
		/**  The text that describes the item. */
		private final String _legend;
		/** Chosen button. */
		private final JPanel _colorButton;

		/**
		 * Creates the pair (item/color).
		 * 
		 * @param legend the legend of the item.
		 * @param numColor the index of the color for the given item.
		 */
		public BindingColor( final String legend, int numColor ) {
			_legend = legend;
			_numColor = numColor;
			_colorButton = new JPanel();
			_colorButton.setPreferredSize( new Dimension(30,30));
			_colorButton.setBorder( BorderFactory.createEtchedBorder());
			_colorButton.setBackground(_colorArray[_numColor]);
			_colorButton.addMouseListener(new MouseAdapter() {
				@Override
				public void mouseClicked( final MouseEvent ev) {
					changeCol(ev.getComponent()); }
			});
			setLayout(new BorderLayout());
			JLabel label=new JLabel(_legend+"   ");
			label.setFont(new Font("Dialog", Font.PLAIN, 12));
			add(label, BorderLayout.CENTER);
			add(_colorButton, BorderLayout.EAST);
		}

		/**
		 * Shows a <code>JColorChooser</code> to select
		 * a color for the related item.
		 * 
		 * @param item the item.
		 */
		private final void changeCol( Component item ) {
			Color col = JColorChooser.showDialog(item, _legend, _colorArray[_numColor]);
			if (col != null) {
				_colorArray[_numColor] = col;
				_colorButton.setBackground(col);
				_colorButton.repaint();
			}
		}
	}
}