/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.utils.GDialog;
import fr.greyc.pantheon.ariane.utils.MutableTreeItem;
import fr.greyc.pantheon.ariane.utils.TreeItem;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.AbstractAction;
import javax.swing.Box;
import javax.swing.ImageIcon;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.border.EmptyBorder;
import javax.swing.border.LineBorder;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

/**
 * The class <code>TreeEditor</code> defines a dialog box
 * that allows users to modify the tree of items
 * (subroutine or operator subtree).
 * 
 * @version Régis CLOUARD - Jul 20, 2010
 */
public abstract class TreeEditor extends GDialog implements MouseListener {
	/** The workspace. */
	private Workspace _workspace;

	/** The root of the current tree of subroutines. */
	private MutableTreeItem _rootNode;

	/** The visual representation of the tree. */
	private JTree _jtree;

	/** The model of the tree. */
	private DefaultTreeModel _treeModel;

	/** The current selected node. */
	private MutableTreeItem _currentNode=null;

	/** The new button. */
	protected JButton _newButton;
	/** The dit button. */
	protected JButton _editButton;
	/** The delete button. */
	protected JButton _deleteButton;
	
	/**
	 * Creates a dialog box that manages subtree of the explorer tree.
	 * 
	 * @param workspace the related workspace.
	 * @param title the title of the dialog box.
	 * @param rootNode the subtree root.
	 */
	public TreeEditor( Workspace workspace, String title, MutableTreeItem rootNode ) {
		super((JFrame)null, title, true);
		setResizable(false);
		_workspace = workspace;
		// Use a copy of the existing tree.
		_rootNode = (MutableTreeItem)rootNode.clone();
		_treeModel = new DefaultTreeModel(_rootNode);
		addWindowListener(new WindowAdapter() {
			// Redirects the window closing event to cancelAction();
			@Override
			public final void windowClosing(WindowEvent e) {
				cancelAction();
			}
		});

		_jtree = new JTree(_treeModel);
		_jtree.getInputMap().put(KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_DELETE, 0), "Delete");
		_jtree.getActionMap().put("Delete", new AbstractAction(null) {
			@Override
			public final void actionPerformed(ActionEvent e) {
				deleteAction(_jtree.getSelectionPath());
			}
		});
		displayContent();
		setLocationRelativeTo(workspace);
	}

	/**
	 * Creates the content of the dialog box.
	 */
	public final void displayContent( ) {
		setSize(300,400);
		_jtree.setEditable(false);
		_jtree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
		_jtree.setShowsRootHandles(true);
		_jtree.addMouseListener(this);
		_jtree.setBorder(new LineBorder(Color.BLACK));

		JPanel scrollPane = new JPanel(new BorderLayout());
		scrollPane.add(new JScrollPane(_jtree), BorderLayout.CENTER);
		scrollPane.setBorder(new EmptyBorder(10,10,10,10));

		_editButton = new JButton(Ariane.lang.getString("TreeEditor.edit"));
		_editButton.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed( ActionEvent e ) {
				modifyAction(_jtree.getSelectionPath());
			}
		} );
		_editButton.setEnabled(false);
		_newButton = new JButton(Ariane.lang.getString("TreeEditor.new"));
		_newButton.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed( ActionEvent e ) {
				addAction(_jtree.getSelectionPath());
			}
		} );
		_newButton.setEnabled(false);
		_deleteButton = new JButton(Ariane.lang.getString("TreeEditor.delete"));
		_deleteButton.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed( ActionEvent e ) {
				deleteAction(_jtree.getSelectionPath());
			}
		} );
		_deleteButton.setEnabled(false);

		JPanel buttonPanel1 = new JPanel();
		buttonPanel1.add(_editButton);
		buttonPanel1.add(_newButton);
		buttonPanel1.add(_deleteButton);

		// save button
		JButton saveButton = new JButton(Ariane.lang.getString("TreeEditor.save"),
				new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"gtk-apply.png")));
		saveButton.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed( ActionEvent e ) {
				saveAction();
			}
		} );
		// cancel button
		JButton cancelButton = new JButton();
		cancelButton.setAction(new AbstractAction(Ariane.lang.getString("TreeEditor.cancel"),
				new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"gtk-cancel.png"))) {
			@Override
			public final void actionPerformed(ActionEvent e) {
				cancelAction();
			}
		});
		Box buttonPanel = Box.createVerticalBox();
		buttonPanel.add(buttonPanel1);
		JPanel buttonPanel2 = new JPanel();
		buttonPanel2.add(saveButton);
		buttonPanel2.add(cancelButton);
		buttonPanel.add(buttonPanel2);
		setDefaultButton(saveButton);
		setCancelButton(cancelButton);

		getContentPane().setLayout(new BorderLayout());
		getContentPane().add(scrollPane, BorderLayout.CENTER);
		getContentPane().add(buttonPanel, BorderLayout.SOUTH);
	}

	/**
	 * Overrides <code>JDialog.dispose()</code>
	 * in order to free all memory.
	 */
	@Override
	public final void dispose( ) {
		_workspace=null;
		_jtree=null;
		_treeModel=null;
		_currentNode=null;
		_rootNode=null;
		super.dispose();
	}

	/**
	 * Invoked when the mouse button has been clicked
	 * (pressed and released) on an tree item.
	 * Opens the dialog box to modify the item in case
	 * of double click.
	 * 
	 * @param event the mouse event.
	 */
	@Override
	public final void mouseClicked( MouseEvent event ) {
		if (event.getClickCount()>1){
			modifyAction(_jtree.getSelectionPath());
		} else {
			nodeSelected(_jtree.getSelectionPath());
		}
	}

	/**
	 * Invoked when the mouse enters a component.
	 * @param e an event.
	 */
	@Override
	public final void mouseEntered( MouseEvent e ) { }

	/**
	 * Invoked when the mouse exits a component.
	 * @param e an event.
	 */
	@Override
	public final void mouseExited( MouseEvent e ) { }

	/**
	 * Invoked when a mouse button has been pressed
	 * on a component.
	 * Opens a menu to add a subitem to the selected item 
	 * or delete the selected item (Unix systems).
	 * 
	 * @param e the mouse event.
	 */
	@Override
	public final void mousePressed( MouseEvent e ) {
		if (e.isPopupTrigger()){
			openPopupMenu(e.getX(), e.getY());
		}
	}

	/**
	 * Invoked when a mouse button has been released
	 * on a component.
	 * Opens a menu to add a subitem to the selected item 
	 * or delete the selected item (Windows systems).
	 * 
	 * @param e the mouse event.
	 */
	@Override
	public final void mouseReleased( MouseEvent e ) {
		if (e.isPopupTrigger()){
			openPopupMenu(e.getX(), e.getY());
		}
	}

	/**
	 * Opens the popup related to the selected item.
	 * 
	 * @param x the x coordinate position to popup the menu.
	 * @param y the y coordinate position to popup the menu.
	 */
	private final void openPopupMenu( int x, int y ) {
		MutableTreeItem selectedNode=(MutableTreeItem)_jtree.getLastSelectedPathComponent();
		if (selectedNode!=null) {
			switch(selectedNode.getLevel()) {
			case 0: openPopupMenuRoot(x,y); break;
			case 1: openPopupMenuCategory(_jtree, x,y); break;
			case 2: openPopupMenuItem(_jtree, x,y); break;
			}
		}
	}

	/**
	 * Creates and displays a popup menu
	 * for the current selected root in the tree.
	 * 
	 * @param x the x coordinate position to popup the menu.
	 * @param y the y coordinate position to popup the menu.
	 */
	private final void openPopupMenuRoot( int x, int y ) {
		JPopupMenu popupMenu = new JPopupMenu();
		JMenuItem item = new JMenuItem(Ariane.lang.getString("TreeEditor.addCategory"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				addAction(_jtree.getSelectionPath());
			}
		});
		popupMenu.add(item);
		popupMenu.show(_jtree, x, y);
	}

	/**
	 * Creates and displays a popup menu
	 * for the current selected category in the tree.
	 * 
	 * @param jtree the current tree.
	 * @param x the x coordinate position to popup the menu.
	 * @param y the y coordinate position to popup the menu.
	 */
	protected abstract void openPopupMenuCategory( final JTree jtree, int x, int y );

	/**
	 * Creates and displays a popup menu
	 * for the current selected item in the tree.
	 * 
	 * @param jtree the current tree.
	 * @param x the x coordinate position to popup the menu.
	 * @param y the y coordinate position to popup the menu.
	 */
	protected abstract void openPopupMenuItem( final JTree jtree, int x, int y );

	///// 
	///// AC T I O N S
	///// 

	/**
	 * Triggered when a node is selected in the tree.
	 * 
	 * @param path the path to the selected node.
	 */
	protected abstract void nodeSelected( TreePath path );

	/**
	 * Adds a new item below the specified item.
	 * 
	 * @param currentSelection the specified item.
	 */
	protected final void addAction( TreePath currentSelection ) {
		MutableTreeItem currentNode;
		if (currentSelection == null)
			currentNode = _rootNode;
		else
			currentNode = (MutableTreeItem)currentSelection.getLastPathComponent();

		if (currentNode == _rootNode){ // Create a category
			_currentNode = new MutableTreeItem("category",true);
			_treeModel.insertNodeInto(_currentNode, currentNode, currentNode.getChildCount());
			_jtree.expandPath(currentSelection);
			CategoryEditor editor = new CategoryEditor(this);
			editor.setVisible(true);
		} else { // Create an operator.
			MutableTreeItem categoryNode;
			if (currentNode.getAllowsChildren()) {
				categoryNode = currentNode;
			} else {
				categoryNode = (MutableTreeItem)currentNode.getParent();
			}
			TreeItem newItem = new TreeItem(TreeItem.Item.OPERATOR, "operator","", null, (String)categoryNode.getUserObject(),0,0,0,"");
			_currentNode = new MutableTreeItem(newItem,false);
			_treeModel.insertNodeInto(_currentNode, categoryNode, categoryNode.getChildCount());
			_jtree.expandPath(currentSelection);
			JDialog editor = getItemEditor(this, _currentNode, _treeModel);
			editor.setVisible(true);
		}
	}

	/**
	 * Modifies the specified item. It means open
	 * a dialog box to specify the new properties.
	 * 
	 * @param currentSelection the specified item.
	 */
	protected final void modifyAction( TreePath currentSelection ) {
		if (currentSelection != null) {
			_currentNode = (MutableTreeItem)currentSelection.getLastPathComponent();
			if (_currentNode.getLevel()<1) return;
			if (_currentNode.getUserObject() instanceof TreeItem) {
				JDialog editor = getItemEditor(this, _currentNode, _treeModel);
				editor.setVisible(true);
			} else {
				CategoryEditor editor = new CategoryEditor(this);
				editor.setVisible(true);
			}
		}
	}

	/**
	 * Returns a new instance of the item editor.
	 * 
	 * @param dialog the parent.
	 * @param currentNode the current Node;
	 * @param treeModel the tree model.
	 * @return the dialog.
	 */
	protected abstract JDialog getItemEditor( GDialog dialog, MutableTreeItem currentNode, DefaultTreeModel treeModel );

	/**
	 * Deletes the specified item.
	 * 
	 * @param currentSelection the specified item.
	 */
	protected final void deleteAction( TreePath currentSelection ) {
		if (currentSelection != null) {
			_currentNode = (MutableTreeItem)currentSelection.getLastPathComponent();
			MutableTreeItem parentNode = (MutableTreeItem)_currentNode.getParent();
			if (_currentNode != _rootNode && parentNode != null) {
				_treeModel.removeNodeFromParent(_currentNode);
				_currentNode = null;
			}
		}
	}

	/**
	 * Saves the current tree of operators.
	 */
	protected final void saveAction( ) {
		store(_rootNode);
		setVisible(false);
		dispose();
	}

	/**
	 * Stores the specified subtree in the current explorer$
	 * tree and in the related file.
	 * 
	 * @param node the root node of the modified subtree.
	 * @return true if all preferences has been saved.
	 */
	protected abstract boolean store( MutableTreeItem node );

	/**
	 * Just disposes the dialog box,
	 * without saving.
	 */
	protected final void cancelAction( ) {
		setVisible(false);
		dispose();
	}

	/**
	 * The class <code>CategoryEditor</code> is a dialog box
	 * for editing category information.
	 */
	private final class CategoryEditor extends JDialog {
		/** The category of the operator. */
		private String _currentCategory;
		/** The current name of category. */ 
		private JTextField _textfield1;

		/**
		 * Creates and shows the dialog box.
		 * 
		 * @param parent the parent frame.
		 */
		public CategoryEditor( JDialog parent ) {
			super(parent, Ariane.lang.getString("TreeEditor.categoryTitle"), true);
			_currentCategory = (String)_currentNode.getUserObject();
			addWindowListener(new WindowAdapter() {
				// Redirects the window closing event to endDialog();
				@Override
				public final void windowClosing(WindowEvent e) {
					saveTree();
				}
			});
			displayContent();
			setLocationRelativeTo(parent);
		}

		/**
		 * Sets the content of the dialog box.
		 */
		private final void displayContent( ) {
			JPanel textPanel = new JPanel();
			textPanel.setLayout(new FlowLayout());

			textPanel.add(new JLabel(Ariane.lang.getString("TreeEditor.categoryNameText")));
			_textfield1 = new JTextField(30);
			_textfield1.setText(_currentCategory);
			textPanel.add(_textfield1);
			JButton validateButton = new JButton(Ariane.lang.getString("TreeEditor.validate"));
			JPanel buttonPanel = new JPanel();
			buttonPanel.add(validateButton);
			validateButton.addActionListener( new ActionListener() {
				@Override
				public final void actionPerformed(ActionEvent e) {
					saveTree();
				}
			} );
			getContentPane().setLayout(new BorderLayout());
			getContentPane().add(textPanel, BorderLayout.NORTH);
			getContentPane().add(buttonPanel, BorderLayout.SOUTH);
			getRootPane().setDefaultButton(validateButton);
			pack();
		}

		/**
		 * Validates informations.
		 */
		private final void saveTree( ) {
			try {
				_currentCategory = _textfield1.getText();
				_currentNode.setUserObject(_currentCategory);
				for (int i=0; i< _currentNode.getChildCount(); i++) {
					MutableTreeItem node = (MutableTreeItem)_currentNode.getChildAt(i);
					TreeItem treeItem = (TreeItem)node.getUserObject();
					treeItem.setCategory(_currentCategory);
				}
				_treeModel.nodeChanged(_currentNode);
				setVisible(false);
			} catch(Exception e) {
				_workspace.printErrorMessage(e.getLocalizedMessage());
			}
			_textfield1=null;
			_currentCategory=null;
			dispose();
		}
	}
}