/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.utils;

import java.awt.AWTException;
import java.awt.Dimension;
import java.awt.image.ColorModel;
import java.awt.image.ImageConsumer;
import java.awt.image.ImageProducer;
import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Random;

/**
 * Extends ImageProducer to decode Pandore images. This code is based
 * on the book "Java AWT references" by John Zukowski.
 * 
 * Custom usage: PandoreImageDecoder pan= new PandoreImageDecoder();
 * pan.readImage(imagebyte); img = Toolkit.getDefaultToolkit().createImage(pan);
 * 
 * @author Régis CLOUARD - Jan 03, 2001
 * @version Régis CLOUARD - Nov 07, 2002 (Pandore images).
 * @version Régis CLOUARD - Jan 16, 2003 (Pandore 4).
 */
public class PandoreImageDecoder implements ImageProducer {
	/** list of Pandore types. */
	//	private static final int Po_Collection = 1;
	//	private static final int Po_Img1duc = 2;
	//	private static final int Po_Img1dsl = 3;
	//	private static final int Po_Img1dsf = 4;
	/** 2D char gray image. */
	private static final int Po_Img2duc = 5;
	/** 2D long gray image. */
	private static final int Po_Img2dsl = 6;
	/** 2D float gray image. */
	private static final int Po_Img2dsf = 7;
	private static final int Po_Img3duc = 8;
	private static final int Po_Img3dsl = 9;
	private static final int Po_Img3dsf = 10;
	//	private static final int Po_Reg1d = 11;
	/** 2D region image. */
	private static final int Po_Reg2d = 12;
	private static final int Po_Reg3d = 13;
	//	private static final int Po_Graph2d = 14;
	//	private static final int Po_Graph3d = 15;
	/** 2D char color image. */
	private static final int Po_Imc2duc = 16;
	/** 2D long color image. */
	private static final int Po_Imc2dsl = 17;
	/** 2D float color image. */
	private static final int Po_Imc2dsf = 18;
	private static final int Po_Imc3duc = 19;
	private static final int Po_Imc3dsl = 20;
	private static final int Po_Imc3dsf = 21;
	//	private static final int Po_Imx1duc = 22;
	//	private static final int Po_Imx1dsl = 23;
	//	private static final int Po_Imx1dsf = 24;
	//	private static final int Po_Imx2duc = 25;
	//	private static final int Po_Imx2dsl = 26;
	//	private static final int Po_Imx2dsf = 27;
	//	private static final int Po_Imx3duc = 28;
	//	private static final int Po_Imx3dsl = 29;
	//	private static final int Po_Imx3dsf = 30;
	/** The magic number. */
	private static final String PANDOREMAGIC = "PANDORE04";
	/* Header attributes */
	/** Magic number : PO_MAGIC */
	private String magic = null;
	/** object type */
	private int potype = 0;
	/** Is it little or big endian encoding. */
	private boolean _reverseByteOrder = false;

	private int _maxSize;

	/**
	 * Build a Pandore image decoder in order to build
	 * an image of maxSize from a file.
	 * @param maxSize max size for both width and height.
	 */
	public PandoreImageDecoder( int maxSize ) {
		_maxSize= maxSize;
	}

	/**
	 * Reads Pandore file header and checks consistency.
	 * 
	 * @param dis buffered image data;
	 * @return identifier of the Pandore objet if 'dis' is a regular Pandore
	 *         file or -1 otherwise.
	 * @exception AWTException when file is not a regular Pandore file.
	 * @exception IOException in case of io exception.
	 */
	private int readHeader( DataInputStream dis ) throws AWTException, IOException {
		byte[] buf = new byte[10];

		// Magic number.
		dis.read(buf, 0, 9);
		magic = new String(buf, 0, 9);
		dis.skipBytes(1); // skip '\n'
		if (!(PANDOREMAGIC.equals(magic))) {
			throw (new AWTException("Invalid Pandore Format " + magic));
		}

		// skip unused bytes.
		dis.skipBytes(2);

		// Hint: When typeoj is wrong -> reverse byte order.
		potype = dis.readInt();
		if (potype > 65535) {
			_reverseByteOrder = true;
			potype = swap(potype);
		}

		// Creator name
		dis.read(buf, 0, 8);
		dis.skipBytes(1); // skip '\0'

		// Date
		dis.read(buf, 0, 9);
		dis.skipBytes(1); // '\0';

		// Skip unused bytes
		dis.skipBytes(1);

		return potype;
	}

	/** Since done in-memory; only one consumer */
	private ImageConsumer _consumer;

	/**
	 * Problem with big-endian and little-endian:
	 * direct reading or reverse byte order?
	 */
	boolean loadError = false;

	private Dimension _thumbnail;

	/** Array of pixels. */
	private int _store[][];

	/**
	 * Format of Pandore file is single pass/frame, w/ complete scan lines in
	 * order
	 */
	private static int PandoreHints = (
			ImageConsumer.TOPDOWNLEFTRIGHT |
			ImageConsumer.COMPLETESCANLINES |
			ImageConsumer.SINGLEPASS |
			ImageConsumer.SINGLEFRAME
	);

	/**
	 * There is only a single consumer. When it registers, produce image. On
	 * error, notify consumer.
	 * 
	 * @param ic new consumer to be added.
	 */
	@Override
	public synchronized void addConsumer( ImageConsumer ic ) {
		_consumer = ic;
		try {
			produce();
		} catch (Exception e) {
			if (_consumer != null)
				_consumer.imageComplete(ImageConsumer.IMAGEERROR);
		}
		_consumer = null;
	}

	/**
	 * If consumer passed to routine is single consumer, return true, else false.
	 * 
	 * @param ic consumer.
	 * @return true if it is a consumer.
	 */
	@Override
	public synchronized boolean isConsumer( ImageConsumer ic ) {
		return (ic == _consumer);
	}

	/**
	 * Disables consumer if currently consuming.
	 * 
	 * @param ic consumer.
	 */
	@Override
	public synchronized void removeConsumer( ImageConsumer ic ) {
		if (_consumer == ic)
			_consumer = null;
	}

	/**
	 * Production is done by adding consumer.
	 * 
	 * @param ic consumer.
	 */
	@Override
	public void startProduction( ImageConsumer ic ) {
		addConsumer(ic);
	}

	/**
	 * Requests, on behalf of the <code>ImageConsumer</code>, 
	 * that the <code>ImageProducer</code> attempt to resend 
	 * the image data one more time in TOPDOWNLEFTRIGHT order 
	 * so that higher quality conversion algorithms which 
	 * depend on receiving pixels in order can be used to 
	 * produce a better output version of the image.
	 * 
	 * @param ic consumer.
	 */
	@Override
	public void requestTopDownLeftRightResend( ImageConsumer ic ) {
		// Not needed. The data is always in this format.
	}

	/**
	 * Production Process: Prerequisite: Image already read into store array.
	 * (readImage) props / width / height already set (readImage) Assumes RGB
	 * Color Model - would need to filter to change. Sends Pandore Image data to
	 * consumer. Pixels sent one row at a time.
	 */
	private void produce( ) {
		ColorModel cm = ColorModel.getRGBdefault();
		if (_consumer != null) {
			if (loadError) {
				_consumer.imageComplete(ImageConsumer.IMAGEERROR);
			} else {
				_consumer.setDimensions(_thumbnail.width, _thumbnail.height);
				_consumer.setColorModel(cm);

				_consumer.setHints(PandoreHints);
				for (int y = 0; y < _thumbnail.height; y++) {
					_consumer.setPixels(0, y, _thumbnail.width, 1, cm, _store[y], 0, _thumbnail.width);
				}
				_consumer.imageComplete(ImageConsumer.STATICIMAGEDONE);
			}
		}
	}

	/**
	 * Swaps integer.
	 * 
	 * @param i the integer to be swapped.
	 * @return the swapped integer.
	 */
	public int swap( int i ) {
		int byte0 = i & 0xff;
		int byte1 = (i >> 8) & 0xff;
		int byte2 = (i >> 16) & 0xff;
		int byte3 = (i >> 24) & 0xff;
		// swap the byte order
		return (byte0 << 24) | (byte1 << 16) | (byte2 << 8) | byte3;
	}

	/**
	 * Allows reading from a filename.
	 * 
	 * @param filename complete name of the image file.
	 */
	public void readImage( String filename ) {
		try {
			FileInputStream fileImage = new FileInputStream(filename);
			if (fileImage.available() > 0) {
				readImage(fileImage);
			}
			fileImage.close();
		} catch (IOException e) {
		}
	}


	/**
	 * Reads an integer format the input stream.
	 * If needed inverts the bytes.
	 * 
	 * @param dis the input stream.
	 * @return the data array.
	 * @throws IOException in case if IO error.
	 */
	private int readInt( DataInputStream dis ) throws IOException {
		byte[] row = new byte[4];
		for (int b=0; b<4; b++)
			row[b]=dis.readByte();
		return parseBytes(row,0,4);
	}

	/**
	 * Converts Pandore pixel to color that can be used by Java.
	 * 
	 * @param row line of pixels
	 * @param x column number of the pixel.
	 * @param bytes bytes per pixel.
	 * @return color to be displayed
	 */
	private int parseBytes( byte[] row, int x, int bytes ) {
		int[] colorByte = new int[4];

		for (int b = 0; b < bytes; b++) {
			colorByte[b] = row[(x*bytes) + b];
			// Problem signed -> unsigned.
			if (colorByte[b] < 0)
				colorByte[b] += 256;
		}
		int color = 0;

		if (_reverseByteOrder) {
			for (int b = 0; b < bytes; b++) {
				int shift = b * 8;
				color |= (colorByte[b] << shift);
			}
		} else {
			for (int b = 0; b < bytes; b++) {
				int shift = (bytes-1-b) * 8;
				color |= (colorByte[b] << shift);
			}
		}

		return color;
	}

	/**
	 * Allows reading to be from internal byte array, in addition to
	 * disk/socket.
	 * 
	 * @param b the byte array.
	 */
	public void readImage( byte b[] ) {
		readImage(new ByteArrayInputStream(b));
	}

	private Dimension computeThumbnailSize( int height, int width ) {
		Dimension d = new Dimension();
		if (width >= height) {
			if (width>_maxSize) {
				int ratio = 1+ width / _maxSize;
				d.width = width / ratio;
				d.height = height / ratio;
			}
		} else {
			if (height>_maxSize) {
				int ratio = 1+ height / _maxSize;
				d.width = width / ratio;
				d.height = height / ratio;
			}
		}
		return d;
	}

	/**
	 * Reads gray data image. First of all read the image size.
	 * 
	 * @param dis buffered data to parse.
	 * @return the data array.
	 * @throws IOException in case of IO errors.
	 */
	public int[][] read2dGrayData( DataInputStream dis ) throws IOException {
		// read image size.
		readInt(dis); // bands = 1 !
		int height = readInt(dis);
		int width = readInt(dis);

		// read data
		_thumbnail = computeThumbnailSize(height, width);
		int[][] store = new int[_thumbnail.height][_thumbnail.width];
		int stepX = width / _thumbnail.width;
		int stepY = height / _thumbnail.height;
		byte row[];
		switch (potype) {
		case Po_Img2duc:
			row = new byte[width];
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 1);

					store[y][x] = (0xff << 24) | (color << 16);
					store[y][x] |= (color << 8);
					store[y][x] |= color;
				}
			}
			break;

		case Po_Img2dsl:
			row = new byte[width * 4];
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*4*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 4);
					store[y][x] = (0xff<< 24) | (color << 16) |
					(color << 8) | color;
				}
			}
			break;

		case Po_Reg2d:
			Random rand = new Random(0);
			int nlabels = readInt(dis);
			int[] lut = new int[nlabels];
			int bytes;
			if (nlabels < 256)
				bytes = 1;
			else if (nlabels < 65536)
				bytes = 2;
			else
				bytes = 4;

			// Colors are set randomly except 0
			lut[0] = 0xff << 24;
			for (int i = 1; i <nlabels; i++)
				lut[i] = (0xff << 24) | (rand.nextInt(256) << 16)
				| (rand.nextInt(256) << 8) | rand.nextInt(256);

			row = new byte[width * bytes];
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*bytes*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, bytes);
					if (color<nlabels)
						store[y][x] = lut[color];
				}
			}
			break;

		case Po_Img2dsf:
			row = new byte[width * 4];
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*4*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 4);
					float toto = Float.intBitsToFloat(color);
					color = (int)toto;
					store[y][x] = (0xff<< 24) | (color << 16) |
					(color << 8) | color;
				}
			}
			break;
		}
		return store;
	}

	/**
	 * Reads color data image.
	 * 
	 * @param dis the input stream.
	 * @return the data array.
	 * @throws IOException in case of IO error.
	 */
	public int[][] read2dColorData( DataInputStream dis ) throws IOException {
		// read image size.
		readInt(dis); // bands = 3;
		int height = readInt(dis);
		int width = readInt(dis);

		// ColorSpace.
		dis.readInt();

		// read data
		_thumbnail = computeThumbnailSize(height, width);
		int[][] store = new int[_thumbnail.height][_thumbnail.width];
		int stepX = width / _thumbnail.width;
		int stepY = height / _thumbnail.height;
		byte row[];

		switch (potype) {

		case Po_Imc2duc: {
			row = new byte[width];
			// red band
			int xx=0;
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*(stepY-1));
				xx+=stepY;
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 1);
					store[y][x] = (0xff << 24) | (color << 16);
				}
			}
			dis.skipBytes(width*((height-(stepY*_thumbnail.height))));
			// green band
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 1);
					store[y][x] |= (color << 8);
				}
			}
			dis.skipBytes(width*((height-(stepY*_thumbnail.height))));
			// blue band
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 1);
					store[y][x] |= color;
				}
			}
		}
		break;

		case Po_Imc2dsl:
			row = new byte[width * 4];
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*4*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 4);
					store[y][x] = (0xff << 24) | (color << 16);
				}
			}
			dis.skipBytes(width*((height-(stepY*_thumbnail.height))));

			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*4*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 4);
					store[y][x] |= (color << 8);
				}
			}
			dis.skipBytes(width*((height-(stepY*_thumbnail.height))));

			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*4*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 4);
					store[y][x] |= color;
				}
			}
			break;

		case Po_Imc2dsf:
			break;
		}
		return store;
	}

	/**
	 * Reads gray data image.
	 * Note that this method is used for displaying
	 * image thumbnail so it does not read all the data but
	 * only the slice in the middle of the image..
	 * 
	 * @param dis buffered data to parse.
	 * @return the data array.
	 * @throws IOException in case of IO errors.
	 */
	public int[][] read3dGrayData( DataInputStream dis ) throws IOException {
		// read image size.
		readInt(dis); // read the number of bands : always 1 !
		final int depth = readInt(dis);
		final int height = readInt(dis);
		final int width = readInt(dis);

		_thumbnail = computeThumbnailSize(height, width);
		int[][] store = new int[_thumbnail.height][_thumbnail.width];
		int stepX = width / _thumbnail.width;
		int stepY = height / _thumbnail.height;
		byte row[];

		// read data
		switch (potype) {
		case Po_Img3duc:
			row = new byte[width];
			dis.skipBytes((depth/2)*width*height);
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 1);

					store[y][x] = (0xff << 24) | (color << 16);
					store[y][x] |= (color << 8);
					store[y][x] |= color;
				}
			}
			break;

		case Po_Img3dsl:
			row = new byte[width*4];
			dis.skipBytes((depth/2)*width*4*height);
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*4*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 4);

					store[y][x] = (0xff << 24) | (color << 16);
					store[y][x] |= (color << 8);
					store[y][x] |= color;
				}
			}
			break;

		case Po_Img3dsf:
			row = new byte[width*4];
			dis.skipBytes((depth/2)*width*4*height);
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*4*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 4);
					float toto = Float.intBitsToFloat(color);
					color = (int)toto;

					store[y][x] = (0xff << 24) | (color << 16);
					store[y][x] |= (color << 8);
					store[y][x] |= color;
				}
			}
			break;

		case Po_Reg3d:
			Random rand = new Random(0);
			int nlabels = readInt(dis);
			int[] lut = new int[nlabels];
			int bytes;
			if (nlabels < 256)
				bytes = 1;
			else if (nlabels < 65536)
				bytes = 2;
			else
				bytes = 4;

			// Colors are set randomly except 0
			lut[0] = 0xff << 24;
			for (int i = 1; i <nlabels; i++)
				lut[i] = (0xff << 24) | (rand.nextInt(256) << 16)
				| (rand.nextInt(256) << 8) | rand.nextInt(256);

			row = new byte[width*bytes];
			dis.skipBytes((depth/2)*width*height);
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*bytes*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, bytes);
					if (color<nlabels)
						store[y][x] = lut[color];
				}
			}
			break;
		}
		return store;
	}

	/**
	 * Reads color data image.
	 * 
	 * @param dis the input stream.
	 * @return the data array.
	 * @throws IOException in case of IO error.
	 */
	public int[][] read3dColorData( DataInputStream dis ) throws IOException {
		// read image size.
		readInt(dis); // bands = 1 !
		int depth = readInt(dis);
		int height = readInt(dis);
		int width = readInt(dis);

		// read data
		_thumbnail = computeThumbnailSize(height, width);
		int[][] store = new int[_thumbnail.height][_thumbnail.width];
		int stepX = width / _thumbnail.width;
		int stepY = height / _thumbnail.height;
		byte row[];
		switch (potype) {
		case Po_Imc3duc:
			row = new byte[width];
			dis.skipBytes((depth/2)*width*height);
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 1);
					store[y][x] = (0xff << 24) | (color << 16);
				}
			}
			dis.skipBytes(width*((height-(stepY*_thumbnail.height))));

			dis.skipBytes((depth-1)*width*height);
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 1);
					store[y][x] |= (color << 8);
				}
			}
			dis.skipBytes(width*((height-(stepY*_thumbnail.height))));

			dis.skipBytes((depth -1)*width*height);
			for (int y = 0; y < _thumbnail.height; y++) {
				dis.readFully(row);
				dis.skipBytes(width*(stepY-1));
				for (int x = 0; x < _thumbnail.width; x++) {
					int color = parseBytes(row, x*stepX, 1);
					store[y][x] |= color;
				}
			}
			break;

		case Po_Imc3dsl:
			break;

		case Po_Imc3dsf:
			break;
		}

		return store;
	}

	/**
	 * Reads image data from Stream.
	 * Parses data for Pandore format Close input
	 * stream when done.
	 * 
	 * @param is input stream.
	 */
	public void readImage( InputStream is ) {
		DataInputStream dis = null;
		BufferedInputStream bis = null;

		try {
			// Try a supported image.
			bis = new BufferedInputStream(is);
			dis = new DataInputStream(bis);

			readHeader(dis);

			switch (potype) {
			case Po_Img2duc:
			case Po_Img2dsl:
			case Po_Img2dsf:
			case Po_Reg2d:
				_store = read2dGrayData(dis);
				break;

			case Po_Imc2duc:
			case Po_Imc2dsl:
			case Po_Imc2dsf:
				_store = read2dColorData(dis);
				break;

			case Po_Img3duc:
			case Po_Img3dsl:
			case Po_Img3dsf:
			case Po_Reg3d:
				_store = read3dGrayData(dis);
				break;

			case Po_Imc3duc:
			case Po_Imc3dsl:
			case Po_Imc3dsf:
				_store = read3dColorData(dis);
				break;
			}
		} catch (Exception io) {
			loadError = true;
		} finally {
			try {
				if (dis != null) dis.close();
				if (bis != null) bis.close();
				if (is != null) is.close();
			} catch (IOException io) {
			}
		}
	}
}