/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.utils;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.ui.ILocalizable;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Cursor;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

import javax.swing.ImageIcon;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

/**
 * The class <code>Tree</code> contains all the available graph items
 * and organized them in a tree.
 * This tree is displayed in the explorer from where the user can drag and
 * drop items into the graph editor.
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Sep 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 03, 2004
 * @version Régis CLOUARD - Dec 12, 2008
 * @version Régis CLOUARD - July 17, 2010 (add the macro-operator library)
 */
public final class Tree extends JPanel implements DragSourceListener, DragGestureListener, ILocalizable {
	/** The displayed tree. */
	private JTree _jtree;

	/** The root of the tree item. */
	private MutableTreeItem _rootNode;

	/** The model of the tree. */
	private DefaultTreeModel _treeModel;

	/** The root of the user-defined operators subtree. */
	private MutableTreeItem _userOperatorSubtree;

	/** The root of the common operators subtree. */
	private MutableTreeItem _commonOperatorSubtree;

	/** The root of the user-defined subroutines subtree. */
	private MutableTreeItem _subRoutineSubtree;

	/** The root of the built-in item subtree. */
	private MutableTreeItem _builtinSubtree;

	/** The drag site. */
	private DragSource _dragSource = null;

	/** The name of the file that contains the documentation index. */
	private String _indexDoc = null;

	/** The cursor for drag and drop. */
	private static final Cursor _itemCursor;

	static {
		_itemCursor=Toolkit.getDefaultToolkit().createCustomCursor(new ImageIcon(Tree.class.getClassLoader().getResource(Ariane.iconDir+"itemcursor.png")).getImage(), new Point(0,0), "itemCursor");
	}

	/**
	 * Creates the tree of available graph items from
	 * the given XML tree.
	 * 
	 * @param readItems the XML tree of graph items.
	 */
	public Tree( LinkedList<XMLElement> readItems ) {
		initWidget();
		if (readItems != null) {
			setCommonOperators(readItems);
		}
		setInternationalText();
	}

	/**
	 * Creates permanent items in the tree such as
	 * Constant, ControlStructure, MacroOperator ...).
	 * Other items are read from the configuration files.
	 */
	private final void initWidget( ) {
		_rootNode = new MutableTreeItem("Library",true);

		// Sub categories.
		_userOperatorSubtree = new MutableTreeItem(null,true);
		_subRoutineSubtree = new MutableTreeItem(null,true);
		_commonOperatorSubtree = new MutableTreeItem(null,true);
		_builtinSubtree = new MutableTreeItem(null,true);
		MutableTreeItem fileItem;

		// Add the various built-in items.
		TreeItem treeItem;
		treeItem = new TreeItem(TreeItem.Item.LOADER, "loader", "loader.html", null, null,0,0,1,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.SAVER, "saver", "saver.html", null, null,1,0,0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.VIEWER, "viewer","viewer.html",null, null,1,0,0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.FILENAME, "filename", "filename.html",null, null,0,0,0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONSTANT, "constant", "constant.html",null, null,0,0,0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.SCRIPT, "script", "script.html",null, null,0,0,0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLEVAL, "eval", "eval.html",null, null,0,2,0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLFORMAT, "format","format.html",null, null,0,2,0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.SYSTEM, "system", "system.html", null, null,0,0,0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLIF, "if","if.html",null, null,0, 0, 0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLSELECT, "select","select.html",null, null,0,0,1,"$0:index");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLMERGE, "merge","merge.html",null, null,4,0,0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLFOR, "for","for.html",null, null,3,0,0, "$0:initial value;$1:final value;$2:increment");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.CONTROLWHILE, "while","while.html",null, null,2,0,0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.MACROOPERATOR, "macro-operator", "macro.html", null, null,0,0,0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		treeItem = new TreeItem(TreeItem.Item.NOTE, "note", "note.html", null, null,0,0,0,"");
		fileItem = new MutableTreeItem(treeItem, false);
		_builtinSubtree.add(fileItem);

		_rootNode.add(_builtinSubtree);
		_rootNode.add(_commonOperatorSubtree);
		_rootNode.add(_subRoutineSubtree);
		_rootNode.add(_userOperatorSubtree);

		// Creates explorer with a Jtree.
		_treeModel = new DefaultTreeModel(_rootNode);
		_jtree = new JTree(_treeModel);
		_jtree.setFont(_jtree.getFont().deriveFont(10.0f));

		_jtree.getSelectionModel().setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);

		ClassLoader classLoader = getClass().getClassLoader();
		_jtree.putClientProperty("JTree.lineStyle", "Angled");
		DefaultTreeCellRenderer renderer = new DefaultTreeCellRenderer();
		renderer.setLeafIcon(new ImageIcon(classLoader.getResource(Ariane.iconDir+"item.png")));
		//		renderer.setOpenIcon(new ImageIcon(classLoader.getResource(Ariane.iconDir+"opencategory.png")));
		//		renderer.setClosedIcon(new ImageIcon(classLoader.getResource(Ariane.iconDir+"category.png")));
		renderer.setOpenIcon(null);
		renderer.setClosedIcon(null);
		_jtree.setCellRenderer(renderer);

		_dragSource = DragSource.getDefaultDragSource();
		_dragSource.createDefaultDragGestureRecognizer(_jtree, DnDConstants.ACTION_MOVE, this);

		setLayout(new GridLayout());
		setBackground(_jtree.getBackground());

		// Listeners event from the tree.
		MouseListener ml = new MouseAdapter() {
			// Case of double-click -> display manual.
			@Override
			public final void mousePressed(MouseEvent e) {
				if (e.getClickCount() > 1) {
					TreePath selectedPath = _jtree.getPathForLocation(e.getX(),e.getY());
					if (selectedPath != null) {
						MutableTreeItem selectedNode = (MutableTreeItem)selectedPath.getLastPathComponent();
						if ((selectedNode.isLeaf()) && (selectedNode.getUserObject() instanceof TreeItem)){
							String doc = ((TreeItem)selectedNode.getUserObject()).getDocURL();
							if (doc != null) {
								Ariane.workspace.displayMan(Ariane.lang.getString("HTMLViewer.manual")+" "+selectedNode.toString(),doc, null);
							}
						}
					}
				}
			}
		};
		_jtree.addMouseListener(ml);

		// Discard key CTRL-A (used by the Graph).
		_jtree.getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_A, KeyEvent.CTRL_MASK),"none");

		add(_jtree);
	}

	/**
	 * Highlights all operators whose name
	 * contains the specified pattern.
	 * 
	 * @param pattern a part of operator name.
	 * @return true if some operators has been found.
	 */
	public final boolean search( String pattern ) {
		ArrayList<TreePath> paths=search(_rootNode,pattern);
		if (paths.size()>0) {
			TreePath[] pathsTab = new TreePath[paths.size()];
			for (int i=0; i< paths.size(); i++) {
				pathsTab[i] = paths.get(i);
			}
			// why paths.toArray(); doesn't work ?
			_jtree.setSelectionPaths(pathsTab);
			return true;
		}
		return false;
	}

	/**
	 * Returns the tree item with the specified name
	 * or null.
	 * 
	 * @param name a tree item name.
	 * @param category the related category (or null).
	 * @return the tree item or null.
	 */
	public final TreeItem getTreeItem( String name, String category ) {
		return getNode(_rootNode, name, category);
	}

	/**
	 * Returns the node with the specified name.
	 * 
	 * @param node the root node or the search.
	 * @param name the name to search.
	 * @param category the related category.
	 * @return the path to the node or null.
	 */
	private TreeItem getNode( MutableTreeItem node, String name, String category ) {
		for (Enumeration<?> e = node.children() ; e.hasMoreElements() ;) {
			MutableTreeItem subnode = (MutableTreeItem)e.nextElement();
			if (subnode.isLeaf()) {
				String operatorName = ""+subnode.getUserObject();
				if (operatorName.equals(name)) {
					TreeItem treeItem = (TreeItem)subnode.getUserObject();
					if (category==null || category.equals(treeItem.getCategory())) {
						return treeItem;
					}
				}
			} else {
				TreeItem t = getNode(subnode,name,category);
				if (t != null) { return t; }
			}
		}
		return null;
	}

	/**
	 * Returns the list of path that locate all
	 * node whose name contains the specified
	 * pattern.
	 * 
	 * @param node the root node or the search.
	 * @param pattern a string.
	 * @return the list of paths.
	 */
	public final ArrayList<TreePath> search( MutableTreeItem node, String pattern ) {
		ArrayList<TreePath> paths = new ArrayList<TreePath>();
		for (Enumeration<?> e = node.children() ; e.hasMoreElements() ;) {
			MutableTreeItem subnode = (MutableTreeItem)e.nextElement();
			if (subnode.isLeaf()) {
				String operatorName = ""+subnode.getUserObject();
				if (operatorName.contains(pattern)) {
					paths.add(new TreePath(subnode.getPath()));
				}
			} else {
				paths.addAll(search(subnode,pattern));
			}
		}
		return paths;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void setInternationalText( ) {
		_subRoutineSubtree.setUserObject(Ariane.lang.getString("Tree.subroutines"));
		_userOperatorSubtree.setUserObject(Ariane.lang.getString("Tree.personalOperators"));
		_commonOperatorSubtree.setUserObject(Ariane.lang.getString("Tree.operators"));
		_builtinSubtree.setUserObject(Ariane.lang.getString("Tree.builtinItems"));
		_treeModel.reload();
	}

	/**
	 * Gets the documentation index page.
	 * 
	 * @return the relative path to the page.
	 */
	public final String getIndexDoc( ) {
		return _indexDoc;
	}

	/**
	 * Returns the list of personal operators
	 * as an XML tree.
	 * 
	 * @return a LinkedList of operators.
	 */
	public final LinkedList<TreeItem> getUserOperatorList( ) {
		LinkedList<TreeItem> operatorList= new LinkedList<TreeItem>();
		Enumeration<?> node=_userOperatorSubtree.depthFirstEnumeration();
		while(node.hasMoreElements()) {
			MutableTreeItem leaf=(MutableTreeItem)node.nextElement();
			if (leaf.getUserObject() instanceof TreeItem) {
				operatorList.add((TreeItem)leaf.getUserObject());
			}
		}
		return operatorList;
	}

	/**
	 * Returns the tree of personal operators.
	 * 
	 * @return the root of the tree.
	 */
	public final MutableTreeItem getUserOperatorTree( ) {
		return _userOperatorSubtree;
	}

	/**
	 * Returns the tree of subroutines.
	 * 
	 * @return the root of the tree.
	 */
	public final MutableTreeItem getSubroutineTree( ) {
		return _subRoutineSubtree;
	}

	/**
	 * Returns the tree of subroutines.
	 * 
	 * @return the root of the tree.
	 */
	public final MutableTreeItem getCommonOperatorTree( ) {
		return _commonOperatorSubtree;
	}

	/**
	 * Sets a new version for the tree of personal operators.
	 * Deletes the previous one first.
	 * 
	 * @param newTree the new tree.
	 */
	public final void setUserOperatorTree( MutableTreeItem newTree ) {
		int i = _treeModel.getIndexOfChild(_rootNode, _userOperatorSubtree);
		_treeModel.removeNodeFromParent(_userOperatorSubtree);
		_userOperatorSubtree = newTree;
		_treeModel.insertNodeInto(_userOperatorSubtree, _rootNode, i);
	}

	/**
	 * Sets a new version for the tree of personal operators.
	 * Deletes the previous one first.
	 * 
	 * @param newTree the new tree.
	 */
	public final void setSubroutineTree( MutableTreeItem newTree ) {
		int i = _treeModel.getIndexOfChild(_rootNode, _subRoutineSubtree);
		_treeModel.removeNodeFromParent(_subRoutineSubtree);
		_subRoutineSubtree = newTree;
		_treeModel.insertNodeInto(_subRoutineSubtree, _rootNode, i);
	}

	/**
	 * Sets the specified tree as the operator tree.
	 * 
	 * @param readItems the XML tree of graph items.
	 */
	public final void setCommonOperators( LinkedList<XMLElement> readItems ) {
		LinkedList<TreeItem> treeItems = load(readItems);
		addCommonOperators(treeItems);
	}

	/**
	 * Adds the list of common operators stored in an XML tree
	 * in the explorer.
	 * 
	 * @param operatorList the XML list of operators.
	 */
	private void addCommonOperators( LinkedList<TreeItem> operatorList ) {
		_commonOperatorSubtree.removeAllChildren();
		addItem(operatorList, _commonOperatorSubtree);
	}

	/**
	 * Adds the list of personal operators stored in an XML tree
	 * in the explorer.
	 * 
	 * @param operatorList the XML list of operators.
	 */
	public final void addUserOperators( LinkedList<XMLElement> operatorList ) {
		addItem(load(operatorList), _userOperatorSubtree);
	}

	/**
	 * Adds the list of subroutines stored in an XML tree
	 * in the explorer.
	 * 
	 * @param operatorList the XML list of subroutines.
	 * @return the root of the tree items.
	 */
	public final MutableTreeItem addSubroutines( LinkedList<XMLElement> operatorList ) {
		addItem(load(operatorList), _subRoutineSubtree);
		return _subRoutineSubtree;
	}

	/**
	 * Adds a new subroutine stored in the explorer.
	 * 
	 * @param item the item to add in the tree.
	 * @throws Exception problem.
	 */
	public final void addSubroutine( TreeItem item ) throws Exception {
		LinkedList<TreeItem> treeItems= new LinkedList<TreeItem>();
		treeItems.add(item);
		addItem(treeItems, _subRoutineSubtree);
	}

	/**
	 * Returns the list of subroutines
	 * as an XML tree.
	 * 
	 * @return a LinkedList of subroutines.
	 */
	public final LinkedList<TreeItem> getSubroutineList( ) {
		LinkedList<TreeItem> subRoutineList= new LinkedList<TreeItem>();
		Enumeration<?> node=_subRoutineSubtree.depthFirstEnumeration();
		while(node.hasMoreElements()) {
			MutableTreeItem leaf=(MutableTreeItem)node.nextElement();
			if (leaf.getUserObject() instanceof TreeItem) {
				subRoutineList.add((TreeItem)leaf.getUserObject());
			}
		}
		return subRoutineList;
	}

	/**
	 * Returns the list of subroutine categories.
	 * 
	 * @return a list of category name.
	 */
	@SuppressWarnings("unchecked")
	public final Vector<String> getSubroutineCategories( ) {
		Vector<String> l=new Vector<String>();
		for (Enumeration<TreeItem> e = _subRoutineSubtree.children(); e.hasMoreElements();) {
			l.add(""+e.nextElement());
		}
		return l;
	}

	/**
	 * Adds the list of items <i>itemList</i> in the tree
	 * under the category <i>node</i>.
	 * 
	 * @param itemList the list of items.
	 * @param node the root of the future subtree (category).
	 */
	private void addItem( LinkedList<TreeItem> itemList, MutableTreeItem node ) {
		MutableTreeItem item = null;
		for ( TreeItem treeItem : itemList ) {
			item = new MutableTreeItem(treeItem, false);
			String categoryName = treeItem.getCategory();
			MutableTreeItem category=node.getNode(categoryName);
			if (category == null) { // create a new category.
				category = new MutableTreeItem( categoryName,true);
				node.add(category); // new subroutine
			}
			if (category.getNode(treeItem.getName())!=null
					&& Ariane.lang!=null
					&& Ariane.frame != null
					&& Ariane.frame.isVisible()) {
				if (JOptionPane.showConfirmDialog(null,
						treeItem.getName()+" "+Ariane.lang.getString("Tree.dialog.overwriteMessage"),
						Ariane.lang.getString("Tree.dialog.overWriteTitle"),
						JOptionPane.YES_NO_OPTION)==JOptionPane.YES_OPTION) {
					category.remove(category.getNode(treeItem.getName()));
					category.add(item); // replace existing subroutine
					_treeModel.reload(node);
					Ariane.workspace.printWarning(Ariane.lang.getString("CompositeItem.message2"));
				}
				continue;
			} else {
				category.add(item);
				_treeModel.reload(node);
			}
		}
	}

	///// 
	///// D R A G   &   D R O P
	/////

	/**
	 * Called when the user has modified the drop gesture.
	 * 
	 * @param event the DragSourceDragEvent.
	 */
	public final void dropActionChanged( DropTargetDragEvent event ) { }

	/**
	 * Invoked to signify that the Drag and Drop operation is complete.
	 * Sets the border to normal looking.
	 * 
	 * @param event the DragSourceDropEvent
	 */
	@Override
	public final void dragDropEnd( DragSourceDropEvent event ) { }

	/**
	 * Called as the cursor's hotspot enters a platform-dependent drop site.
	 * Changes the cursor.
	 * 
	 * @param event the DragSourceDragEvent.
	 */
	@Override
	public final void dragEnter( DragSourceDragEvent event ) { }

	/**
	 * Called as the cursor's hot spot exits a platform-dependent drop site.
	 * Do nothing.
	 * 
	 * @param event the DragSourceEvent.
	 */
	@Override
	public final void dragExit( DragSourceEvent event ) { }

	/**
	 * Called as the cursor's hot spot moves over a platform-dependent
	 * drop site (a Graph or a Slot).
	 * 
	 * @param event the DragSourceDragEvent.
	 */
	@Override
	public final void dragOver( DragSourceDragEvent event ) { }

	/**
	 * Called when the user has modified the drop gesture.
	 * 
	 * @param event the DragSourceDragEvent.
	 */
	@Override
	public final void dropActionChanged( DragSourceDragEvent event ) { }

	/**
	 * A DragGestureRecognizer has detected a platform-dependent drag
	 * initiating gesture and is notifying this listener in order for
	 * it to initiate the action for the user.
	 * 
	 * @param event the gesture that has just occurred.
	 */
	@Override
	public final void dragGestureRecognized( DragGestureEvent event ) {
		// Detect the select tree item.
		TreePath selectedPath = _jtree.getSelectionPath();
		if (selectedPath != null) {
			MutableTreeItem selectedNode=(MutableTreeItem)selectedPath.getLastPathComponent();
			// Is a valid drag only if it is a leaf.
			if (selectedNode.isLeaf() && (selectedNode.getUserObject() instanceof TreeItem)) {
				TreeItem selected = (TreeItem)selectedNode.getUserObject();
				_dragSource.startDrag(event, _itemCursor, selected, this);
			}
		}
	}

	///// 
	///// I N P U T   /   O U P U T
	/////

	/**
	 * Loads the tree from an XML tree.
	 * 
	 * @param xmlTree the XNML tree of items.
	 * @return the list of all operators.
	 */
	private LinkedList<TreeItem> load( LinkedList<XMLElement> xmlTree ) {
		LinkedList<TreeItem> treeItems= new LinkedList<TreeItem>();

		// Discard tag <ARIANE>.
		xmlTree = xmlTree.getFirst().getChildren();

		if (xmlTree==null) return treeItems;
		Iterator<XMLElement> it = xmlTree.iterator();

		XMLElement e;
		while (it.hasNext()) {
			e = it.next();
			if (e.getName().equalsIgnoreCase("OperatorLibrary") || e.getName().equalsIgnoreCase("Library") || e.getName().equalsIgnoreCase("Bibliotheque")) {
				xmlTree = e.getChildren();
				it = xmlTree.iterator();
				while (it.hasNext()) {
					e = it.next();
					// kept for sake of compatibility
					if (e.getName().equalsIgnoreCase("DocPath")) {
						// kept for sake of compatibility
						Ariane.setDocPath(e.getValue());
					} else if (e.getName().equalsIgnoreCase("CommandStatus")) {
						GraphItem.setCommandStatus(e.getValue());
					} else if (e.getName().equalsIgnoreCase("CommandVisu")) {
						GraphItem.setCommandVisu(e.getValue());
					} else if (e.getName().equalsIgnoreCase("Version")) {
						// kept for sake of compatibility
						// Gets the version number of library.
						_rootNode.setUserObject(e.getValue());
						_treeModel.reload(_rootNode);
						if (e.getValue().startsWith("Pandore")) {
							GraphItem.setCommandStatus("pstatus");
							GraphItem.setCommandVisu("pvisu");
						}
					} else if (e.getName().equalsIgnoreCase("Index")){
						// kept for sake of compatibility
						// Gets the documentation index page.
						_indexDoc = e.getValue();
					} else if (e.getName().equalsIgnoreCase("Operator") || e.getName().equalsIgnoreCase("Operateur"))
						try {
							treeItems.add(new TreeItem(e.getChildren()));
						} catch(Exception exc) {
							// Cannot build the branch,
							// so do not build it that's all.
						}
				}
				// subroutines
			} else if (e.getName().equalsIgnoreCase("SubroutineLibrary")) {
				xmlTree = e.getChildren();
				it = xmlTree.iterator();
				while (it.hasNext()) {
					e = it.next();
					// kept for sake of compatibility
					if (e.getName().equalsIgnoreCase("Subroutine"))
						try {
							treeItems.add(new SubroutineTreeItem(e.getChildren()));
						} catch(Exception exc) {
							// Cannot build the branch,
							// so do not build it .. that's all.
						}
				}
			}
		}
		return treeItems;
	}
}