/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.xml;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;

/**
 * A <code>XMLFileOutputStream</code> defines an output stream for
 * XML files.
 *
 * @author Régis CLOUARD - Nov 25, 2002
 * @version Régis CLOUARD - Nov 30, 2004
 */
public final class XMLFileOutputStream extends FileOutputStream {
	/**
	 * Creates and opens a new output stream with the specified file.
	 * Begins the file with the convenient XML directives.
	 * 
	 * @param file the file to be opened.
	 * @param  DTDFilename save the file with the specified dtd filename if not NULL.
	 * @throws FileNotFoundException file not found.
	 * @throws IOException any IO exception.
	 */
	public XMLFileOutputStream( File file, String DTDFilename ) throws FileNotFoundException, IOException {
		super(file);
		if (DTDFilename != null)
			// prolog.
			write("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n<!DOCTYPE ARIANE SYSTEM '"+DTDFilename+"'>\n",0);
		else
			write("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n",0);
		write("<ARIANE version=\"1.1\">\n",0);
	}

	/**
	 * Writes the given string as it is (i.e. without encoding &lt; &amp;).
	 * 
	 * @param string the string to be written.
	 * @param depth the depth in the XML tree.
	 * @throws IOException any IO exception.
	 */
	public final void write( String string, int depth ) throws IOException {
		for(int i=0; i< depth*2; i++)
			super.write(' ');
		super.write(string.getBytes());
	}

	/**
	 * Writes the given string as an encoded string :
	 *  '<' by "&lt;", '&' by "&amp;".
	 *
	 * @param string the string to be written.
	 * @throws IOException any IO exception.
	 */
	private final void writeEncoded( String string ) throws IOException {
		if (string == null) return;
		// Replace character '&' by &amp; '<' by &lt;
		byte[] encoded = string.getBytes("ISO-8859-1");
		for (int i=0; i< encoded.length;i++) {
			switch(encoded[i]){
			case '&' : super.write("&amp;".getBytes()); break;
			case '<' : super.write("&lt;".getBytes()); break;
			default : super.write(encoded[i]); break;
			}
		}
	}

	/**
	 * Writes the given XML tree in the file.
	 * Uses depth to indent the file.
	 * 
	 * @param xmlRoot the root of the tree to be written.
	 * @param depth the current depth.
	 * @throws IOException any IO exception.
	 */
	private final void write( XMLElement xmlRoot, int depth ) throws IOException {
		if (xmlRoot == null) return;
		if (xmlRoot.getChildren() != null) { // A node.
			write("<"+xmlRoot.getName()+">\n",depth);
			for (XMLElement e1 : xmlRoot.getChildren()) {
				write(e1,depth+1);
			}
			write("</"+xmlRoot.getName()+">\n",depth);
		} else {							// A leaf.
			write("<"+xmlRoot.getName()+">", depth);
			writeEncoded(xmlRoot.getValue());
			write("</"+xmlRoot.getName()+">\n",0);
		}
	}

	/**
	 * Builds the begin tag for the given element.
	 * Adds attributes if needed.
	 * 
	 * @param element an XML element.
	 * @throws IOException any IO exception.
	 */
	public final void write( XMLElement element ) throws IOException {
		write(element,0);
	}

	/**
	 * Closes the file and adds the end tag <code>&lt;/ARIANE&gt;</code>.
	 * 
	 * @throws IOException any IO exception.
	 */
	@Override
	public final void close( ) throws IOException {
		write("</ARIANE>",0);
		super.close();
	}
}